//  BMPx - The Dumb Music Player
//  Copyright (C) 2005-2006 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non GPL-compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glib/gi18n.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <cairomm/cairomm.h>
#include <cmath>

#include "main.hpp"
#include "ui_util.hpp"
#include "paths.hpp"

#include "about_dialog.hpp"


namespace
{
  const int mask_alpha_threshold = 128;

  // Animation settings

  const int animation_fps = 24;
  const int animation_frame_period_ms = 1000 / animation_fps;

  const char  *text_font          = "Sans";
  const int    text_size_px       = 15;
  const double text_colour[3]     = { 1.0, 1.0, 1.0 };
  const double text_bottom_margin = 30;
  const double text_fade_in_time  = 0.8;
  const double text_fade_out_time = 0.5;
  const double text_hold_time     = 1.3;
  const double text_time          = text_fade_in_time + text_fade_out_time + text_hold_time;
  const double text_full_alpha    = 0.75;

  const double initial_delay      = 2.0;

  // Credits text
  // Each list of names is sorted in dictionary order, first according
  // to last name, then according to first name.

  const char *credit_text[] =
  {
      N_("<big><b>BMPx is brought to you by</b></big>"),
      "Chong Kai Xiong",
      "Milosz Derezynski",
      "David Le Brun",
      N_("<big><b>With contributions from</b></big>"),
      "Saleem Abdulrasool",
      "Olivier Blin",
      "Roman Bogorodskiy",
      "Shlomi Fish",
      "Kenneth Ostby",
      "Mart Raudsepp",
      "Martin Schlemmer",
      N_("<big><b>Translations by</b></big>"),
      "Afrikaan: Martin Schlemmer",
      "Brazilian Portuguese: Renato Atilio",
      "Bulgarian: Rostislav Raykov",
      "Czech: Ondrej Novy",
      "Dutch: Niels Abspoel",
      "Estonian: Hannes Tarien",
      "Finnish: Tero Ratilainen",
      "French: David Le Brun, Olivier Blin",
      "German: Milosz Derezynski",
      "Greek: Stavros Giannouris",
      "Hungarian: Balint Erdosi",
      "Italian: Andrea Scialpi",
      "Japanese: Nobuyuki Ito",
      "Lithuanian: Ernestas Liubarskij",
      "Polish: Jacek Wolszczak",
      "Russian: Youri",
      "Simplified Chinese: Huang Huan, Chong Kai Xiong",
      "Spanish: Andres Suarez",
      "Swedish: Peter Asplund",
      "Thai: Prach Pongpanich",
  };

  const unsigned int n_lines = G_N_ELEMENTS (credit_text);

  const double total_animation_time = n_lines * text_time;
  const double start_time = initial_delay;
  const double end_time   = start_time + total_animation_time;

  inline double
  cos_smooth (double x)
  {
      return (1.0 - std::cos (x * G_PI)) / 2.0;
  }

  const char *
  get_text_at_time (double time)
  {
      if (time >= start_time && time <= end_time)
      {
          unsigned int line = static_cast<unsigned int> ((time - start_time) / text_time);
          return credit_text[line];
      }
      else
      {
          return 0;
      }
  }

  double
  get_text_alpha_at_time (double time)
  {
      if (time >= start_time && time <= end_time)
      {
          double offset = std::fmod ((time - start_time), text_time);

          if (offset < text_fade_in_time)
          {
              return text_full_alpha * cos_smooth (offset / text_fade_in_time);
          }
          else if (offset < text_fade_in_time + text_hold_time)
          {
              return text_full_alpha;
          }
          else
          {
              return text_full_alpha * cos_smooth (1.0 - (offset - text_fade_in_time - text_hold_time) / text_fade_out_time);
          }
      }
      else
      {
          return 0;
      }
  }
}

namespace Bmp
{
  AboutDialog::AboutDialog ()
      : has_alpha (false)
  {
      set_title (_("About BMPx"));
      bmp_window_set_icon_list (GTK_WIDGET (gobj ()), "player");

      set_position (Gtk::WIN_POS_CENTER);
      set_resizable (false);

      set_type_hint (Gdk::WINDOW_TYPE_HINT_DIALOG);
      set_decorated (false);

      set_app_paintable (true);
      add_events (Gdk::ALL_EVENTS_MASK);

      Glib::RefPtr<Gdk::Screen> screen = Gdk::Screen::get_default ();
      Glib::RefPtr<Gdk::Colormap> colormap;

      Glib::RefPtr<Gdk::Visual> visual = Util::screen_get_rgba_visual (screen);
      if (visual)
      {
          colormap = Util::screen_get_rgba_colormap (screen);
          has_alpha = true;
      }
      else
      {
          has_alpha = false;
      }

      g_message ("RGBA visuals availability: %s\n", has_alpha ? "yes" : "no");

      const gchar *filename = BMP_IMAGE_DIR G_DIR_SEPARATOR_S "about-logo.png";
      background = Gdk::Pixbuf::create_from_file (filename);

      set_size_request (background->get_width (), background->get_height ());

      if (has_alpha)
      {
          set_colormap (colormap);
      }
      else
      {
          Glib::RefPtr<Gdk::Pixmap> pixmap;
          Glib::RefPtr<Gdk::Bitmap> mask;

          background->render_pixmap_and_mask (pixmap, mask, mask_alpha_threshold);

          shape_combine_mask (mask, 0, 0);
      }

      timer.stop ();
      timer.reset ();
  }

  bool
  AboutDialog::on_key_press_event (GdkEventKey *event)
  {
      if (event->keyval == GDK_Escape)
          hide ();

      return false;
  }

  bool
  AboutDialog::on_button_press_event (GdkEventButton *event)
  {
      if (event->button == 1)
          hide ();

      return false;
  }

  bool
  AboutDialog::on_expose_event (GdkEventExpose *event)
  {
      draw_frame ();

      return false;
  }

  bool
  AboutDialog::on_delete_event (GdkEventAny *event)
  {
      hide ();
      return true;
  }

  void
  AboutDialog::on_map ()
  {
      Gtk::Window::on_map ();

      if (timer.elapsed () == 0.0)
      {
          timer.start ();

          update_connection = Glib::signal_timeout ().connect (sigc::mem_fun (this, &AboutDialog::update_frame),
                                                               animation_frame_period_ms);
      }
  }

  void
  AboutDialog::on_unmap ()
  {
      if (timer.elapsed () > 0.0)
      {
          update_connection.disconnect ();

          timer.stop ();
          timer.reset ();
      }

      Gtk::Window::on_unmap ();
  }

  void
  AboutDialog::draw_frame ()
  {
      Cairo::RefPtr<Cairo::Context> cr (new Cairo::Context (gdk_cairo_create (get_window ()->gobj ()), true));

      if (has_alpha)
      {
          cr->set_source_rgba (.0, .0, .0, .0);
          cr->paint ();
      }

      gdk_cairo_set_source_pixbuf (cr->cobj (), background->gobj (), 0, 0);
      cr->set_operator (Cairo::OPERATOR_SOURCE);
      cr->paint ();

      double current_time = timer.elapsed ();

      if (current_time >= start_time && current_time <= end_time)
      {
          Pango::FontDescription font_desc (text_font);

          int text_size_pt = static_cast<int> ((text_size_px * 72) / Util::screen_get_y_resolution (Gdk::Screen::get_default ()));
          font_desc.set_size (text_size_pt * PANGO_SCALE);

          const char *text  = get_text_at_time (current_time);
          double      alpha = get_text_alpha_at_time (current_time);

          // NOTE: For some stupid reason, pangomm refuses to Glib::wrap() a PangoLayout object created by
          // pango_cairo_create_layout(), so we're forced to use PangoLayout directly.

          PangoLayout *layout = pango_cairo_create_layout (cr->cobj ());
          pango_layout_set_font_description (layout, font_desc.gobj ());

          pango_layout_set_markup (layout, _(text), -1);

          int width, height;
          pango_layout_get_pixel_size (layout, &width, &height);

          cr->move_to ((background->get_width () - width) / 2, background->get_height () - height - text_bottom_margin);
          cr->set_source_rgba (text_colour[0], text_colour[1], text_colour[2], alpha);
          cr->set_operator (Cairo::OPERATOR_ATOP);

          pango_cairo_show_layout (cr->cobj (), layout);

          g_object_unref (layout);
      }
  }

  bool
  AboutDialog::update_frame ()
  {
      queue_draw ();

      return true;
  }

} // namespace Bmp
