#include <iostream>
#include <sstream>

#include <glibmm.h>
#include <sqlite3_bmp.h>

#include <boost/tuple/tuple.hpp>
#include <boost/variant.hpp>
#include <boost/format.hpp>

#include "debug.hpp"
#include "database.hpp"

namespace
{
  Glib::ustring
  sql_uprintf (const char *format, ...)
  {
    va_list args;
    va_start (args, format);
    char *v = sqlite3_vmprintf (format, args);
    va_end (args);
    Glib::ustring r (v);
    sqlite3_free (v);
    return r;
  }

  const char* sql_valuenames[] =
  {
      "BOOL", 
      "INTEGER",
      "REAL",
      "TEXT"
  };

  static boost::format format_int  ("%d");
  static boost::format format_bool ("%d");
  static boost::format format_real ("%f");

  int statement_prepare (sqlite3_stmt **pStmt, const Glib::ustring& sql, sqlite3 *sqlite)
  {
      int status;
      const char *tail;
      status = sqlite3_prepare (sqlite,
                                sql.c_str(), 
                                strlen (sql.c_str()),
                                pStmt,
                                &tail);      
  
      if (status)
    {
      g_warning ("SQL Error: '%s', SQL Statement: '%s'", sqlite3_errmsg (sqlite), sql.c_str ());
      sqlite3_reset (*pStmt);
    };

      return status;
  }

  void get_row_from_statement (sqlite3_stmt *pStmt, Bmp::DB::Row& row, Bmp::DB::ValueMap const& map)
  {
      using namespace Bmp::DB;
      using namespace Glib; 

      unsigned int counter = 0;
      for (ValueMap::const_iterator iter = map.begin (); iter != map.end (); iter++)
      {
        const PairValueType &pair = (*iter);
        switch (pair.second)
          {
            case VALUE_TYPE_BOOL:
              row.insert (std::make_pair (pair.first, bool(sqlite3_column_int (pStmt, counter++)))); break;

            case VALUE_TYPE_INT:
              row.insert (std::make_pair (pair.first, int(sqlite3_column_int (pStmt, counter++)))); break;

            case VALUE_TYPE_REAL:
              row.insert (std::make_pair (pair.first, double(sqlite3_column_double (pStmt, counter++)))); break;

            case VALUE_TYPE_STRING:
              {
                const char* value = (const char*)sqlite3_column_text (pStmt, counter++);        
                if (value)
                  {
                    row.insert (std::make_pair (pair.first, ustring(value)));
                    break;
                  }
                else
                  {
                    row.insert (std::make_pair (pair.first, ustring()));
                  }
              }
            default: break;
          }
      }
  }

  void
  print_compiled_statement (sqlite3_stmt *stmt)
  {
    int count = sqlite3_bind_parameter_count (stmt);
    for (int n = 0; n < count; ++n)
    {
      g_message ("Parameter %d: %s", n, sqlite3_bind_parameter_name (stmt, n));
    }
  }

  void
  append_attributes (Glib::ustring &sql, const Bmp::DB::AttributeList& attributes)
  {
      using namespace Bmp;
      using namespace Bmp::DB;
      using namespace Glib; 

      AttributeList::const_iterator a_last (attributes.end ());
      a_last--;
      for (AttributeList::const_iterator a_iter  = attributes.begin  (),
                     a_end   = attributes.end    ();
                     a_iter != a_end ; ++a_iter)
      {
        Bmp::DB::Attribute const& Attr      = (*a_iter);
        const MatchStyle match_style        = boost::get<0>(Attr);
        std::string const& attribute_name   = boost::get<1>(Attr);
        Variant const& value                = boost::get<2>(Attr);

        if (match_style == FUZZY)
          {
              switch (value.which())
                {
                  case VALUE_TYPE_STRING: 
                    {
                      sql.append (sql_uprintf (" %s LIKE '%%%q%%' ", attribute_name.c_str(), boost::get<ustring>(value).c_str())); 
                      break;
                    }

                  case VALUE_TYPE_INT:
                    {
                      sql .append (attribute_name)  
                          .append (" LIKE '%")
                          .append ((format_int % boost::get<int>(value)).str()).append ("%' ");
                      break;
                    }

                  case VALUE_TYPE_BOOL:
                    {
                      sql .append (attribute_name)
                          .append (" LIKE '%")
                          .append ((format_int % int(boost::get<bool>(value))).str()).append ("%' ");
                      break;
                    }
                }
          }
        else
          {
            ustring op;

            switch (match_style)  
              {
                case FUZZY: continue; //Should have been handled above;
                case EXACT:
                  {
                    op = "="; break;
                  }

                case LESSER_THAN:
                  {
                    op = "<"; break;
                  }

                case GREATER_THAN:
                  {
                    op = ">"; break;
                  }

                case LESSER_THAN_OR_EQUAL:
                  {
                    op = "<="; break;
                  }

                case GREATER_THAN_OR_EQUAL:
                  {
                    op = ">="; break;
                  }
              }
  
          switch (value.which())
            {
              case VALUE_TYPE_STRING:
                {
                  sql.append (sql_uprintf (" %s %s '%q' ", attribute_name.c_str(), op.c_str(), boost::get<Glib::ustring>(value).c_str())); 
                  break;
                }

              case VALUE_TYPE_INT:
                {
                  sql .append (attribute_name)
                      .append (" ")
                      .append (op)
                      .append (" '")
                      .append ((format_int % boost::get<int>(value)).str()).append ("' ");
                  break;
                }

              case VALUE_TYPE_BOOL:
                {
                  sql .append (attribute_name)
                      .append (" ")
                      .append (op)
                      .append (" '")
                      .append ((format_int % int(boost::get<bool>(value))).str()).append ("' ");
                  break;
                }
            }
          } 

      if (a_iter != a_last) sql.append (" AND ");
   }
  }
};

namespace Bmp
{
  namespace DB
  {
      void
      Database::sqlite_trace   (void         *data,
                                const char   *sqltext)
      {
        g_message (G_STRLOC ": %s", sqltext);
      }

      unsigned int
      Database::sqlite_exec_simple (const char* sql) 
      {
        int            status = 0;
        sqlite3_stmt  *pStmt  = 0;
        const char    *tail   = 0;
        unsigned int   rows   = 0;

        debug ("sql", "exec-simple: '%s'", sql);

        status = sqlite3_prepare (sqlite,
                                  sql,    
                                  strlen (sql),
                                  &pStmt,
                                  &tail);      

        if (status != SQLITE_OK)
          {
            sqlite3_reset (pStmt);
            throw Exception (sqlite3_errmsg (sqlite), sqlite3_errcode (sqlite), sql);
          }

        for (status = -1; status != SQLITE_DONE; )
          {
            status = sqlite3_step (pStmt);
            if (status == SQLITE_ROW) { rows++; continue; }
            if (status == SQLITE_BUSY) continue;
            if (status == SQLITE_DONE) break;
            if (status == SQLITE_ERROR || status == SQLITE_MISUSE)
              {
                sqlite3_reset (pStmt);
                throw Exception (sqlite3_errmsg (sqlite), sqlite3_errcode (sqlite), sql);
              }
          }
        sqlite3_finalize (pStmt);
        return rows;
      }

      Database::Database (std::string name, std::string path, bool truncate) : name (name), path (path)
      {
        using namespace std;
        using namespace Glib;

        string db_filename (build_filename (path, name) + ".mlib");        

          if (file_test (db_filename, FILE_TEST_EXISTS) && truncate)
            {
              unlink (db_filename.c_str());
            }

          if (sqlite3_open (db_filename.c_str(), &sqlite))
            {
              throw Exception (sqlite3_errmsg (sqlite), sqlite3_errcode (sqlite), db_filename);
            }
#ifdef SQLITE_TRACE
          sqlite3_trace (sqlite, Database::sqlite_trace, this);
#endif
      };

      Database::~Database ()
      {
          sqlite3_close (sqlite);
      };

      void
      Database::create_attribute_view     (std::string const&   name,
                                           std::string const&   view_name,
                                           std::string const&   attribute,
                                           ValueMap    const&   map)  
      {
        using namespace Glib;

        mapmap.insert (std::make_pair (name, map));

        if (table_exists (view_name)) return;

        ustring sql;

        sql   .append ("CREATE VIEW ")
              .append (name)
              .append (".")
              .append (view_name)
              .append (" AS SELECT DISTINCT ")
              .append (attribute)
              .append (" FROM " )
              .append (name)
              .append (";");
                        
        sqlite_exec_simple (sql.c_str());
      }

      void
      Database::create_view_manual    (std::string const&  name,
                                       std::string const&  view_name,
                                       std::string const&  select,
                                       ValueMap    const&  map)
      {
        using namespace Glib;

        mapmap.insert (std::make_pair (view_name, map));

        if (table_exists (view_name)) return;

        ustring sql;

        sql .append ("CREATE VIEW ")
            .append (name)
            .append (".")
            .append (view_name)
            .append (" AS ")
            .append (select)
            .append (";");

        sqlite_exec_simple (sql.c_str());
      }

      void
      Database::add  (std::string const&       name,
                      Bmp::DB::Row const&      row)
      {
        using namespace Glib;

        if (mapmap.find (name) == mapmap.end ())
        {
          g_error ("No value-map for table '%s'", name.c_str());
        }

        const ValueMap& map = (*(mapmap.find(name))).second;

        ustring sql_keys,
                sql_values;

        bool had_value = false; 

        sql_keys    .append ("(");
        sql_values  .append ("("); 

        for (ValueMap::const_iterator iter = map.begin (); iter != map.end (); iter++)
        {
          const PairValueType &pair = (*iter);
          const Bmp::DB::Row::const_iterator data_iter = row.find (pair.first);

          if (data_iter != row. end ())
            {
              const Variant &value = (*data_iter).second;

              if (had_value)
                {
                  sql_keys    .append (" , ");
                  sql_values  .append (" , ");
                }

                //add key
                sql_keys.append (pair.first);

                debug ("sql", "::add: Key [%s], Type: %d", pair.first.c_str(), int(value.which()));

                //add its value
                sql_values.append ("'");
                switch (pair.second)
                {
                  case VALUE_TYPE_STRING:
                    {
                      sql_values.append (sql_uprintf ("%q", boost::get<ustring>(value).c_str())); 
                      break;
                    }
  
                  case VALUE_TYPE_INT:
                    sql_values.append ((format_int % boost::get<int>(value)).str());
                    break;

                  case VALUE_TYPE_BOOL:
                    sql_values.append ((format_bool % boost::get<bool>(value)).str()); 
                    break;

                  case VALUE_TYPE_REAL:
                    sql_values.append ((format_real % boost::get<double>(value)).str()); 
                    break;

                }
              sql_values.append ("'");
              had_value = true;
            }
        }

        sql_keys      .append (")");
        sql_values    .append (")");

        ustring statement;
        statement   .append ("INSERT INTO ")
                    .append (name)
                    .append (" ")
                    .append (sql_keys)
                    .append (" VALUES ")
                    .append (sql_values)
                    .append (";");
        debug ("sql", "Statement: %s", statement.c_str()); 
        sqlite_exec_simple (statement.c_str());
      }

      void
      Database::del   (std::string const&     name,
                        AttributeList  const& attributes)
      {
        using namespace Glib;

        ustring sql;
        sql.append ("DELETE FROM ").append(name).append(" WHERE ");
        append_attributes (sql, attributes);
        sql.append (";");
        sqlite_exec_simple (sql.c_str());
      }

      void 
      Database::project     (std::string const&       name,
                             std::string const&       project, 
                             VRow&                    vector,
                             const AttributeList&     attributes)
      {
        using namespace std;
        using namespace Glib;

        if (mapmap.find (name) == mapmap.end ())
          {
            g_error ("No value-map for table '%s'", name.c_str());
          }

        const ValueMap& map = (*(mapmap.find(name))).second;
        ValueMap::const_iterator vm_last (map.end());   
        vm_last--;

        ustring sql;
        if (!attributes.empty ())
          {
            sql .append ("SELECT DISTINCT ")
                .append (project)
                .append (" FROM ")
                .append (name)
                .append (" WHERE ");
            append_attributes (sql, attributes);
            sql .append (";");
          }
        else
          {
            sql .append ("SELECT DISTINCT ")
                .append (project)
                .append (" FROM ")
                .append (name)
                .append (";");
          }

        sqlite3_stmt  *pStmt;
        if (statement_prepare (&pStmt, sql, sqlite))
          {
            throw Exception (sqlite3_errmsg (sqlite), sqlite3_errcode (sqlite), sql);
          }

        for (int status = -1 ; status != SQLITE_DONE ; )
          {
            status = sqlite3_step (pStmt);
            switch (status)
              {
                case SQLITE_BUSY: continue;
                case SQLITE_DONE: break;
                case SQLITE_ROW:            
                {
                  Bmp::DB::Row row;
                  Bmp::DB::ValueType type = (*(map.find (project))).second;

                  switch (type)
                  {
                    case VALUE_TYPE_BOOL:
                      row.insert (make_pair (project, bool(sqlite3_column_int (pStmt, 0)))); break;

                    case VALUE_TYPE_INT:
                      row.insert (make_pair (project, int(sqlite3_column_int (pStmt, 0)))); break;

                    case VALUE_TYPE_REAL:
                      row.insert (make_pair (project, double(sqlite3_column_double (pStmt, 0)))); break;

                    case VALUE_TYPE_STRING:
                    {
                      const char* v = (const char*)sqlite3_column_text (pStmt, 0); 
                      if (v)
                        {
                          row.insert (make_pair (project, ustring(v)));
                        }
                      break;
                    }
                    default: break;
                  }

                  vector.push_back (row);
                  break;
                }
                case SQLITE_ERROR:  sqlite3_reset (pStmt); return; 
                case SQLITE_MISUSE: sqlite3_reset (pStmt); return; 
              } 
          } 
        sqlite3_finalize (pStmt);
      }

      void
      Database::txn_begin()
      {
        Glib::ustring sql;
        sql.append ("BEGIN;");
        sqlite_exec_simple (sql.c_str());
      }

      void
      Database::txn_commit()
      {
        Glib::ustring sql;
        sql.append("COMMIT;");
        sqlite_exec_simple (sql.c_str());
      }

      void
      Database::txn_rollback()
    
      {
        Glib::ustring sql;
        sql.append("ROLLBACK;");
        sqlite_exec_simple (sql.c_str());
      }

      void 
      Database::set     (std::string const&       name,
                         std::string const&       pkey,
                         std::string const&       pkey_value,
                         AttributeListSet const&  attributes)
      {
        using namespace Glib;

        ustring sql;
        sql.append ("UPDATE ").append(name).append(" SET ");
        AttributeListSet::const_iterator a_last = attributes.end ();
        a_last--;
        for (AttributeListSet::const_iterator a_iter  = attributes.begin (),
                                              a_end   = attributes.end ();
                                              a_iter != a_end; ++a_iter)
        {
            Bmp::DB::AttributeSet const& Attr   = (*a_iter);
            std::string const& attribute_name   = boost::get<0>(Attr);
            Variant const& value                = boost::get<1>(Attr);

            switch (value.which())
            {
                case VALUE_TYPE_STRING:
                  {
                    sql.append (sql_uprintf ("%s='%q' ", attribute_name.c_str(), boost::get<ustring>(value).c_str())); 
                    break;
                  }

                case VALUE_TYPE_INT:
                  {
                    sql .append (attribute_name)
                        .append ("='")
                        .append ((format_int % boost::get<int>(value)).str()).append("' ");
                    break;
                  }

                case VALUE_TYPE_BOOL:
                  {
                    sql .append (attribute_name)
                        .append ("='")
                        .append((format_int % int(boost::get<bool>(value))).str()).append("' ");
                    break;
                  }
            }

            if (a_iter != a_last) sql.append (" , ");
        }

        sql.append (sql_uprintf (" WHERE %s='%q';", pkey.c_str(), pkey_value.c_str())); 
        sqlite_exec_simple (sql.c_str());
      }

      void 
      Database::get_by_rowid (std::string const&    name,
                              sqlite_int64          rowid,
                              Row&                  row)
      {
        using namespace Glib;
        static boost::format lld ("%lld");

        if (mapmap.find (name) == mapmap.end ())
          {
            g_error ("No value-map for table '%s'", name.c_str());
          }

        const ValueMap& map = (*(mapmap.find(name))).second;
        ValueMap::const_iterator vm_last (map.end());   
        vm_last--;

        ustring sql;
        sql .append ("SELECT ");

        for (ValueMap::const_iterator vm_iter = map.begin (),
                                      vm_end  = map.end   ();
                                      vm_iter != vm_end ; ++vm_iter)
          {
            const PairValueType &pair = (*vm_iter);
            sql.append (pair.first);
            if (vm_iter != vm_last) sql.append (",");
          }

        sql .append (" FROM ")
            .append (name)
            .append (" WHERE rowid='")
            .append ((lld % rowid).str())
            .append ("';"); 

        sqlite3_stmt *pStmt;
        if (statement_prepare (&pStmt, sql, sqlite))
          { 
            throw Exception (sqlite3_errmsg (sqlite), sqlite3_errcode (sqlite), sql);
          }

        for (int status = -1 ; status != SQLITE_DONE ; )
          {
            status = sqlite3_step (pStmt);
            switch (status)
              {
                case SQLITE_BUSY:   continue;
                case SQLITE_DONE:   break;
                case SQLITE_ERROR:  sqlite3_reset (pStmt); return;
                case SQLITE_MISUSE: sqlite3_reset (pStmt); return;
                case SQLITE_ROW:            
                  {
                    get_row_from_statement (pStmt, row, map);
                    sqlite3_finalize (pStmt);
                    return;
                  }
              } 
          } 
        sqlite3_finalize (pStmt);
      }

      void 
      Database::get         (std::string const&     name,
                             AttributeList const&   attributes,
                             VRow&                  vector,
                             std::string const&     suffix)
      {
        using namespace Glib;

        if (mapmap.find (name) == mapmap.end ())
        {
          g_error ("No value-map for table '%s'", name.c_str());
        }

        ValueMap const& map = (*(mapmap.find(name))).second;
        ValueMap::const_iterator vm_last (map.end());   
        vm_last--;

        ustring sql;
        sql .append ("SELECT ");

        for (ValueMap::const_iterator vm_iter = map.begin (),
                                      vm_end  = map.end   ();
                                      vm_iter != vm_end ; ++vm_iter)
          {
            PairValueType const& pair = (*vm_iter);
            sql.append (pair.first);
            if (vm_iter != vm_last) sql.append (",");
          }

        if (!attributes.empty ())
          {
            sql.append (" FROM ").append(name).append(" WHERE ");
            append_attributes (sql, attributes);
            if (suffix.empty())
              {
                sql.append (";");
              }  
            else
              {
                sql.append (" ").append(suffix).append(" ").append(";");
              }
          }
        else
          {
            sql.append (" FROM ").append(name);
            if (suffix.empty())
              sql.append (";");
            else
              sql.append (" ").append(suffix).append(";");
          }

        sqlite3_stmt *pStmt = 0;
        if (statement_prepare (&pStmt, sql, sqlite))
          {
            throw Exception (sqlite3_errmsg (sqlite), sqlite3_errcode (sqlite), sql);
          }
 
        for (int status = -1 ; status != SQLITE_DONE ; )
          {
            status = sqlite3_step (pStmt);
            switch (status)
              {
                case SQLITE_BUSY:   continue;
                case SQLITE_DONE:   break;
                case SQLITE_MISUSE: sqlite3_reset (pStmt); return;
                case SQLITE_ERROR:  sqlite3_reset (pStmt); return;
                case SQLITE_ROW:            
                  {
                    Bmp::DB::Row row;
                    get_row_from_statement (pStmt, row, map);
                    vector.push_back (row);
                    break;
                  }
              } 
          } 
          sqlite3_finalize (pStmt);
      }

      void
      Database::insert_map_for_table
                        (std::string const      &name,
                         const ValueMap         &map)
      {
        mapmap.insert (std::make_pair (name, map));
      }

      void
      Database::create_table (std::string const   &name,
                              std::string const   &pkey,
                              const ValueMap      &map)
      {
        using namespace Glib;

        mapmap.insert (std::make_pair (name, map));

        if (table_exists (name)) return;

        ustring sql;
        sql .append ("CREATE TABLE IF NOT EXISTS ")
            .append (name)
            .append (" (");

        ValueMap::const_iterator last;
        last = map.end ();
        last--;

        for (ValueMap::const_iterator iter = map.begin (); iter != map.end (); iter++)
          {
            PairValueType const& pair = (*iter);
            if (!pkey.compare (pair.first))
              {
                sql   .append ("'")
                      .append (pair.first)
                      .append ("' ")
                      .append (sql_valuenames[pair.second])
                      .append (" NOT NULL PRIMARY KEY");
              }
            else
              {
                if (pair.second != VALUE_TYPE_INT)
                  {
                    sql .append ("'")
                        .append (pair.first)
                        .append ("' ")
                        .append (sql_valuenames[pair.second])
                        .append (" ");
                  }
                else
                  {
                    sql .append ("'")
                        .append (pair.first)
                        .append ("' ")
                        .append (sql_valuenames[pair.second])
                        .append (" DEFAULT 0");
                  }
              }
            if (iter != last) sql.append (",");
          }

        sql.append (");");
        sqlite_exec_simple (sql.c_str());
      }

      void
      Database::drop_table (std::string const& name)
      {
        using namespace Glib;

        ustring sql;

        sql .append ("DROP TABLE ")
            .append (name)
            .append (";");

        sqlite_exec_simple (sql.c_str());
      }

      bool
      Database::table_exists    (std::string const& name)
      {
        using namespace Glib;

        ustring sql; 

        sql .append ("SELECT name FROM sqlite_master WHERE name='")
            .append (name)
            .append("';");

        unsigned int nrow = sqlite_exec_simple (sql.c_str()); 
        return (nrow != 0); 
      }

      bool
      Database::attr_exists (std::string const&   name, 
                             std::string const&   pkey_name,
                             const Glib::ustring& pkey_value)
      {
        using namespace Glib;

        ustring sql; 
        sql .append ("SELECT ")
            .append (pkey_name)
            .append (" FROM ")
            .append (name)
            .append (" WHERE ")
            .append (pkey_name)
            .append ("='")
            .append (sql_uprintf ("%q", pkey_value.c_str()))
            .append ("';");

        unsigned int nrow = sqlite_exec_simple (sql.c_str()); 
        return (nrow != 0); 
      }

      void
      Database::remove_by_attr  (std::string const&   name, 
                                 std::string const&   pkey_name,
                                 Glib::ustring const& pkey_value)
      {
        using namespace Glib;

        ustring sql; 
        sql .append ("DELETE FROM ")
            .append (name)
            .append (" WHERE ")
            .append (pkey_name)
            .append ("='")
            .append (sql_uprintf ("%q", pkey_value.c_str()))
            .append ("';");
        sqlite_exec_simple (sql.c_str()); 
      }
  } 
}; 
