//  BMPx - The Dumb Music Player
//  Copyright (C) 2005-2006 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non GPL-compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#include <config.h>

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <istream>
#include <ostream>
#include <fstream>
#include <boost/format.hpp>

#include <unistd.h>
#include <errno.h>

#include "debug.hpp"
#include "util.hpp"
#include "uri++.hpp"
#include "file_utils.hpp"
#include "audio.hpp"

#include "x_play.hpp"

#ifndef __USE_BSD
# define __USE_BSD      /* Needed for mkdtemp () */
#endif

#include "main.hpp"

//////////////// Archive handling functions
namespace Bmp
{
  namespace Util
  {
    namespace
    {
      typedef std::list<std::string> DirEntries;

      enum ArchiveType
      {
          ARCHIVE_UNKNOWN = 0,
          ARCHIVE_DIR,
          ARCHIVE_TAR,
          ARCHIVE_TGZ,
          ARCHIVE_ZIP,
          ARCHIVE_TBZ2
      };

      struct ArchiveExtensionType
      {
          ArchiveType  type;
          const char  *ext;
      };

      ArchiveExtensionType archive_extensions[] =
      {
          {ARCHIVE_TAR, "tar"},
          {ARCHIVE_ZIP, "wsz"},
          {ARCHIVE_ZIP, "zip"},
          {ARCHIVE_TGZ, "tar.gz"},
          {ARCHIVE_TGZ, "tgz"},
          {ARCHIVE_TBZ2, "tar.bz2"},
          {ARCHIVE_TBZ2, "bz2"},
          {ARCHIVE_UNKNOWN, NULL}
      };

      boost::format archive_extract_commands[] =
      {
          boost::format (""),
          boost::format (""),
          boost::format ("tar >/dev/null xf %1% -C %2%"),
          boost::format ("tar >/dev/null xzf %1% -C %2%"),
          boost::format ("unzip >/dev/null -o -j %1% -d %2%"),
          boost::format ("bzip2 -dc %1% | tar >/dev/null xf - -C %2%")
      };

      std::string
      escape_shell_chars (const std::string &string)
      {
          char *out_c = g_shell_quote (string.c_str ());
          std::string out (out_c);
          free (out_c);
          return out;
      }

      ArchiveType
      archive_get_type (const std::string &filename)
      {
          int i = 0;

          if (Glib::file_test (filename, Glib::FILE_TEST_IS_DIR))
              return ARCHIVE_DIR;

          while (archive_extensions[i].ext)
          {
              if (Util::str_has_suffix_nocase (filename.c_str (), archive_extensions[i].ext))
                  return archive_extensions[i].type;

              i++;
          }

          return ARCHIVE_UNKNOWN;
      }

      bool
      find_file_actual (const std::string &dir_path,
                        const std::string &filename,
                        std::string       &found_path,
                        int                max_depth,
                        int                depth)
      {
          if (max_depth != -1 && depth > max_depth)
          {
              return false;
          }

          Glib::Dir dir (dir_path);
          DirEntries entries (dir.begin (), dir.end ());

          for (DirEntries::const_iterator iter = entries.begin (), end = entries.end ();
               iter != end;
               ++iter)
          {
              std::string path = Glib::build_filename (dir_path, *iter);
              if (path.length () > FILENAME_MAX)
              {
                  g_warning (G_STRLOC ": ignoring path: name too long (%s)",
                             (path).c_str ());
                  continue;
              }

              if (Glib::file_test (path, Glib::FILE_TEST_IS_REGULAR))
              {
                  try
                    {
                      // NOTE: We do a straight comparison first so that should G_FILENAME_ENCODING
                      // be wrongly set, we at least can match filenames character for character.
                      // - descender

                      if (*iter == filename)
                        {
                          found_path = path;
                          return true;
                        }

                     Glib::ustring filename_utf8 = Glib::filename_to_utf8 (filename).casefold ();
                      if (Glib::filename_to_utf8 (*iter).casefold () == filename_utf8)
                      {
                          found_path = path;
                          return true;
                      }
                    }
                  catch (Glib::ConvertError &error)
                    {
                      debug ("file-utils", "%s Cannot convert filename %s to UTF-8. G_FILENAME_ENCODING may not be properly set", G_STRLOC, (*iter).c_str());
                    }
              }
              else if (Glib::file_test (path, Glib::FILE_TEST_IS_DIR))
                {
                  if (find_file_actual (path, filename, found_path, max_depth, depth + 1))
                      return true;
                }
          }

          return false;
      }

    } // <anonymous> namespace

    // FIXME: Return a std::string
    char *
    create_glob (const std::string &suffix)
    {
        GString *out;
        int n = 0;

        out = g_string_new ("");

        while (suffix[n])
        {
            g_string_append_printf (out, "[%c%c]", g_ascii_tolower (suffix[n]), g_ascii_toupper (suffix[n]));
            n++;
        }

        return g_string_free (out, FALSE);
    }

    bool
    file_is_archive (const std::string &filename)
    {
        return (archive_get_type (filename) > ARCHIVE_DIR);
    }

    std::string
    archive_basename (const std::string &filename)
    {
        int i = 0;

        while (archive_extensions[i].ext)
        {
            if (Util::str_has_suffix_nocase (filename.c_str (), archive_extensions[i].ext))
                return std::string (filename, 0, filename.length () - std::strlen (archive_extensions[i].ext) - 1); 
            i++;
        }

        return std::string ();
    }

    std::string
    archive_decompress (const std::string &filename)
    {
        int type = archive_get_type (filename);
        if (type <= ARCHIVE_DIR)
            return std::string ();

        char *c_tmp_dir = g_strdup (Glib::build_filename (g_get_tmp_dir (), "bmp.XXXXXXXX").c_str ());

        if (!mkdtemp (c_tmp_dir))
        {
            g_message ("Unable to decompress archive '%s': could not create temporary directory: %s",
                       (filename).c_str (), g_strerror (errno));

            return std::string ();
        }

        std::string tmp_dir (c_tmp_dir);
        g_free (c_tmp_dir);

        std::string escaped_filename (escape_shell_chars (filename));
        std::string command = (archive_extract_commands[type] % escaped_filename % tmp_dir).str ();

        if (command.empty ())
            return std::string ();

        if (system (command.c_str ()))
        {
            g_message ("Unable to decompress archive '%s'", (filename).c_str ());
            return std::string ();
        }

        return tmp_dir;
    }

    bool
    find_file (const std::string &dir_path,
               const std::string &filename,
               std::string       &found_path,
               int                max_depth)
    {
        found_path = "";
        return find_file_actual (dir_path, filename, found_path, max_depth, 0);
    }

    // Deletes a directory recursively
    void
    del_directory (const std::string &dir_path)
    {
        Glib::Dir dir (dir_path);
        DirEntries entries (dir.begin(), dir.end());

        for (DirEntries::const_iterator iter = entries.begin (), end = entries.end ();
             iter != end;
             ++iter)
        {
            std::string full_path = Glib::build_filename (dir_path, *iter);

            if (Glib::file_test (full_path, Glib::FILE_TEST_IS_DIR))
                del_directory (full_path);
            else
                unlink (full_path.c_str ());
        }

        rmdir (dir_path.c_str ());
    }

    // "Collects" a path of audio files recursively
    void
    collect_path (const std::string &dir_path,
                  FileList          &collection,
                  bool               clear)
    {
        if (clear)
        {
            collection.clear ();
        }

        Glib::Dir dir (dir_path);
        DirEntries entries (dir.begin(), dir.end ());

        for (DirEntries::const_iterator iter = entries.begin (), end = entries.end ();
             iter != end;
             ++iter)
        {
            std::string full_path = Glib::build_filename (dir_path, *iter);

            if (Glib::file_test (full_path, Glib::FILE_TEST_IS_REGULAR))
              {
                if (play->is_audio_file (*iter))
                {
                    collection.push_back (full_path);
                }
              }
            else if (Glib::file_test (full_path, Glib::FILE_TEST_IS_DIR))
              {
                collect_path (full_path, collection, false);
              }
        }
    }

    void
    dir_for_each_entry (const std::string &path,
                        DirForeachSlot     slot)
    {
        Glib::Dir dir (path);
        DirEntries entries (dir.begin(), dir.end ());

        for (DirEntries::const_iterator iter = entries.begin (), end = entries.end ();
             iter != end;
             ++iter)
        {
            std::string full_path = Glib::build_filename (path, *iter);

            if (slot (full_path))
                break;
        }
    }

    void
    files_writable (const FileList &list,
                    FileList       &non_writable)
    {

        for (FileList::const_iterator file  = list.begin (), end = list.end ();
                                      file != end; ++file)
        {
            std::string filename = (*file); 

            struct stat info;
            if (stat (filename.c_str (), &info))
            {
                non_writable.push_back (*file);
                continue;
            }

            // Get stat info
            uid_t  stat_uid  = info.st_uid;
            gid_t  stat_gid  = info.st_gid;
            mode_t stat_mode = info.st_mode;

            // Check if world-writable
            if (stat_mode & S_IWOTH)
                continue;

            // Get the process uid/gid
            uid_t p_uid = geteuid ();
            gid_t p_gid = getegid ();

            // Check whether group has write permission
            if ((p_gid == stat_gid) && (stat_mode & S_IWGRP))
                continue;

            //Check whether user has write permission
            if ((p_uid == stat_uid) && (stat_mode & S_IWUSR))
                continue;

            non_writable.push_back (*file);
        }
    }

    bool
    files_taggable (const FileList &list)
    {
         char* const types[] = {
            "application/x-id3",
            "application/ogg",
#ifdef HAVE_MP4V2
            "audio/x-m4a",
#endif
            "video/x-ms-asf",
            "audio/x-flac",
            "application/x-apetag",
            0,
        };

        for (FileList::const_iterator file  = list.begin (), end = list.end ();
                                      file != end; ++file)
        {
            std::string type;
            if (!Audio::typefind (*file, type)) return false;

            bool found = false; 
            for (unsigned int n = 0; types[n]; ++n)
              {
                if (type == types[n])
                  {
                    found = true;
                    break;
                  }
              }

            if (!found) return false; 
        }
        return true;
    }

    // compares files wrg to existence and filesize
    bool
    compare_files (const std::string &file1,
                   const std::string &file2)
    {
        if (!Glib::file_test (file1, Glib::FILE_TEST_EXISTS))
            return false;

        if (!Glib::file_test (file2, Glib::FILE_TEST_EXISTS))
            return false;

        struct stat stat_a;
        struct stat stat_b;

        stat (file1.c_str (), &stat_a);
        stat (file2.c_str (), &stat_b);

        return (stat_a.st_size == stat_b.st_size);
    }

    void
    copy_file (const std::string &s,
               const std::string &d)
    {
        using namespace std;

        if (s == d) return;

        std::ifstream i (s.c_str (), std::ios::binary);
        std::ofstream o (d.c_str (), std::ios::binary);
        o << i.rdbuf();
        o.close ();
        i.close ();
    }

  } // namespace Util

} // namespace Bmp

/////////////////////////////////////////////////////////////////////
