#include <gtkmm.h>
#include <glibmm/i18n.h>

#include "x_play.hpp"
#include "x_system_control.hpp"
#include "x_vfs.hpp"

#include "main.hpp"
#include "filebrowser.hpp"

namespace Bmp
{
      // Virtual overrides from Gtk::Widget
      bool
      FileBrowser::on_key_press_event (GdkEventKey *event)
      {
        if (event->keyval == GDK_Escape)
        {
          this->hide (); 
          return true;
        }
        return false;
      }

      bool
      FileBrowser::on_delete_event (GdkEventAny *event)
      {
        hide ();
        return true;
      } 

      // Ctor/Creation
      FileBrowser*
      FileBrowser::create ()
      {
        const std::string path = Glib::build_filename (DATA_DIR, "glade" G_DIR_SEPARATOR_S "filebrowser.glade");
        Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);
        FileBrowser *browser = 0;
        glade_xml->get_widget_derived ("filebrowser", browser);
        return browser;
      }

      FileBrowser::FileBrowser (BaseObjectType *cobject,
                const Glib::RefPtr<Gnome::Glade::Xml> &xml)
      : Gtk::Window (cobject),
        ref_xml (xml),
        mode (ADD),
        block_toggle (false)
      {
        ref_xml->get_widget ("chooser", chooser);
        ref_xml->get_widget ("action", action);
        ref_xml->get_widget ("close", close);
        ref_xml->get_widget ("select_all", select_all);
        ref_xml->get_widget ("deselect_all", deselect_all);
        ref_xml->get_widget ("toggle", toggle);

        // Connect buttons
        action->signal_clicked().connect (sigc::mem_fun (*this, &Bmp::FileBrowser::execute));
        close->signal_clicked().connect (sigc::mem_fun (*this, &Bmp::FileBrowser::hide));
        toggle->signal_toggled().connect (sigc::mem_fun (*this, &Bmp::FileBrowser::toggle_toggled));

        select_all->signal_clicked().connect (sigc::mem_fun (*chooser, &Gtk::FileChooser::select_all));
        deselect_all->signal_clicked().connect (sigc::mem_fun (*chooser, &Gtk::FileChooser::unselect_all));
        
        chooser->signal_file_activated().connect (sigc::mem_fun (*this, &Bmp::FileBrowser::execute));

        struct {
          const char *name;
          const char *glob;
        } filter_list[] = {
          { "FLAC", "*.[fF][lL][aA][cC]" },
          { "M4A" , "*.[mM]4[aA]" },
          { "MP3" , "*.[mM][pP]3" },
          { "MPC" , "*.[mM][pP][cC]" },
          { "OGG" , "*.[oO][gG][gG]" },
          { "WMA" , "*.[wW][mM][aA]" }
        };

        // Add "all" filter
        Gtk::FileFilter all;
        all.set_name (_("All Files"));
        all.add_pattern ("*");
        chooser->add_filter (all);

        // Add audiofiles filter
        Gtk::FileFilter audios;
        audios.set_name (_("Audio Files"));
        audios.add_custom (Gtk::FILE_FILTER_URI, sigc::mem_fun (*this, &Bmp::FileBrowser::audio_files_filter));
        chooser->add_filter (audios);

        // Add container plugin filters
        Bmp::VFS::ExportDataList containers;
        vfs->get_containers (containers);

        for (Bmp::VFS::ExportDataList::const_iterator container = containers.begin ();
                              container != containers.end (); ++container)
        {
              std::string glob = Util::create_glob (container->extension);
              std::string glob_suffix = std::string ("*.") + glob;

          Gtk::FileFilter filter_container;
          filter_container.set_name (container->description);
          filter_container.add_pattern (glob_suffix);
          chooser->add_filter (filter_container);
        }

        // Add glob filefilters
        for (unsigned int n = 0; n < G_N_ELEMENTS(filter_list); ++n) 
        {
          Gtk::FileFilter filter;
          filter.set_name (filter_list[n].name);
          filter.add_pattern (filter_list[n].glob);
          chooser->add_filter (filter);
        }

        // Set current folder to the filechooser
        chooser->set_current_folder (mcs->key_get<std::string>("bmp", "file-chooser-path")); 
      }

      void
      FileBrowser::hide ()
      {
        mcs->key_set<std::string>("bmp","file-chooser-path", chooser->get_current_folder());
        Gtk::Widget::hide ();
      }

      bool
      FileBrowser::audio_files_filter (const Gtk::FileFilter::Info& info)
      {
        return (play->is_audio_file(info.uri));
      } 

      void
      FileBrowser::execute ()
      {
        typedef Glib::SListHandle<Glib::ustring> URIList;

        URIList uris = chooser->get_uris (); 
        mcs->key_set<std::string>("bmp","file-chooser-path", chooser->get_current_folder());
        if (!uris.empty())
        {
              char **uri_list = (char**)calloc (uris.size()+1, sizeof(char*));
              unsigned int counter = 0;
              for (URIList::const_iterator iter = uris.begin (),
                           end  = uris.end (); iter != end; ++iter)
              {
                  uri_list[counter++] = strdup ((*iter).c_str ());
              }
              uri_list[uris.size()] = 0;
     
          int out = 0;
          bmp_system_control_add_uri_list (bmp_system_control,
                           (const char**)uri_list,
                           -1,
                           false, 
                           (mode == PLAY),
                           -1,
                           &out,
                           NULL);
          chooser->unselect_all ();
          g_strfreev (uri_list);
          if (toggle->get_active ()) hide ();
        }
      }

      void
      FileBrowser::toggle_toggled ()
      {
        if (block_toggle) return;

        switch (mode)
        {
            case PLAY:
            {
              mcs->key_set<bool>("bmp", "file-chooser-close-on-open", toggle->get_active ()); 
              break;
            }

            case ADD:
            {
              mcs->key_set<bool>("bmp", "file-chooser-close-on-add", toggle->get_active ()); 
              break;
            }
        }
      }

      void
      FileBrowser::run (Mode mode)
      {
          show ();
          Gtk::Window::set_keep_above (mcs->key_get<bool>("bmp", "keep-above"));

          this->mode = mode;

          switch (mode)
          {
                case PLAY:
                {
                  set_title (_("Play Files - BMP"));
                  action->set_label (GTK_STOCK_MEDIA_PLAY);
                  toggle->set_label (_("Close dialog on Play"));
                    
                  block_toggle = true;
                  toggle->set_active (mcs->key_get<bool>("bmp", "file-chooser-close-on-open"));
                  block_toggle = false;
                  break;
                }

                case ADD:
                {
                  set_title (_("Add Files - BMP"));
                  action->set_label (GTK_STOCK_ADD);
                  toggle->set_label (_("Close dialog on Add"));

                  block_toggle = true;
                  toggle->set_active (mcs->key_get<bool>("bmp", "file-chooser-close-on-add"));
                  block_toggle = false;
                  break;
                }
          }
      }
}
