#include <glibmm.h>
#include <glibmm/i18n.h>
#include <gtkmm.h>

#include <config.h>
#include <stdio.h>
#include <dirent.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include <libhal.h>
#include <libhal-storage.h>

#include "hal.hpp"
#include "debug.hpp"
#include "uri++.hpp"

#include "x_system_control.hpp"

namespace
{
  const char* HALBus[] = {
    "<Unknown>",
    "IDE",
    "SCSI",
    "USB",
    "IEEE1394",
    "CCW"
  };

  void
  print_hal_volume (const Bmp::Library::HALVolume& volume)
  {
    g_message ("Volume UDI [%s]", volume.volume_udi.c_str());
    g_message ("\tDevice UDI....: [%s]", volume.device_udi.c_str());
    g_message ("\tMount Path....: [%s]", volume.mount_path.c_str());
    g_message ("\tDevice Serial.: [%s]", volume.device_serial.c_str());
    g_message ("\tVolume Name...: [%s]", volume.volume_name.c_str());
    g_message ("\tDevice File...: [%s]", volume.device_file.c_str());
    g_message ("\tDisc..........: [%s]", volume.is_disc ? "TRUE" : "FALSE");
    g_message ("\tBus...........: [%s]", HALBus[volume.device_bus]); 
  }
}

namespace Bmp
{
  namespace Library
  {
      HAL::HAL () : dbus_connection (0), ctx (0)
      {
          msg_domain_id = bmp_system_control_message_domain_register (bmp_system_control, "hal", _("BMP: Volume/Device Management"));

          if (!hal_init())
          {
            bmp_system_control_message_dispatch (bmp_system_control,
                   msg_domain_id,
                   GTK_MESSAGE_ERROR,
                   _("Volume/Device Management could not be initialized; HAL monitoring of"
                     "volumes in the library, permanent addition of items to the library"
                     "and other parts of BMP will not be functional.\n\nPlease check "
                     "your HAL installation and/or contact your distributor."));
            initialized = false;
          }
          else
          {
            initialized = true;
          }
      }

      HALVolume
      HAL::get_volume_for_uri (Glib::ustring const& uri)
      {
        if (!initialized)
          {
            throw HalException ("No volume for given URI", HalException::HAL_NO_VOLUME_FOR_URI, uri);
          }

        unsigned int length = 0;
        std::string  target_mount_path;
        std::string  filename = Glib::filename_from_uri (uri);
        for (HALVolumes::const_iterator vIter = volumes.begin (), vEnd = volumes.end (); vIter != vEnd; ++vIter)
        {
            HALVolume const& volume = vIter->second;

            unsigned int l_vmp  = volume.mount_path.length (); 
            std::string  substr = filename.substr (0, l_vmp);
      
            debug ("hal", "vmp: '%s', substr: '%s'", volume.mount_path.c_str(), substr.c_str());

            if ((volume.mount_path == substr) && (l_vmp > length))
            {
              target_mount_path = volume.mount_path;
              length = l_vmp;
            } 
	      }

        if (target_mount_path.empty())
          {
            throw HalException ("No volume for given URI", HalException::HAL_NO_VOLUME_FOR_URI, uri);
          }

        HALVolumes::iterator vIter = volumes.find (target_mount_path); 
        if (vIter == volumes.end())
          {
            throw HalException ("No volume for given URI", HalException::HAL_NO_VOLUME_FOR_URI, uri);
          }
        return vIter->second;
      }

      std::string
      HAL::get_mount_path_for_volume_udi (std::string const& volume_udi)
      {
        if (!initialized)
          {
            throw HalException ("No volume for given URI", HalException::HAL_NO_VOLUME_FOR_URI, volume_udi);
          }
 
        for (HALVolumes::const_iterator vIter = volumes.begin (), vEnd = volumes.end (); vIter != vEnd; ++vIter)
            {
              HALVolume const& volume = vIter->second;
              if (volume.volume_udi == volume_udi) return volume.mount_path;
            }
        throw HalException ("No mount path for given URI", HalException::HAL_NO_MOUNT_PATH_FOR_UDI, volume_udi);
      }

      void
      HAL::scan_devices (const char *udi)
      {
        debug ("hal", "Processing UDI '%s'", udi);  

        char **path_elements = g_strsplit (udi+1, "/", -1);
        if (!g_ascii_strncasecmp (path_elements[4], "volume", strlen ("volume")))
          {
            LibHalVolume    *hal_volume;
            LibHalDrive	    *hal_drive;
            HALVolume	     volume;

            volume.volume_udi = udi; 
            hal_volume = libhal_volume_from_udi (ctx, udi); 

            if (hal_volume)
              {
                if (!libhal_volume_is_mounted (hal_volume))
                  {
                    libhal_volume_free (hal_volume);
                    g_strfreev (path_elements);
                    return;
                  }

                volume.mount_path     = libhal_volume_get_mount_point (hal_volume);
                volume.is_disc	      = libhal_volume_is_disc (hal_volume);
                volume.volume_name    = (libhal_volume_get_label (hal_volume)) ? libhal_volume_get_label (hal_volume) : "";
                volume.device_file    = (libhal_volume_get_device_file (hal_volume)) ? libhal_volume_get_device_file (hal_volume) : "";
                volume.device_udi     = libhal_volume_get_storage_device_udi (hal_volume);
                hal_drive = libhal_drive_from_udi (ctx, volume.device_udi.c_str());
                volume.device_serial  = (libhal_drive_get_serial (hal_drive)) ? libhal_drive_get_serial (hal_drive) : "";
                volume.device_bus     = libhal_drive_get_bus (hal_drive);

                libhal_volume_free (hal_volume);
                libhal_drive_free  (hal_drive);
              }
            else
              {
                g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, G_STRLOC ": Unable to retrieve information from HAL for volume UDI '%s'", udi);
              }
        
            volumes.insert (std::make_pair (volume.mount_path, volume)); 
            volumes_by_udi.insert (std::make_pair (volume.volume_udi, volume)); 

            print_hal_volume (volume);

            signal_volume_added_.emit (volume); 
          }
          if ((!g_ascii_strncasecmp (path_elements[4], "storage", strlen ("storage"))) || g_strrstr (path_elements[4], "ide"))
          { 
            LibHalDrive	*hal_drive = libhal_drive_from_udi (ctx, udi); 
            if (hal_drive)
              { 
                // Check for optical device
                if (libhal_drive_get_type (hal_drive) == LIBHAL_DRIVE_TYPE_CDROM)
                {
                    HALDevice device;

                    device.device_udi	= udi;
                    device.device_name	= libhal_drive_get_model (hal_drive);
                    device.device_file  = libhal_drive_get_device_file (hal_drive);

                    optical.push_back (device);
                }
              }
            libhal_drive_free  (hal_drive);
          }
        g_strfreev (path_elements);
      }

      bool
      HAL::hal_init ()
      {
        DBusError     error;
        char	    **devices;
        int	      nr;
        int	      n;

        if (!(ctx = libhal_ctx_new ()))
          {
            debug ("hal", G_STRLOC ": Failed to initialize HAL!");
            ctx = 0;
            return false;
          }

        dbus_error_init (&error);
        if (!hal_mainloop_integration (ctx, &error))
          {
            debug ("hal", G_STRLOC ": Mainloop integration failed!");
            dbus_error_free (&error);
            return false;
          }

        libhal_ctx_set_user_data (ctx, this);
	
#if 0
        libhal_ctx_set_device_new_capability (ctx, hal_device_new_capability);
        libhal_ctx_set_device_lost_capability (ctx, hal_device_lost_capability);
        libhal_ctx_set_device_condition (ctx, hal_device_condition);
#endif

        libhal_ctx_set_device_added		(ctx, Bmp::Library::HAL::device_added);
        libhal_ctx_set_device_removed		(ctx, Bmp::Library::HAL::device_removed);
        libhal_ctx_set_device_property_modified	(ctx, Bmp::Library::HAL::property_modified);

        if (!libhal_device_property_watch_all (ctx, &error))
          {
            debug ("hal", G_STRLOC ": Failed to set up HAL properties watch: %s", error.message);
            dbus_error_free (&error);
            libhal_ctx_free (ctx);
            ctx = 0;
            return false;
          }

        if (!libhal_ctx_init (ctx, &error))
          {
            debug ("hal", G_STRLOC ": hal_ctx_init failed: %s", error.message);
            dbus_error_free (&error);
            libhal_ctx_free (ctx);
            ctx = 0;
            return false;
          }

        if (!(devices = libhal_get_all_devices (ctx, &nr, &error)))
          {
            debug ("hal", G_STRLOC ": get_all_devices failed: %s", error.message);
            dbus_error_free (&error);
            libhal_ctx_shutdown (ctx, NULL);
            libhal_ctx_free (ctx);
            ctx = 0;
            return false;
          }

          for (n = 0; n < nr; n++)
            {
              scan_devices (devices[n]);
            }

        libhal_free_string_array (devices);
        	return true;
      }

      dbus_bool_t
      HAL::hal_mainloop_integration (LibHalContext *ctx,
			                               DBusError     *error)
      {
        dbus_connection = dbus_bus_get (DBUS_BUS_SYSTEM, error);

        if (dbus_error_is_set (error))
          {
            debug ("hal", G_STRLOC ": Error conntecting to DBus");
            dbus_connection = 0;
            return false;
          }

        dbus_connection_setup_with_g_main (dbus_connection, NULL);
        libhal_ctx_set_dbus_connection (ctx, dbus_connection);
        return true;
      }

      HAL::~HAL ()
      {
        DBusError error;
        dbus_error_init (&error);

        if (ctx)
          {
            libhal_ctx_shutdown (ctx, &error);
            libhal_ctx_free (ctx);
          }
        if (dbus_connection)
          {
            dbus_connection_close (dbus_connection);
          }
      }

      void
      HAL::device_added	  (LibHalContext *ctx,
			                     const char    *udi)
      {
      }

      void
      HAL::device_removed (LibHalContext *ctx,
                           const char	   *udi) 
      {
      }

      void
      HAL::property_modified	
		      (LibHalContext    *ctx,
                       const char       *udi,
                       const char       *key,
                       dbus_bool_t       is_removed,
                       dbus_bool_t       is_added)
      {
        using namespace Bmp::Library;

        HAL *hal = reinterpret_cast<Bmp::Library::HAL *>(libhal_ctx_get_user_data (ctx)); 
        bool mounted;

        if (strcmp (key, "volume.is_mounted") != 0) return;
        mounted = libhal_device_get_property_bool (ctx, udi, key, NULL);

        if (mounted)
          {
            hal->scan_devices (udi);
          }
        else
          {
            for (HALVolumes::iterator vIter = hal->volumes.begin (), vEnd = hal->volumes.end (); vIter != vEnd; ++vIter)
              {
                HALVolume const& volume = vIter->second;
                if (volume.volume_udi == udi)
                {
                  HALVolume vol_erased = volume;
                  hal->volumes_by_udi.erase (volume.volume_udi);
                  hal->volumes.erase (vIter);
                  hal->signal_volume_removed_.emit (vol_erased); 
                  break;
                }
              }
          }
      }

      SignalVolume&
      HAL::signal_volume_removed ()
      {
        return signal_volume_removed_;
      }

      SignalVolume&
      HAL::signal_volume_added ()
      {
        return signal_volume_added_;
      }
  }
}
