#include <gtkmm.h>

#include "history.hpp"
#include "debug.hpp"

#include "x_playlist.hpp"

#define NONE (-1)
#define PLAYBACK_HISTORY_LIMIT 1000

using namespace Bmp;

static void
history_free (BmpPlaybackHistory *history)
{
    g_queue_foreach (history->history, GFunc (row_guid_free), NULL);
    g_queue_free (history->history);
    free (history);
}


BmpPlaybackHistory*
bmp_playback_history_new (void)
{
    BmpPlaybackHistory *history;

    history = g_new0 (BmpPlaybackHistory, 1);
    history->history  = g_queue_new ();
    history->mark     = 0;
    history->dirty    = false;
    history->mapping  = g_hash_table_new_full (GHashFunc      (row_guid_hash_func),
                                               GEqualFunc     (row_guid_equal_func),
                                               GDestroyNotify (row_guid_free),
                                               NULL);	

    return history;
}


void
bmp_playback_history_free (BmpPlaybackHistory *history)
{
    if (!history) return;
    history_free (history);
}

namespace
{
  bool
  history_item_prepend (BmpPlaybackHistory     *history,
		                    BmpPluginFlowInterface *plugin)
  {
    RowGUID               *guid;
    GtkListStore          *store;
    GtkTreeIter            iter;

    store = bmp_playlist_get_playlist (bmp_playlist);
    if (!bmp_plugin_flow_item_prev ((BmpPluginFlowInterface*)plugin, store, &iter))
      {
        debug ("history", "No prev item from the flow plugin.");
        return false;
      }

    if (history->history->length == PLAYBACK_HISTORY_LIMIT)
      { 
        debug ("history", "Popping tail.");
        g_hash_table_remove (history->mapping, g_queue_peek_tail (history->history));
        g_queue_pop_tail (history->history);
      }

    gtk_tree_model_get (GTK_TREE_MODEL (store), &iter, COLUMN_GUID, &guid, -1);
    g_queue_push_head (history->history, guid);
    g_hash_table_insert (history->mapping, row_guid_copy (guid), g_queue_peek_head_link (history->history));
    history->mark = 1;

    debug ("history", "Item prepended.");

    return true;
  }

  bool
  history_item_append (BmpPlaybackHistory     *history,
		                   BmpPluginFlowInterface *plugin)
  {
    RowGUID               *guid;
    GtkListStore          *store;
    GtkTreeIter            iter;

    store = bmp_playlist_get_playlist (bmp_playlist);
    if (!bmp_plugin_flow_item_next ((BmpPluginFlowInterface*)plugin, store, &iter))
      {
        debug ("history", "No next item from the flow plugin.");
        return false;
      }

    if (history->history->length == PLAYBACK_HISTORY_LIMIT)
      {
        debug ("history", "Popping head.");
        g_hash_table_remove (history->mapping, g_queue_peek_head (history->history));
        g_queue_pop_head (history->history);
      }

    gtk_tree_model_get (GTK_TREE_MODEL (store), &iter, COLUMN_GUID, &guid, -1);
    g_queue_push_tail (history->history, guid);
    g_hash_table_insert (history->mapping, row_guid_copy (guid), g_queue_peek_tail_link (history->history));
    history->mark++;

    debug ("history", "Item appended.");

    return true;
  }
}

int
bmp_playback_history_rewind (BmpPlaybackHistory *history)
{
    history->mark = 1;
    int row = 0;

    if (history->dirty && (history->history->length > 0))
      {
        row = bmp_playlist_get_row_by_guid (bmp_playlist, 
                  static_cast<RowGUID *> (g_queue_peek_nth (history->history, history->mark-1)));
      }

    return row;
}

void
bmp_playback_history_set (BmpPlaybackHistory *history,
                          int                 position)

{
    GtkListStore *store;
    GtkTreeIter   iter;
    GtkTreePath  *path;
    RowGUID	     *guid;

    history->dirty = true;

    if (int(history->mark) < int(history->history->length))
      {
        for (unsigned int n = 0; n < (history->history->length - history->mark); n++)
          {
              RowGUID *guid = static_cast<RowGUID *>(g_queue_peek_tail (history->history));
              g_hash_table_remove (history->mapping, guid);
              row_guid_free (guid);
              g_queue_pop_tail (history->history);
          }
      }

    store = bmp_playlist_get_playlist (bmp_playlist);
    path  = gtk_tree_path_new_from_indices (position, -1);
    gtk_tree_model_get_iter (GTK_TREE_MODEL (store), &iter, path);
    gtk_tree_path_free (path);
    gtk_tree_model_get (GTK_TREE_MODEL (store), &iter, COLUMN_GUID, &guid, -1);

    RowGUID *guid2 = (RowGUID*)(g_queue_peek_tail (history->history));
    if ((guid2 && !row_guid_equal_func (guid, guid2)) || !guid2)
      {
        g_queue_push_tail (history->history, guid);
        g_hash_table_insert (history->mapping, row_guid_copy (guid), g_queue_peek_tail_link (history->history)); 
        history->mark++;
      }

    debug ("history", "Item %d set at position %d", position, history->mark-1);
}

int
bmp_playback_history_prev (BmpPlaybackHistory     *history,
                           BmpPluginFlowInterface *plugin)
{
    history->dirty = true;
    int row = NONE;

    if (history->mark == 0)
      {
        if (history_item_prepend (history, plugin))
          {
            row = bmp_playlist_get_row_by_guid (bmp_playlist,
                      static_cast<RowGUID *> (g_queue_peek_nth (history->history, history->mark-1)));

            debug ("history", "Item prepended, row %d", row);
          }
      }
    else
      {
        if (--history->mark == 0)
          {
            return bmp_playback_history_prev (history, plugin);
          }
        else
          {
            row = bmp_playlist_get_row_by_guid (bmp_playlist, 
                      static_cast<RowGUID *> (g_queue_peek_nth (history->history, history->mark-1)));

            debug ("history", "Went back, row %d", row);
          }
      }

    return row;
}

int
bmp_playback_history_next (BmpPlaybackHistory       *history,
                           BmpPluginFlowInterface   *plugin)
{
    history->dirty = true;
    int row = NONE;

    if (int(history->mark) == int(history->history->length))
      {
        if (history_item_append (history, plugin))
          {
            row = bmp_playlist_get_row_by_guid (bmp_playlist,
                      static_cast<RowGUID *> (g_queue_peek_nth (history->history, history->mark-1)));

            debug ("history", "Item appended, row %d", row);
          }
      }
    else
      {
        history->mark++;
        row = bmp_playlist_get_row_by_guid (bmp_playlist, 
                  static_cast<RowGUID *> (g_queue_peek_nth (history->history, history->mark-1)));

        debug ("history", "Went forward, row %d", row);
      }

    return row;
}

void
bmp_playback_history_remove_guid (BmpPlaybackHistory *history, RowGUID *guid)
{
    GList *link = (GList*) g_hash_table_lookup (history->mapping, guid);
    if (!link) return;
    if ((history->mark-1) >= g_queue_link_index (history->history, link))
      { 
        history->mark--;
      }
    row_guid_free ((RowGUID*)link->data);
    g_queue_delete_link (history->history, link);
    g_hash_table_remove (history->mapping, guid);
}
