//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 2 
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glibmm.h>
#include <boost/algorithm/string.hpp>

// Musicbrainz
#include <musicbrainz/musicbrainz.h>

#include <glib/gi18n.h>
#include <glib/gstdio.h>
#include <gtkmm.h>

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <unistd.h>
#include <sys/stat.h>

#include <iostream>
#include <sstream>

#include <iostream>
#include <boost/shared_ptr.hpp>
#include <boost/format.hpp>
#include <boost/optional.hpp>

// BMP
#include "uri++.hpp"
#include "database.hpp"
#include "library.hpp"
#include "file_utils.hpp"
#include <sqlite3_bmp.h>

// TagLib
#include <taglib_plugins/tplugins.h>
#ifdef HAVE_MP4V2
#  include <taglib_plugins/mp4/mp4file.h>
#  include <taglib_plugins/mp4/mp4tag.h>
#endif

#include <taglib_plugins/wma/wmafile.h>
#include <taglib_plugins/wma/wmatag.h>

#include <taglib/taglib.h>
#include <taglib/fileref.h>
#include <taglib/audioproperties.h>
#include <taglib/tfile.h>
#include <taglib/tag.h>
#include <taglib/id3v2tag.h>
#include <taglib/mpegfile.h>
#include <taglib/oggfile.h>
#include <taglib/vorbisfile.h>
#include <taglib/vorbisproperties.h>
#include <taglib/xiphcomment.h>
#include <taglib/mpcfile.h>
#include <taglib/vorbisfile.h>
#include <taglib/flacfile.h>
#include <taglib/xiphcomment.h>
#include <taglib/id3v2framefactory.h>
#include <taglib/textidentificationframe.h>
#include <taglib/uniquefileidentifierframe.h>
#include <taglib/apetag.h>
#include <taglib/apeitem.h>
#include <taglib/apefooter.h>

// GStreamer
#include <gst/gst.h>
#include <gst/gstelement.h>
#include "audio.hpp"

#include "main.hpp"
#include "paths.hpp"
#include "debug.hpp"

#include "x_library.hpp"

#ifdef HAVE_HAL
#  include "x_hal.hpp"
#endif

#define TAG_READ_TUPLE(_data) ((TagReadTuple*)(_data))

#define MLIB_VERSION "10"

namespace Bmp
{
  namespace Library
  {
    char* const Exception::_message_types[] = { "NO_METADATA", "NO_ASIN", "NO_RELEASE", "NO_FILE", "FILE_INVALID", "FILE_NON_READABLE", "FILE_NON_WRITABLE", "CONVERSION_ERROR" }; 

    namespace
    {
      struct TagReadTuple
      {
          DB::Row *row;
          bool     eos;
      };

      DatumDefine track_attributes[] =
      {
          {
              N_("Location"),
              N_("Locations"),
              N_("The location of the track (file, internet stream, etc.)"),
              "location",
              DB::VALUE_TYPE_STRING,
              true,
          },

          {
              N_("Artist"),
              N_("Artists"),
              N_("The artist/peformer of the track"),
              "artist",
              DB::VALUE_TYPE_STRING,
              true,
          },

          {
              N_("Album"),
              N_("Albums"),
              N_("The album of the track"),
              "album",
              DB::VALUE_TYPE_STRING,
              true,
          },

          {
              N_("Title"),
              N_("Titles"),
              N_("The title of the current track"),
              "title",
              DB::VALUE_TYPE_STRING,
              true,
          },

          {
              N_("Track"),
              N_("Tracks"),
              N_("The track number of the track"),
              "tracknumber",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Time"),
              N_("Times"),
              N_("The length of the track"),
              "time",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Genre"),
              N_("Genres"),
              N_("The genre (kind of music) of the track"),
              "genre",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("Comment"),
              N_("Comments"),
              N_("Comments added to the track"),
              "comment",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("Rating"),
              N_("Ratings"),
              N_("The rating of the track"),
              "rating",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Date"),
              N_("Dates"),
              N_("The date (year) of the track"),
              "date",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("MTIME"),
              N_("MTIMEs"),
              N_("The modification time of the track"),
              "mtime",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Bitrate"),
              N_("Bitrates"),
              N_("The bitrate of the track"),
              "bitrate",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Samplerate"),
              N_("Samplerates"),
              N_("The samplerate of the track"),
              "samplerate",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Play count"),
              N_("Play counts"),
              N_("The number of times this track was played"),
              "count",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Last played date"),
              N_("Last played dates"),
              N_("The last date of when this track was played"),
              "playdate",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("ASIN"),
              N_("ASINs"),
              N_("Amazon Standard Index Number"),
              "asin",
              DB::VALUE_TYPE_STRING,
              false,
          },

#ifdef HAVE_HAL
          {
              N_("Volume UDI"),
              N_("Volume UDIs"),
              N_("HAL Volume UDI this track resides on"),
              "volume_udi",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("Device UDI"),
              N_("Device UDIs"),
              N_("HAL Device UDI this track resides on"),
              "device_udi",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("Volume Relative Path"),
              N_("Volume Relative Paths"),
              N_("Path of the track relative to it's volume's root"),
              "vrp",
              DB::VALUE_TYPE_STRING,
              false,
          },
#endif

          {
              N_("MusicBrainz Album Artist"),
              N_("MusicBrainz Album Artists"),
              N_("The MusicBrainz Release Artist"),
              "musicbrainz_albumartist",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("MusicBrainz Album Artist Id"),
              N_("MusicBrainz Album Artist Ids"),
              N_("The MusicBrainz Release Artist ID (The ID of the Artist this release is accredited to)"),
              "musicbrainz_albumartist_id",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("MusicBrainz Album Artist Sortname"),
              N_("MusicBrainz Album Artist Sortnames"),
              N_("The MusicBrainz Release Sort Name"),
              "musicbrainz_albumartistsortname",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("MusicBrainz Track ID"),
              N_("MusicBrainz Track IDs"),
              N_("The MusicBrainz Track ID"),
              "musicbrainz_track_id",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("MusicBrainz Album Id"),
              N_("MusicBrainz Album Ids"),
              N_("The MusicBrainz Release ID"),
              "musicbrainz_album_id",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("MusicBrainz Artist Id"),
              N_("MusicBrainz Artist Ids"),
              N_("The MusicBrainz Artist ID"),
              "musicbrainz_artist_id",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("MusicBrainz Artist Sortname"),
              N_("MusicBrainz Artist Sortnames"),
              N_("The MusicBrainz Artist Sort Name"),
              "musicbrainz_artistsortname",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("New Item"),
              N_("New Items"),
              N_("Whether this item was just addded and/or yet not approved"),
              "new_item",
              DB::VALUE_TYPE_BOOL,
              false,
          },
      };

      TagLib::ID3v2::UserTextIdentificationFrame*
      find_utif (TagLib::ID3v2::Tag *tag, const TagLib::String &description)
      {
          TagLib::ID3v2::FrameList l = tag->frameList("TXXX");
          for(TagLib::ID3v2::FrameList::Iterator it = l.begin(); it != l.end(); ++it) {
            TagLib::ID3v2::UserTextIdentificationFrame *f = dynamic_cast<TagLib::ID3v2::UserTextIdentificationFrame *>(*it);
            if(f && f->description() == description)
              return f;
          }
          return 0;
      }

      // Based on SoundJuicer code, (C) Ross Burton <ross@burtonini.com> 
      void
      add_txxx_tag (TagLib::ID3v2::Tag    *id3v2tag,
                    const Glib::ustring&   spec_id,
                    const Glib::ustring&   realworld_id,
                    const Glib::ustring&   id_str) 
      {
        using namespace Bmp;
        using namespace TagLib;

        ID3v2::UserTextIdentificationFrame * frame;

        if (id_str.empty()) return;

        if (!spec_id.empty())
        {
          frame = new ID3v2::UserTextIdentificationFrame (String::UTF8);
          id3v2tag->addFrame (frame);
          frame->setDescription (String (spec_id.c_str(), String::UTF8));
          frame->setText (String (id_str.c_str(), String::UTF8));
        }

        if (!realworld_id.empty())
        {
          frame = new ID3v2::UserTextIdentificationFrame (String::UTF8);
          id3v2tag->addFrame (frame);
          frame->setDescription (String (realworld_id.c_str(), String::UTF8));
          frame->setText (String (id_str.c_str(), String::UTF8));;
        }
      }

      int
      get_duration_from_sectors (int sectors)
      {
          const int bytes_per_sector   = 2352;
          const int bytes_per_seconds  = (44100 / 8) / 16 / 2;

          return (sectors * bytes_per_sector / bytes_per_seconds)/1000;
      }

      void
      tag_foreach (const GstTagList *list,
                   const gchar      *tag,
                   TagReadTuple     *tuple)
      {
          int count = gst_tag_list_get_tag_size (list, tag);

          for (int i = 0; i < count; i++)
            {
              const GValue *value = gst_tag_list_get_value_index (list, tag, i);

              if (G_VALUE_HOLDS (value, G_TYPE_INT))
                {
                  tuple->row->insert (std::make_pair (std::string (tag), int (g_value_get_int(value))));
                }
              else
              if (G_VALUE_HOLDS (value, G_TYPE_UINT))
                {
                  tuple->row->insert (std::make_pair (std::string (tag), int (g_value_get_uint(value))));
                }
              else
              if (G_VALUE_HOLDS (value, G_TYPE_STRING))
                {
                  const char *_v_C = g_value_get_string (value);
                  if (*_v_C)
                    {
                      std::string v = _v_C; 
                      boost::trim (v);
                      tuple->row->insert (std::make_pair (std::string (tag), v)); 
                    }
                }
            }
      }

      gboolean
      metadata_bus_handler (GstBus       *bus,
                            GstMessage   *message,
                            TagReadTuple *tuple)
      {
          switch (GST_MESSAGE_TYPE (message))
          {
            case GST_MESSAGE_EOS:
                {
                  tuple->eos = true;
                  break;
                }

            case GST_MESSAGE_TAG:
                {
                  GstTagList *tag_list = 0;
                  gst_message_parse_tag (message, &tag_list);
                  if (tag_list)
                    {
                      gst_tag_list_foreach (tag_list, (GstTagForeachFunc) tag_foreach, tuple);
                      gst_tag_list_free (tag_list);
                    }
                  break;
                }

            case GST_MESSAGE_ERROR:
                {
                  GError *error = 0;
                  gst_message_parse_error (message, &error, 0);
                  g_print ("%s ERROR: %s\n", G_STRLOC, error->message);
                  g_error_free (error);
                  break;
                }

            default: break;
          }
          return FALSE;
      }

      void
      metadata_event_loop (TagReadTuple *tuple,
                           GstElement   *element,
                           gboolean      block)
      {
          GstBus *bus = gst_element_get_bus (element);
          g_return_if_fail (bus != 0);

          bool done = false;
          while (!done && !tuple->eos)
          {
              GstMessage *message;

              if (block)
                  message = gst_bus_poll (bus, GST_MESSAGE_ANY, -1);
              else
                  message = gst_bus_pop  (bus);

              if (message == 0)
              {
                  gst_object_unref (bus);
                  return;
              }

              done = metadata_bus_handler (bus, message, tuple);
              gst_message_unref (message);
          }
          gst_object_unref (bus);
      }

    } // anonymous namespace

    void
    get_audiocd_metadata_offline (MusicBrainz &o, DB::VRow *vector)
    {
        using namespace std;

        if (o.Query (string (MBQ_GetCDTOC)))
          {
            g_log (BMP_LIBRARY_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s: Unable to get read CD TOC", G_STRLOC);
            return;
          }

        unsigned int n_tracks = o.DataInt (MBE_TOCGetLastTrack);

        for (unsigned int i = 1; i <= n_tracks; i++)
          {
            DB::Row row;
            row.insert (make_pair (track_attributes[DATUM_LOCATION].id, (boost::format ("cdda:///%d") % i).str ()));
            row.insert (make_pair (track_attributes[DATUM_TITLE].id,    (boost::format ("Track %d") % i).str ()));
            row.insert (make_pair (track_attributes[DATUM_TRACK].id,    int (i)));
            row.insert (make_pair (track_attributes[DATUM_TIME].id,     int (get_duration_from_sectors(o.DataInt(MBE_TOCGetTrackNumSectors, i)))));
            vector->push_back (row);
          }
        debug ("library", "AudioCD: Everything allright, status is MB_OK (Offline)");
    }

    void
    get_audiocd_metadata (int &status, DB::VRow &vector)
    {
        using namespace std;

        // Create the musicbrainz object, which will be needed for subsequent calls
        MusicBrainz o;
        o.UseUTF8 (true);
        o.SetDevice (mcs->key_get<string>("audio", "cdrom-device"));

        bool isMultipleArtist = false;
        int trackNum = 0;

        int ret;
        ret = o.Query (MBQ_GetCDTOC);
        if (!ret)
        {
            string error;
            o.GetQueryError (error);
            debug ("library", "%s: Query failed: %s", G_STRLOC, error.c_str());
            status = Bmp::Library::Library::MB_FAILURE;
            return;
        }
        debug ("library", "AudioCD: Got TOC");


        string cdindex = o.Data(string (MBE_TOCGetCDIndexId));
        if (cdindex.empty())
        {
            string error;
            o.GetQueryError (error);
            debug ("library", "%s: Query failed: %s", G_STRLOC, error.c_str());
            status = Bmp::Library::Library::MB_FAILURE;
            return;
        }
        debug ("library", "AudioCD: Got cdindexid");

        ret = o.Query(string (MBQ_GetCDInfo));
        if (!ret)
        {
            string error;
            o.GetQueryError (error);
            debug ("library", "%s: Query failed: %s", G_STRLOC, error.c_str());
            status = Bmp::Library::Library::MB_OFFLINE;
            get_audiocd_metadata_offline (o, &vector);
            return;
        }
        debug ("library", "AudioCD: Got CDInfo");

        // Select the first album
        list<int> argList;
        argList.push_back (1);
        if (!o.Select (string (MBS_SelectAlbum), &argList))
        {
            status = Bmp::Library::Library::MB_OFFLINE;
            get_audiocd_metadata_offline (o, &vector);
            return;
        }
        debug ("library", "AudioCD: Selected First Album");

        string album_id_url = o.Data(string (MBE_AlbumGetAlbumId));
        if (album_id_url.empty ())
          {
            debug ("library", "No AlbumIDURL");
            return;
          }

        debug ("library", "AudioCD: Got Album ID URL");

        string album_id;
        o.GetIDFromURL (album_id_url, album_id);

        // Extract the album name
        string album_name = o.Data(string (MBE_AlbumGetAlbumName));

        // Extract the number of tracks
        int n_tracks = o.DataInt(string (MBE_AlbumGetNumTracks));

        if (n_tracks < 1)
        {
          status = Bmp::Library::Library::MB_OFFLINE;
          get_audiocd_metadata_offline (o, &vector);
          return;
        }
        debug ("library", "More than 1 track, proceeding to get online data");

        // Check to see if there is more than one artist for this album
        string artist_check;
        for (int i = 1; i <= n_tracks; i++)
        {
            string artist = o.Data (string (MBE_AlbumGetArtistId), i);
            if (artist.empty ())
              {
                debug ("library", "No Artist");
                return;
              }

            if (i == 1)
            {
              artist_check = artist;
              continue;
            }

            if (artist.compare (artist_check))
            {
              isMultipleArtist = true;
              break;
            }
        }

        string artist_name, artist_id;

        if (!isMultipleArtist)
          {
            // Extract the artist name from the album
            artist_name = o.Data (string (MBE_AlbumGetArtistName), 1);

            // Extract the artist id from the album
            artist_id   = o.Data (string (MBE_AlbumGetArtistId), 1);
          }

        for(int i = 1; i <= n_tracks; i++)
          {
            DB::Row row;

            row.insert (make_pair (track_attributes[DATUM_LOCATION].id, (boost::format ("cdda://%s/%d") % cdindex % i).str ()));
            row.insert (make_pair (track_attributes[DATUM_ALBUM].id, Glib::ustring (album_name)));
            row.insert (make_pair (track_attributes[DATUM_BITRATE].id, int(1411)));
            row.insert (make_pair (track_attributes[DATUM_SAMPLERATE].id, int(44100)));

            // Extract the track name from the album.
            string title = o.Data (string (MBE_AlbumGetTrackName), i);
            if (!title.empty ())
              {
                row.insert (make_pair (track_attributes[DATUM_TITLE].id, DB::Variant (Glib::ustring (title))));
              }

            // Extract the album id from the track. Just use the
            // first album that this track appears on
            string track_id = o.Data (string (MBE_AlbumGetTrackId), i);
            if (!track_id.empty ())
              {
                // Extract the track number
                trackNum = o.GetOrdinalFromList (string (MBE_AlbumGetTrackList), track_id);
                if (trackNum > 0 && trackNum < 100)
                  {
                    row.insert (make_pair (track_attributes[DATUM_TRACK].id, DB::Variant (int(trackNum))));
                  }
              }

            // If its a multple artist album, print out the artist for each track
            if (isMultipleArtist)
              {
                // Extract the artist name from this track
                string individual_artist = o.Data(string (MBE_AlbumGetArtistName), i);
                if (!individual_artist.empty ())
                  {
                    row.insert (make_pair (track_attributes[DATUM_ARTIST].id, DB::Variant (Glib::ustring (individual_artist))));
                  }
              }
            else
              {
                row.insert (make_pair (track_attributes[DATUM_ARTIST].id, DB::Variant (Glib::ustring (artist_name))));
              }
            vector.push_back (row);
          }

        ret = o.Query(MBQ_GetCDTOC);
        if (!ret)
          {
            string error;
            o.GetQueryError (error);
            debug ("library", "%s: Query failed: %s", G_STRLOC, error.c_str());
            status = Bmp::Library::Library::MB_FAILURE;
            return;
          }
        debug ("library", "AudioCD: Got TOC again");

        for(int i = 1; i <= n_tracks; i++)
          {
            DB::Row& row = vector[i-1];
            int duration = get_duration_from_sectors(o.DataInt(MBE_TOCGetTrackNumSectors, i+1));
            row.insert (make_pair (track_attributes[DATUM_TIME].id, duration)); 
          } 

        status = Bmp::Library::Library::MB_OK;
        debug ("library", "AudioCD: Everything allright, status is MB_OK");
    }

    void
    Library::metadata_set_taglib_common (const UpdateTrack& track,
                                         TagLib::Tag* tag)
    {
      using namespace TagLib;
      if (!tag) return;

      if (track.title)
          tag->setTitle   (String (track.title.get().c_str(), String::UTF8)); 

      if (track.artist)
          tag->setArtist  (String (track.artist.get().c_str(), String::UTF8));

      if (track.album)
          tag->setAlbum   (String (track.album.get().c_str(), String::UTF8));

      if (track.comment)
          tag->setComment (String (track.comment.get().c_str(), String::UTF8));

      if (track.genre)
          tag->setGenre   (String (track.genre.get().c_str(), String::UTF8));

      if (track.date)
          tag->setYear    (track.date.get());

      if (track.tracknumber)
          tag->setTrack   (track.tracknumber.get());
    }

    void 
    Library::metadata_set_taglib (const Glib::ustring&  location,
                                  const UpdateTrack&    track)
    {
        using namespace std;
        using namespace Glib;
        using namespace TagLib;

        string filename = Glib::filename_from_uri (location); 
        string type;

        if (!Audio::typefind (filename, type))
          {
            debug ("library", "Unable to modify file %s", filename.c_str());
            throw Exception ("Unable to modify file", Exception::UNABLE_TO_MODIFY, filename);
          }

        if (type == "application/x-id3")
          {
            ID3v2::FrameFactory *factory = ID3v2::FrameFactory::instance(); 
            factory->setDefaultTextEncoding (String::UTF8);
            MPEG::File mpegfile (filename.c_str(), factory);

            if (!mpegfile.isOpen() && mpegfile.isValid())
                throw Exception ("Unable to get an MPEG::File* from filename", Exception::UNABLE_TO_MODIFY, filename);

            debug ("library", "Modifying MPEG::File: %s", filename.c_str());

            ID3v2::Tag *tag = mpegfile.ID3v2Tag (true);

            metadata_set_taglib_common (track, tag);

            if (track.mb_album_artist)
              {
                add_txxx_tag (tag,
                              track_attributes[DATUM_MB_ALBUM_ARTIST].description,
                              track_attributes[DATUM_MB_ALBUM_ARTIST].id,
                              track.mb_album_artist.get());
              }

            if (track.mb_album_artist_id)
              {
                add_txxx_tag (tag,
                              track_attributes[DATUM_MB_ALBUM_ARTIST_ID].description,
                              track_attributes[DATUM_MB_ALBUM_ARTIST_ID].id,
                              track.mb_album_artist_id.get());
              }

            if (track.mb_album_artist_sort_name)
              {
                add_txxx_tag (tag,
                              track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].description,
                              track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id,
                              track.mb_album_artist_sort_name.get());
              }

            if (track.mb_track_id)
              {
                ID3v2::UniqueFileIdentifierFrame *frame = new ID3v2::UniqueFileIdentifierFrame ("http://musicbrainz.org", track.mb_track_id.get().c_str());
                tag->addFrame (frame);
              }

            if (track.mb_album_id)
              {
                add_txxx_tag (tag,
                              track_attributes[DATUM_MB_ALBUM_ID].description,
                              track_attributes[DATUM_MB_ALBUM_ID].id,
                              track.mb_album_id.get());
              }

            if (track.mb_artist_id)
              {
                add_txxx_tag (tag,
                              track_attributes[DATUM_MB_ARTIST_ID].description,
                              track_attributes[DATUM_MB_ARTIST_ID].id,
                              track.mb_artist_id.get());
              }

            if (track.mb_artist_sort_name)
              {
                add_txxx_tag (tag,
                              track_attributes[DATUM_MB_ARTIST_SORTNAME].description,
                              track_attributes[DATUM_MB_ARTIST_SORTNAME].id,
                              track.mb_artist_sort_name.get());
              }

            if (track.asin)
              {
                add_txxx_tag (tag,
                            track_attributes[DATUM_ASIN].description,
                            track_attributes[DATUM_ASIN].id,
                            track.asin.get());
              }

            mpegfile.save(TagLib::MPEG::File::ID3v1 | TagLib::MPEG::File::ID3v2);
            return;
          }
        else 
        if (type == "application/ogg")
          {
            Ogg::Vorbis::File oggfile (filename.c_str());

            if (!oggfile.isOpen() && oggfile.isValid())
                throw Exception ("Unable to get an Ogg::Vorbis::File from filename", Exception::UNABLE_TO_MODIFY, filename);

            debug ("library", "Modifying OGG::File: %s", filename.c_str());

            metadata_set_taglib_common (track, oggfile.tag());

            TagLib::Ogg::XiphComment *comment = oggfile.tag(); 

            if (comment)
              {
                if (track.mb_album_artist)
                  comment->addField (String ("MUSICBRAINZ_ALBUMARTIST",             String::UTF8),
                                     String (track.mb_album_artist.get(),           String::UTF8));
                if (track.mb_album_artist_id)
                  comment->addField (String ("MUSICBRAINZ_ALBUMARTISTID",           String::UTF8),
                                     String (track.mb_album_artist_id.get(),        String::UTF8));
                if (track.mb_album_artist_sort_name)
                  comment->addField (String ("MUSICBRAINZ_ALBUMARTISTSORTNAME",     String::UTF8),
                                     String (track.mb_album_artist_sort_name.get(),  String::UTF8));
                if (track.mb_track_id)
                  comment->addField (String ("MUSICBRAINZ_TRACKID",                 String::UTF8),
                                     String (track.mb_track_id.get(),               String::UTF8));
                if (track.mb_album_id)
                  comment->addField (String ("MUSICBRAINZ_ALBUMID",                 String::UTF8),
                                     String(track.mb_album_id.get(),                String::UTF8));
                if (track.mb_artist_id)
                  comment->addField (String ("MUSICBRAINZ_ARTISTID",                String::UTF8),
                                     String (track.mb_artist_id.get(),              String::UTF8));
                if (track.mb_artist_sort_name)
                  comment->addField (String ("MUSICBRAINZ_SORTNAME",                String::UTF8),
                                     String (track.mb_artist_sort_name.get(),        String::UTF8));
                if (track.asin)
                  comment->addField (String ("ASIN",           String::UTF8),
                                     String (track.asin.get(), String::UTF8));
              }
            oggfile.save ();
            return;
        }
#ifdef HAVE_MP4V2
      else
      if (type == "audio/x-m4a")
        {
          TagLib::MP4::File mp4file (filename.c_str());

          if (!mp4file.isOpen() && mp4file.isValid())
            throw Exception ("Unable to get an MP4::File from filename", Exception::UNABLE_TO_MODIFY, filename);

          MP4::Tag *tag = mp4file.getMP4Tag ();
          if (tag)
            {
              debug ("library", "Modifying MP4::File: %s", filename.c_str());

              metadata_set_taglib_common (track, tag);

              MP4FileHandle handle = mp4file.getMP4File ();

              if (track.mb_album_artist_sort_name)    
              MP4SetMetadataFreeForm (handle,
                                      "MusicBrainz Sortname",
                                      (uint8_t*)track.mb_album_artist_sort_name.get().c_str(),
                                      strlen(track.mb_album_artist_sort_name.get().c_str()) + 1);

              if (track.mb_track_id)    
              MP4SetMetadataFreeForm (handle,
                                      "MusicBrainz Track Id",
                                      (uint8_t*)track.mb_track_id.get().c_str(),
                                      strlen(track.mb_track_id.get().c_str()) + 1);

              if (track.mb_album_id)    
              MP4SetMetadataFreeForm (handle,
                                      "MusicBrainz Album Id",
                                      (uint8_t*)track.mb_album_id.get().c_str(),
                                      strlen(track.mb_album_id.get().c_str()) + 1);

              if (track.mb_artist_id)    
              MP4SetMetadataFreeForm (handle,
                                      "MusicBrainz Artist Id",
                                      (uint8_t*)track.mb_artist_id.get().c_str(),
                                      strlen(track.mb_artist_id.get().c_str()) + 1);

              if (track.mb_album_artist_id)    
              MP4SetMetadataFreeForm (handle,
                                      "MusicBrainz Album Artist Id",
                                      (uint8_t*)track.mb_album_artist_id.get().c_str(),
                                      strlen(track.mb_album_artist_id.get().c_str()) + 1);

              if (track.mb_album_artist_sort_name)    
              MP4SetMetadataFreeForm (handle,
                                      "MusicBrainz Album Artist Sortname",
                                      (uint8_t*)track.mb_album_artist_sort_name.get().c_str(),
                                      strlen(track.mb_album_artist_sort_name.get().c_str()) + 1);

              if (track.mb_album_artist)    
              MP4SetMetadataFreeForm (handle,
                                      "MusicBrainz Album Artist",
                                      (uint8_t*)track.mb_album_artist.get().c_str(),
                                      strlen(track.mb_album_artist.get().c_str()) + 1);
              mp4file.save ();
              return;
            }
          else
            {
              throw Exception ("Unable to modify file", Exception::UNABLE_TO_MODIFY, filename);
            }
        }
#endif
      else
      if (type == "video/x-ms-asf")
        {
          TagLib::WMA::File wmafile (filename.c_str());

          if (!wmafile.isOpen() && wmafile.isValid())
            throw Exception ("Unable to get an WMA::File from filename", Exception::UNABLE_TO_MODIFY, filename);

            TagLib::WMA::Tag *tag = wmafile.WMATag();
            if (tag)
              {
                debug ("library", "Modifying WMA::File: %s", filename.c_str());

                metadata_set_taglib_common (track, tag);      

                if (track.mb_track_id)
                tag->setAttribute("MusicBrainz/TrackId",
                                   TagLib::String(track.mb_track_id.get().c_str(),
                                   TagLib::String::UTF8));

                if (track.mb_artist_id)
                tag->setAttribute("MusicBrainz/ArtistId",
                                   TagLib::String(track.mb_artist_id.get().c_str(),
                                   TagLib::String::UTF8));

                if (track.mb_artist_id)
                tag->setAttribute("MusicBrainz/SortName",
                                   TagLib::String(track.mb_artist_id.get().c_str(),
                                   TagLib::String::UTF8));

                if (track.mb_album_id)
                tag->setAttribute("MusicBrainz/AlbumId",
                                   TagLib::String(track.mb_album_id.get().c_str(),
                                   TagLib::String::UTF8));

                if (track.mb_album_artist_id)
                tag->setAttribute("MusicBrainz/AlbumArtistId",
                                   TagLib::String(track.mb_album_artist_id.get().c_str(),
                                   TagLib::String::UTF8));

                if (track.mb_album_artist_sort_name)
                tag->setAttribute("MusicBrainz/AlbumArtistSortName",
                                   TagLib::String(track.mb_album_artist_id.get().c_str(),
                                   TagLib::String::UTF8));

                wmafile.save();
                return;
              }
            else
              {
                throw Exception ("Unable to modify file", Exception::UNABLE_TO_MODIFY, filename);
              }
        }
      else
      if (type == "audio/x-flac")
        {
          TagLib::FLAC::File file (filename.c_str());
          if (file.isValid() && file.tag() && !file.readOnly())
            {
              debug ("library", "Modifying FLAC::File: %s", filename.c_str());
              metadata_set_taglib_common (track, file.tag());
              file.save();
              return;
            }
        }
      else
      if (type == "application/x-apetag")
        {
          TagLib::MPC::File file (filename.c_str());
          if (file.isValid() && file.tag() && !file.readOnly())
            {
              debug ("library", "Modifying MPC::File: %s", filename.c_str());
              metadata_set_taglib_common (track, file.tag());
              APE::Tag *tag = file.APETag (true);
              if (tag)
                {
                  if (track.mb_album_artist)
                    tag->addValue ("MUSICBRAINZ_ALBUMARTIST",
                                    String (track.mb_album_artist.get(),            String::UTF8));

                  if (track.mb_album_artist_id)
                    tag->addValue ("MUSICBRAINZ_ALBUMARTISTID",
                                    String (track.mb_album_artist_id.get(),         String::UTF8));

                  if (track.mb_album_artist_sort_name)
                    tag->addValue ("MUSICBRAINZ_ALBUMARTISTSORTNAME",
                                    String (track.mb_album_artist_sort_name.get(),  String::UTF8));

                  if (track.mb_track_id)
                    tag->addValue ("MUSICBRAINZ_TRACKID",
                                    String (track.mb_track_id.get(),                String::UTF8));

                  if (track.mb_album_id)
                    tag->addValue ("MUSICBRAINZ_ALBUMID",
                                    String(track.mb_album_id.get(),                 String::UTF8));

                  if (track.mb_artist_id)
                    tag->addValue ("MUSICBRAINZ_ARTISTID",
                                    String (track.mb_artist_id.get(),               String::UTF8));

                  if (track.mb_artist_sort_name)
                    tag->addValue ("MUSICBRAINZ_SORTNAME",
                                    String (track.mb_artist_sort_name.get(),        String::UTF8));

                  if (track.asin)
                    tag->addValue ("ASIN",
                                    String (track.asin.get(),                       String::UTF8));
                } 
              file.save();
              return;
            }
        }

      debug ("library", "Modifying Unable to modify file %s", filename.c_str());
      throw Exception ("Unable to modify file", Exception::UNABLE_TO_MODIFY, filename);
    }

    void 
    Library::metadata_get_taglib (Glib::ustring const&  uri,
                                  DB::Row&              row)
    {
      using namespace TagLib;
      using namespace boost;
      using namespace std;
      using namespace Glib;

      string filename;
      try {
          filename = filename_from_uri (uri);
        }
      catch (ConvertError& cxe)
        {
          throw Exception ("Unable to convert URI to filename", Exception::CONVERSION_ERROR, uri);
        } 

      if (!File::isReadable (filename.c_str())) throw Exception ("File is not readable", Exception::FILE_NON_READABLE, filename);

      File *track = 0;

      const Datum extra_tags_mpeg[] =
      {
        DATUM_MB_ALBUM_ARTIST,
        DATUM_MB_ALBUM_ARTIST_ID,
        DATUM_MB_ALBUM_ARTIST_SORTNAME,
        DATUM_MB_ALBUM_ID,
        DATUM_MB_ARTIST_ID,
        DATUM_MB_ARTIST_SORTNAME,
        DATUM_ASIN,
      };

      const Datum extra_tags_vorbis[] =
      {
        DATUM_MB_ALBUM_ARTIST,
        DATUM_MB_ALBUM_ARTIST_ID,
        DATUM_MB_ALBUM_ARTIST_SORTNAME,
        DATUM_MB_TRACK_ID,
        DATUM_MB_ALBUM_ID,
        DATUM_MB_ARTIST_ID,
        DATUM_MB_ARTIST_SORTNAME,
        DATUM_ASIN,
      };

      const char* extra_tags_vorbis_ids[] =
      {
        "MUSICBRAINZ_ALBUMARTIST",
        "MUSICBRAINZ_ALBUMARTISTID",
        "MUSICBRAINZ_ALBUMARTISTSORTNAME",
        "MUSICBRAINZ_TRACKID",
        "MUSCIBRAINZ_ALBUMID",
        "MUSICBRAINZ_ARTISTID",
        "MUSICBRAINZ_SORTNAME",
        "ASIN",
      };

      TagLib::FileRef fileref;
      fileref = TagLib::FileRef (filename.c_str());

      if (fileref.isNull()) throw Exception ("Unable to create a TagLib FileRef for this file", Exception::NO_METADATA, filename);
      if (!fileref.file()->isValid()) throw Exception ("TagLib::FileRef is not valid", Exception::NO_METADATA, filename);

      boost::optional<ustring> type;
      string type_found;

      if (Audio::typefind (filename, type_found))
        {
          type = type_found;
        }

      if (type && (type.get() == "application/x-id3"))
        {
          using namespace TagLib;

          MPEG::File *file;
          if (!(file = dynamic_cast<MPEG::File*> (fileref.file())))
            {
                  throw Exception ("Unable to cast FileRef to MPEG::File",
                                    Exception::NO_METADATA, filename);
            }

          ID3v2::Tag *tag = file->ID3v2Tag (false);
          if (tag) 
            {
              using boost::algorithm::split;
              using boost::algorithm::is_any_of;

              ID3v2::UserTextIdentificationFrame *frame = 0;

              for (unsigned int n = 0; n < G_N_ELEMENTS(extra_tags_mpeg); ++n) 
              {
                  frame = find_utif (tag, String (track_attributes[extra_tags_mpeg[n]].id, String::UTF8));
                  if (frame)
                    {
                      vector<string> subs;
                      string str = frame->toString().to8Bit(true);
                      split (subs, str, is_any_of (" "));

                      row.insert (make_pair (track_attributes[extra_tags_mpeg[n]].id, str.substr (subs[0].length()+1))); 
                    }
              }

              // Check for UFID
              using TagLib::ID3v2::FrameList;
              FrameList const& map = tag->frameListMap()["UFID"];
              if (!map.isEmpty())
                {
                  for (FrameList::ConstIterator iter = map.begin(), end = map.end (); iter != end; ++iter)
                    {
                    ID3v2::UniqueFileIdentifierFrame *ufid = reinterpret_cast<ID3v2::UniqueFileIdentifierFrame*>(*iter);
                    if (ufid->owner() == "http://musicbrainz.org")
                      {
                        ByteVector vec = ufid->identifier();
                        vec.append ('\0');
                        row.insert (make_pair (track_attributes[DATUM_MB_TRACK_ID].id, ustring(vec.data())));
                        break;
                      }
                    }
                }
              }
          track = file;
        }
      else
      if (type && (type.get() == "application/ogg"))
        {
          using namespace TagLib;

          Ogg::Vorbis::File *file = 0;

          if (!(file = dynamic_cast<Ogg::Vorbis::File*> (fileref.file())))
            {
                      throw Exception ("Unable to cast FileRef to Ogg::Vorbis::File",
                                        Exception::NO_METADATA, filename);
            }

          Ogg::XiphComment *comment = dynamic_cast<Ogg::Vorbis::File*>(file)->tag (); 

          if (comment)
            {
              TagLib::Map<TagLib::String, TagLib::StringList> const& map = comment->fieldListMap();
              TagLib::Map<TagLib::String, TagLib::StringList>::ConstIterator iter;

              for (unsigned int n = 0; n < G_N_ELEMENTS (extra_tags_vorbis); ++n)
              {
                iter = map.find (extra_tags_vorbis_ids[n]);
                if (iter == map.end()) continue;
                TagLib::StringList const& list = iter->second;
                row.insert (make_pair (track_attributes[extra_tags_vorbis[n]].id, list[0].to8Bit(true)));
              }
            }
          track = file;
        }
      else
      if (type && (type.get() == "video/x-ms-asf"))
        {
          WMA::File *file;
          if (!(file = dynamic_cast<WMA::File*> (fileref.file())))
            {
                  throw Exception ("Unable to cast FileRef to WMA::File",
                                    Exception::NO_METADATA, filename);
            }
          WMA::Tag *tag = file->WMATag ();
          const TagLib::WMA::AttributeMap &attrMap = tag->attributeMap();

          struct { 
              Bmp::Library::Datum   datum;
              char                 *id;
          } extra_tags_wma[] = {
              { DATUM_MB_TRACK_ID,                "MusicBrainz/TrackId"             },
              { DATUM_MB_ARTIST_ID,               "MusicBrainz/ArtistId"            },
              { DATUM_MB_ALBUM_ID,                "MusicBrainz/AlbumId"             },
              { DATUM_MB_ALBUM_ARTIST,            "MusicBrainz/AlbumArtist"         },
              { DATUM_MB_ALBUM_ARTIST_ID,         "MusicBrainz/AlbumArtistId"       },
              { DATUM_MB_ALBUM_ARTIST_SORTNAME,  "MusicBrainz/AlbumArtistSortName"  },
              { DATUM_MB_ARTIST_SORTNAME,        "MusicBrainz/SortName"             },
          };

          for (unsigned int n = 0; n < G_N_ELEMENTS (extra_tags_wma); ++n)
            {
              if (attrMap.contains(extra_tags_wma[n].id))
                  row.insert (make_pair (track_attributes[extra_tags_wma[n].datum].id, 
                                         ustring (attrMap[extra_tags_wma[n].id]->toString().toCString(true))));
            }
        }
#ifdef HAVE_MP4V2
      else
      if (type && (type.get() == "audio/x-m4a"))
        {
          MP4::File *file;
          if (!(file = dynamic_cast<MP4::File*> (fileref.file())))
            {
                  throw Exception ("Unable to cast FileRef to MP4::File",
                                    Exception::NO_METADATA, filename);
            }

          MP4FileHandle mp4file = file->getMP4File();

          struct { 
              Bmp::Library::Datum   datum;
              char                 *id;
          } extra_tags_mp4[] = {
              { DATUM_MB_ARTIST_SORTNAME,         "MusicBrainz Sortname"              },
              { DATUM_MB_TRACK_ID,                "MusicBrainz Track Id"              },
              { DATUM_MB_ALBUM_ID,                "MusicBrainz Album Id"              },
              { DATUM_MB_ARTIST_ID,               "MusicBrainz Artist Id"             },
              { DATUM_MB_ALBUM_ARTIST_ID,         "MusicBrainz Album Artist Id"       },
              { DATUM_MB_ALBUM_ARTIST_SORTNAME,   "MusicBrainz Album Artist Sortname" },
              { DATUM_MB_ALBUM_ARTIST,            "MusicBrainz Album Artist"          },
          };

          for (unsigned int n = 0; n < G_N_ELEMENTS (extra_tags_mp4); ++n) 
            {
              char *value = 0;
              uint32_t size = 0;
              if (MP4GetMetadataFreeForm (mp4file, extra_tags_mp4[n].id, 
                                                (uint8_t**)&value, &size) && value != NULL)
                {
                  value = (char*)realloc (value, size+1);
                  value[size] = '\0';
                  row.insert (make_pair (track_attributes[extra_tags_mp4[n].datum].id, ustring(value)));
                  free (value);
                }
            }
        }
#endif
      else
      if (type && (type.get() == "application/x-apetag"))
        {
          MPC::File *file;
          if (!(file = dynamic_cast<MPC::File*> (fileref.file())))
            {
                  throw Exception ("Unable to cast FileRef to MPC::File",
                                    Exception::NO_METADATA, filename);
            }

          APE::Tag *tag = file->APETag();
          if (tag)
            {
              const TagLib::APE::ItemListMap &items = tag->itemListMap();
 
              for (unsigned int n = 0; n < G_N_ELEMENTS (extra_tags_vorbis); ++n)
              {
                if (!items[extra_tags_vorbis_ids[n]].isEmpty())
                   row.insert (make_pair (track_attributes[extra_tags_vorbis[n]].id, 
                                          ustring (items[extra_tags_vorbis_ids[n]].toString().toCString(true))));
              }
            }
        }
      else
        {
          track = fileref.file ();
        }

      if (!track) throw Exception ("Unable to read metadata with taglib", Exception::NO_METADATA, filename);

      struct stat file_stat;
      stat (filename.c_str(), &file_stat);
      string value;

      row.insert (make_pair (track_attributes[DATUM_TRACK].id,    int     (track->tag()->track())));
      row.insert (make_pair (track_attributes[DATUM_DATE].id,     int     (track->tag()->year())));

      row.insert (make_pair (track_attributes[DATUM_LOCATION].id, ustring (uri)));

      value = track->tag()->artist().to8Bit (true);

      if (!value.empty())
        {
          trim (value);
          row.insert (make_pair (track_attributes[DATUM_ARTIST].id, value)); 
        }

      value = track->tag()->album().to8Bit (true);
      if (!value.empty())
        {
          trim (value);
          row.insert (make_pair (track_attributes[DATUM_ALBUM].id, value)); 
        }

      value = track->tag()->genre().to8Bit (true);
      if (!value.empty())
        {
          trim (value);
          row.insert (make_pair (track_attributes[DATUM_GENRE].id, value));
        }

      value = track->tag()->comment().to8Bit (true);
      if (!value.empty())
        {
          trim (value);
          row.insert (make_pair (track_attributes[DATUM_COMMENT].id, value)); 
        }

      row.insert (make_pair (track_attributes[DATUM_RATING].id, int (0)));
      row.insert (make_pair (track_attributes[DATUM_COUNT].id, int (0)));

      if (track->audioProperties())
        {
          row.insert (make_pair (track_attributes[DATUM_BITRATE].id, int (track->audioProperties()->bitrate())));
          row.insert (make_pair (track_attributes[DATUM_SAMPLERATE].id, int (track->audioProperties()->sampleRate())));
          row.insert (make_pair (track_attributes[DATUM_TIME].id, int (track->audioProperties()->length())));
        }
      else
        {
          row.insert (make_pair (track_attributes[DATUM_BITRATE].id, int (0)));
          row.insert (make_pair (track_attributes[DATUM_SAMPLERATE].id, int (0)));
          row.insert (make_pair (track_attributes[DATUM_TIME].id, int (0)));
        }

      if  ((!track->tag()->title().toCString(true)) || (!strlen(track->tag()->title().toCString(true))))
        {
          row.insert (make_pair (track_attributes[DATUM_TITLE].id, string (path_get_basename(filename))));
        }
      else
        {
          value = track->tag()->title().to8Bit(true);
          if (!value.empty())
            {
              trim (value);
              row.insert (make_pair (track_attributes[DATUM_TITLE].id, value)); 
            }
        }
    }

    void 
    Library::metadata_get_gst  (const Glib::ustring&  uri,
                                DB::Row&          row)
    {
      using namespace std;
      using namespace Glib;

      string filename;

      try {
          filename = filename_from_uri (uri);
        }
      catch (ConvertError& cxe)
        {
          throw Exception ("Unable to convert URI to filename", Exception::CONVERSION_ERROR, uri);
        } 

      boost::shared_ptr<TagReadTuple> ttuple;

      ttuple = boost::shared_ptr<TagReadTuple> (new TagReadTuple);
      ttuple->eos = false;
      ttuple->row = &row;

      ttuple->row->insert (std::make_pair (track_attributes[DATUM_LOCATION].id, std::string (uri)));

      GstElement *element   = gst_element_factory_make ("playbin",  "playbin0");
      GstElement *fakesink  = gst_element_factory_make ("fakesink", "fakesink0");
      GstElement *videosink = gst_element_factory_make ("fakesink", "fakesink1");

      g_object_set (G_OBJECT (element), "uri", uri.c_str(), "audio-sink", fakesink, "video-sink", videosink, NULL);

      GstStateChangeReturn state_ret = gst_element_set_state (element, GST_STATE_PAUSED);
      int change_timeout = 5;

      while (state_ret == GST_STATE_CHANGE_ASYNC && change_timeout > 0)
        {
          GstState state;
          state_ret = gst_element_get_state (GST_ELEMENT (element), &state, 0, 1 * GST_SECOND);
          change_timeout--;
        }

      if (state_ret != GST_STATE_CHANGE_FAILURE && state_ret != GST_STATE_CHANGE_ASYNC)
        {
          // Post application specific message so we'll know when to stop the message loop
          GstBus *bus;
          bus = gst_element_get_bus (GST_ELEMENT (element));
          if (bus)
            {
              gst_bus_post (bus, gst_message_new_application (GST_OBJECT (element), 0));
              gst_object_unref (bus);
            }
          // Poll the bus for messages
          metadata_event_loop (ttuple.get(), GST_ELEMENT (element), FALSE);
        }

      GstFormat  format = GST_FORMAT_TIME;
      GstQuery  *query  = gst_query_new_duration (format);

      if (gst_element_query (GST_ELEMENT(element), query))
        {
          gst_element_get_state (GST_ELEMENT(element), 0, 0, 100 * GST_MSECOND);
          gint64 length_in_nanoseconds;
          gst_query_parse_duration (query, &format, &length_in_nanoseconds);
          int length = length_in_nanoseconds / GST_SECOND;
          ttuple->row->insert (std::make_pair (get_metadatum_id (DATUM_TIME), DB::Variant (length)));
        }
      gst_query_unref (query);
      gst_element_set_state (GST_ELEMENT (element), GST_STATE_NULL);
      gst_object_unref (GST_ELEMENT (element));
    }

    Glib::ustring
    Library::get_titlestring (const Glib::ustring& uri, std::string const& format_)
    {
      GString     *string_out   = 0;
      GString     *string_field = 0;
      char        *field;
      const char  *format = format_.c_str();
      int          n;
      bool         hit = false;

      Bmp::DB::Row row;
      get_metadata (uri, row); 

      string_out = g_string_new ("");
      while (*format)
      {
        // Begin sequence
        if ((*format == '%') && (*(format+1) != 0) && (*(format+1) == '{'))
        {
          string_field = g_string_new ("");
          format += 2;
          while (*format && (*format != '}') && (*format != ' '))
            {
              g_string_append_c (string_field, *format);
              format++;
            }

          field = g_string_free (string_field, FALSE);
          if (!*format)
            {
              free (field);
              char *str = g_string_free (string_out, FALSE);
              std::string stdstr = str;
              free (str);
              return stdstr;
            }

          // Append metadata field
          for (n = 0; n < Bmp::Library::DATUM_LAST; ++n)
            {
              if (!g_ascii_strcasecmp (field, get_metadatum_id (Bmp::Library::Datum (n)).c_str())) 
                {
                  Bmp::DB::Variant v = row[std::string(field)];
                  char *value = 0;

                  if (Bmp::DB::ValueType(v.which()) == Bmp::DB::VALUE_TYPE_INT)
                    {
                      value = g_strdup_printf ("%d", boost::get<int>(v)); 
                    }
                  else if (Bmp::DB::ValueType(v.which()) == Bmp::DB::VALUE_TYPE_STRING)
                    {
                      value = strdup (boost::get<Glib::ustring>(v).c_str());
                    }

                  if (value)
                    {  
                      g_strstrip (value);
                      if (strlen(value) && strcmp(value, "0") != 0)
                        {
                          hit = true;
                          g_string_append (string_out, value);
                        }
                      free (value);
                    } 
                  break;
               }
            }
          free (field);
        }
      else
        {
          g_string_append_c (string_out, *format);
        }
      format++;
    }

    if (!hit)
      {
        g_string_free (string_out, TRUE);
        return Glib::path_get_basename (uri);
      }
    else
      {
        char *str = g_string_free (string_out, FALSE);
        Glib::ustring ustr = str;
        free (str);
        return ustr;
      }
    }

    void 
    Library::add_track (const Glib::ustring &uri)
    {
      using namespace Bmp::DB;
      using namespace Glib; 

      Row row;

      try {
          metadata_get_taglib (uri, row);
        }
      catch (Bmp::Library::Exception& cxe)
        {
          if (cxe.id() != Exception::FILE_INVALID) metadata_get_gst (uri, row);
        }

      if (!row.empty ())
        {
                Track track (row);
                if (track.artist && track.album && track.title)
                  {
                    debug ("library", "Track '%s' qualifies for insertion", uri.c_str());
#ifdef HAVE_HAL
                    if (hal->is_initialized())
                    try {
                          HALVolume volume = hal->get_volume_for_uri (uri);
                          row.insert (std::make_pair (track_attributes[DATUM_HAL_VOLUME_UDI].id, volume.volume_udi));
                          row.insert (std::make_pair (track_attributes[DATUM_HAL_DEVICE_UDI].id, volume.device_udi));
                          Bmp::URI u (track.location.get());

                          ustring volume_relative_path = filename_from_uri (uri).substr (volume.mount_path.length());

                          row.insert (std::make_pair (track_attributes[DATUM_VOLUME_RELATIVE_PATH].id, volume_relative_path));
                        }
                     catch (HalException& cxe) { return; }
#endif
                    // Mark as new
                    debug ("library", "Marking track '%s' as new", uri.c_str());
                    row.insert (make_pair (track_attributes[DATUM_NEW_ITEM].id, bool(true))); 

                    // Try to add the row
                    try {
                          debug ("library", "Attempting to add row for track '%s'", uri.c_str());
                          db->add ("main", row);
                      }
                    catch (Bmp::DB::Exception& cxe)
                      {
                        debug ("library", "Unable to add row for track '%s', checking for duplicate", uri.c_str());
                        // Does the pkey exist?
                        if (db->attr_exists ("main", track_attributes[DATUM_LOCATION].id, track.location.get())); 
                          {
                            debug ("library", "Duplicate exists, replacing.", uri.c_str());
                            // Looks like it. Replace it 
                            try {
                              db->remove_by_attr ("main", track_attributes[DATUM_LOCATION].id, track.location.get()); 
                              db->add ("main", row);
                              UpdateTrack update_track;
                              update_track = track;
                              signal_row_updated_.emit (track.location.get(), update_track);   
                              }
                            catch (Bmp::DB::Exception& cxe)
                              {
                                debug ("library", "Unable to replace", uri.c_str());
                                return;
                              }
                          }
                      }
                  }
                else
                  {
                    debug ("library", "Track '%s' does NOT qualify for insertion", uri.c_str());
                  }
        }
    }

    void
    Library::del (const DB::AttributeList &attrs)
    {
      signal_processing_start_.emit ();
      db->del("main", attrs);
      signal_processing_end_.emit ();
    }

    void
    Library::query (const DB::AttributeList &attrs,
                    DB::VRow            &vector)
    {
      Glib::ustring suffix;

      suffix.append (" ORDER BY ");
      suffix.append (track_attributes[DATUM_ARTIST].id);

      suffix.append (", "); 
      suffix.append (track_attributes[DATUM_ALBUM].id);

      suffix.append (", ");
      suffix.append (track_attributes[DATUM_TRACK].id);

      suffix.append (", ");
      suffix.append (track_attributes[DATUM_TITLE].id);

      db->get ("main", attrs, vector, suffix);
    }

    bool
    Library::is_compilation           (const Glib::ustring& album)
    {
      return (compilations.find (album) != compilations.end());
    }

    bool
    Library::is_library_track         (const Glib::ustring &uri)
    {
      return (db->attr_exists ("main", track_attributes[DATUM_LOCATION].id, uri)); 
    }

    void
    Library::set_asin_for_album       (const Glib::ustring& artist, const Glib::ustring& album, const Glib::ustring& asin)
    {
      char *query = sqlite3_mprintf ("UPDATE main SET asin='%q' WHERE artist='%q' AND album='%q';",
                                     asin.c_str(), artist.c_str(), album.c_str()); 
      db->sqlite_exec_simple (query);
      free (query);
    }

    bool Library::is_new
         (const Glib::ustring& artist, const Glib::ustring& album)
    {
      using namespace Bmp::DB;

      AttributeList attrs;
      attrs.push_back (Attribute (DB::EXACT, track_attributes[DATUM_ARTIST].id, artist));
      attrs.push_back (Attribute (DB::EXACT, track_attributes[DATUM_ALBUM].id,  album));

      VRow vector;
      project (track_attributes[DATUM_NEW_ITEM].id, vector, attrs);
      
      if (vector.empty ()) throw Exception ("No new_item column (This error is fatal and should never occur)", Exception::UNKNOWN, std::string("")); 
      Row row = vector[0];
      Row::iterator i = row.find (track_attributes[DATUM_NEW_ITEM].id);
      return boost::get<bool>(i->second); 
    }

    Glib::ustring Library::get_sort_name_for_album
                  (Glib::ustring const& artist,
                   Glib::ustring const& album)
    {
      using namespace Bmp::DB;
      using namespace Glib; 

      DB::AttributeList attrs;
      attrs.push_back (DB::Attribute (DB::EXACT, track_attributes[DATUM_ARTIST].id, artist));
      attrs.push_back (DB::Attribute (DB::EXACT, track_attributes[DATUM_ALBUM].id,  album));

      VRow vector;
      db->get ("sort_names", attrs, vector);
      if (vector.empty ()) throw Exception ("No Artist for this sort_name", Exception::NO_RELEASE, std::string()); 
         
      Row row = vector[0];
      return boost::get<ustring>(row[track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id]);

    }

#if 0
    Glib::ustring Library::get_name_for_sort_name
                  (Glib::ustring const& name)
    {
      using namespace Bmp::DB;
      using namespace Glib; 

      DB::AttributeList attrs;
      attrs.push_back (DB::Attribute (DB::EXACT, track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id, name));

      VRow vector;
      db->get ("sort_names", attrs, vector);
      if (vector.empty ()) throw Exception ("No Sortname for this artist", Exception::NO_RELEASE, std::string(name)); 
         
      Row row = vector[0];
      return boost::get<ustring>(row[track_attributes[DATUM_ARTIST].id]);
    }
#endif

    Glib::ustring Library::get_release_for_album
                  (const Glib::ustring& artist, 
                   const Glib::ustring& album) 
    {
      using namespace Bmp::DB;
      using namespace Glib; 

      DB::AttributeList attrs;
      attrs.push_back (DB::Attribute (DB::EXACT, track_attributes[DATUM_ARTIST].id, artist));
      attrs.push_back (DB::Attribute (DB::EXACT, track_attributes[DATUM_ALBUM].id,  album));

      VRow vector;
      db->get ("mbreleases", attrs, vector);
      if (vector.empty ()) throw Exception ("No MB release id for this album", Exception::NO_RELEASE, std::string(album)); 
         
      Row row = vector[0];
      return boost::get<ustring>(row[track_attributes[DATUM_MB_ALBUM_ID].id]);
    }

    void Library::get_albums
         (DB::VRow &vector, bool new_albums) 
    {
      using namespace Bmp::DB;
      using namespace Glib; 

      DB::AttributeList attrs;
      attrs.push_back (DB::Attribute (DB::EXACT, track_attributes[DATUM_NEW_ITEM].id, bool(new_albums)));
      db->get ("albums", attrs, vector);
    }

    Glib::ustring Library::get_asin_for_album 
                  (Glib::ustring const& artist,
                   Glib::ustring const& album)    
    {
      using namespace Bmp::DB;
      using namespace Glib; 

      DB::AttributeList attrs;
      attrs.push_back (DB::Attribute (DB::EXACT, track_attributes[DATUM_ARTIST].id, artist));
      attrs.push_back (DB::Attribute (DB::EXACT, track_attributes[DATUM_ALBUM].id,  album));

      VRow vector;
      db->get ("asin", attrs, vector);
      if (vector.empty ()) throw Exception ("No ASIN for this album", Exception::NO_ASIN, std::string(album)); 
         
      Row row = vector[0];
      return boost::get<ustring>(row[track_attributes[DATUM_ASIN].id]);
    }

    /** The caller must be aware that each row in this vector contains ONLY the requrested column, re-marshaling/transforming
      * the result into something more suitable, i.e. an std::vector<> of std::string would be just too much overhead
      */
    void
    Library::project (std::string const&        project,
                      DB::VRow&             vector,
                      const DB::AttributeList&  attrs)
    {
      db->project ("main", project, vector, attrs);
    }

    GHashTable*
    Library::get_hashtable (const DB::Row& row)
    {
      using namespace Glib;
      using namespace std;

      GHashTable *table;
      GValue     *value;

      table = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);

      for (unsigned int n = 0; n < G_N_ELEMENTS (track_attributes); n++)
      {
        DB::Row::const_iterator i = row.find (track_attributes[n].id);
        if (i == row.end())
          {
            g_warning (G_STRLOC ": DB::Attribute '%s' not found in Row", track_attributes[n].id);
            continue;
          }

        value = g_new0 (GValue,1);
        switch (Bmp::DB::ValueType(i->second.which()))
          {
          case Bmp::DB::VALUE_TYPE_STRING:
              g_value_init (value, G_TYPE_STRING);
              g_value_set_string (value, (boost::get<ustring>(i->second)).c_str());
              break;

          case Bmp::DB::VALUE_TYPE_INT:
              g_value_init (value, G_TYPE_INT);
              g_value_set_int (value, (boost::get<int>(i->second)));
              break;

          case Bmp::DB::VALUE_TYPE_BOOL:
              g_value_init (value, G_TYPE_BOOLEAN);
              g_value_set_boolean (value, (boost::get<bool>(i->second)));
              break;

          case Bmp::DB::VALUE_TYPE_REAL:
              g_value_init (value, G_TYPE_DOUBLE);
              g_value_set_double (value, (boost::get<double>(i->second)));
              break;
        }
        g_hash_table_insert (table, g_strdup(track_attributes[n].id), value);
      }
      return table;
    }

    void
    Library::get_metadata (const Glib::ustring& uri,
                           DB::Row&             row)
    {
      using namespace Bmp::DB;
      using namespace std;
      using namespace Glib;

      if ( (uri.substr (0, 5) == "http:") || (uri.substr (0, 3) == "mms") )
      {
        row.insert (std::make_pair (track_attributes[DATUM_LOCATION].id, uri)); 
        row.insert (std::make_pair (track_attributes[DATUM_SAMPLERATE].id, int(0))); 
        row.insert (std::make_pair (track_attributes[DATUM_BITRATE].id, int(0)));
        row.insert (std::make_pair (track_attributes[DATUM_TITLE].id, uri));
        return;
      }

      AttributeList attrs;
      VRow vector;

      debug ("library", "Querying for metadata for URI %s", uri.c_str());

      attrs.push_back (Attribute (EXACT, track_attributes[DATUM_LOCATION].id, uri)); 
      query (attrs, vector);

      if (vector.empty())
        {
          try {
              metadata_get_taglib (uri, row);
            }
          catch (Bmp::Library::Exception& cxe)
            {
              if (cxe.id() != Exception::FILE_INVALID)
                {
                  metadata_get_gst (uri, row);
                }
            }

          if (row.empty())
            {
              row.insert (make_pair (track_attributes[DATUM_LOCATION].id, uri)); 
              row.insert (make_pair (track_attributes[DATUM_SAMPLERATE].id, int(0))); 
              row.insert (make_pair (track_attributes[DATUM_BITRATE].id, int(0)));
              row.insert (make_pair (track_attributes[DATUM_TITLE].id, path_get_basename (filename_from_uri(uri))));
            }
        }
      else
        {
          row = vector[0]; 

#ifdef HAVE_HAL
          if (hal->is_initialized())
            {
              // Let's see if the file was moved by checking with HAL
              Bmp::URI u (boost::get<ustring>(row[track_attributes[DATUM_LOCATION].id]));
              if ((u.get_protocol () == Bmp::URI::PROTOCOL_FILE))
                {
                  string filename = filename_from_uri (boost::get<ustring>(row[track_attributes[DATUM_LOCATION].id]));

                  if ((file_test (filename, FILE_TEST_EXISTS))) return;

                  try {
                    string  mount_path           = hal->get_mount_path_for_volume_udi (boost::get<ustring>(row[track_attributes[DATUM_HAL_VOLUME_UDI].id]));
                    ustring volume_relative_path = boost::get<ustring>(row[track_attributes[DATUM_VOLUME_RELATIVE_PATH].id]);

                    ustring new_uri;
                    new_uri .append ("file://")
                            .append (mount_path)
                            .append (volume_relative_path);

                    if (file_test (filename_from_uri (new_uri), FILE_TEST_EXISTS))
                      {
                          char *query = sqlite3_mprintf ("UPDATE main SET location='%q' WHERE location='%q';", new_uri.c_str(),
                                                              boost::get<ustring>(row[track_attributes[DATUM_LOCATION].id]).c_str());
                          db->sqlite_exec_simple (query);
                          free (query);
                          row[track_attributes[DATUM_LOCATION].id] = new_uri; 
                      }
                    }
                  catch (...) { return ; }
                }
            }
#endif
        } 
    }
    


    std::string const
    Library::get_metadatum_id (Bmp::Library::Datum metadatum) const
    {
        return std::string (track_attributes[metadatum].id);
    }



    const DatumDefine 
    Library::get_metadatum_define (Bmp::Library::Datum metadatum) const
    {
        return track_attributes[metadatum];
    }


    void
    Library::drop_items ()
    {
      char *query = sqlite3_mprintf ("DELETE FROM main WHERE new_item > 0;"); 
      db->sqlite_exec_simple (query);
      free (query);
    }


    void
    Library::approve_items ()
    {
      char *query = sqlite3_mprintf ("UPDATE main SET new_item=0;"); 
      db->sqlite_exec_simple (query);
      free (query);
    }


    void
    Library::approve_items (const Glib::ustring& artist, const Glib::ustring& album, bool sort_name)
    {
      char *query = 0;
      if (sort_name)
        {
          query = sqlite3_mprintf ("UPDATE main SET new_item=0 WHERE %s='%q' AND %s='%q';",
                                          track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id, artist.c_str(),
                                          track_attributes[DATUM_ALBUM].id, album.c_str()); 
        }

      else
        {
          query = sqlite3_mprintf ("UPDATE main SET new_item=0 WHERE %s='%q' AND %s='%q';",
                                          track_attributes[DATUM_ARTIST].id, artist.c_str(),
                                          track_attributes[DATUM_ALBUM].id, album.c_str()); 
        }
      db->sqlite_exec_simple (query);
      free (query);
    }


    void
    Library::increase_play_count (const Glib::ustring& uri)
    {
      using namespace DB;

      char *sql = sqlite3_mprintf ("UPDATE main SET %s='%d', %s = %s +1 WHERE %s='%q';",
                                   track_attributes[DATUM_PLAYDATE].id,
                                   time (NULL),
                                   track_attributes[DATUM_COUNT].id,
                                   track_attributes[DATUM_COUNT].id,
                                   track_attributes[DATUM_LOCATION].id,
                                   uri.c_str());
      db->sqlite_exec_simple (sql);
      free (sql);

      DB::AttributeList attrs; 
      VRow v;
      attrs.push_back (DB::Attribute (DB::EXACT, track_attributes[DATUM_LOCATION].id, uri)); 
      db->get ("main", attrs, v);
      if (!v.empty())
        {
          Track track (v[0]);
          UpdateTrack update_track;
          update_track = track; 
          signal_row_updated_.emit (update_track.location_cur.get(), update_track);   
        }
    }

#if 0
    void
    Library::increase_pair_count   (const Glib::ustring    &uri1,
                                    const Glib::ustring    &uri2)
    {
        char *sql = sqlite3_mprintf ("SELECT playcount FROM transitions WHERE track_from='%q' AND "
                                     "track_to='%q';", uri1.c_str(), uri2.c_str());
        unsigned int nrow = db->sqlite_exec_simple (sql);
        free (sql);

        if (nrow != 0)
          {
            char *sql = sqlite3_mprintf ("UPDATE transitions SET playcount = playcount + 1 WHERE track_from='%q' AND "
                         "track_to='%q';", uri1.c_str(), uri2.c_str());
            db->sqlite_exec_simple (sql);
            free (sql);
          }
        else
          {
            char *sql = sqlite3_mprintf ("INSERT INTO transitions (track_from, track_to, playcount) VALUES ("
                         "'%q', '%q', 1);", uri1.c_str(), uri2.c_str());
            db->sqlite_exec_simple (sql);
            free (sql);
          }
    }
#endif

#if 0
    Glib::ustring
    Library::get_autoplay_track (const Glib::ustring  &pkey)
    {
      using namespace Bmp::DB;

      DB::AttributeList attrs;
      attrs.push_back (DB::Attribute (DB::EXACT, "track_from", pkey));

      VRow vector;
      db->get ("transitions", attrs, vector, std::string(" ORDER BY playcount DESC"));

      if (vector.empty ())
        {
          throw Exception ("No Autoplay Track", Exception::NO_AUTOPLAY_TRACK, pkey); 
        }

      Row row = vector[0];
      return boost::get<Glib::ustring>(row["track_to"]);
    }
#endif


    void
    Library::add_audiocd (std::string const& cdindex)
    {
        using namespace Bmp::DB;
        using namespace std; 

        int status;
        VRow rows;

        debug ("library", "Calling get_audiocd_metadata");

        get_audiocd_metadata (status, rows);

        if (status == Bmp::Library::Library::MB_FAILURE)
          {
            debug ("library", "get_audiocd_metadata() status is MB_FAILURE");
            throw Exception ("Unable to acquire metadata for AudioCD", Exception::NO_METADATA, cdindex);
          }

        AttributeList attrs;
        VRow vector;
        attrs.push_back (Attribute (DB::EXACT, track_attributes[DATUM_LOCATION].id, Glib::ustring((boost::format ("cdda://%s/1") % cdindex).str())));
      
        query (attrs, vector);

        if (vector.empty())
        {
          debug ("library", "CD Seems to be not cached yet");
          for (VRow::const_iterator iter = rows.begin (); iter != rows.end (); ++iter)
          {
            db->add ("main", (*iter));
          }
        }
    }

    void
    Library::recreate_db ()
    {
        using namespace Bmp::DB;
        using namespace std;
        using namespace Glib;

        delete db;
        db = new Database ("library", BMP_PATH_DATA_DIR, true);

        ValueMap meta;
        meta.insert (make_pair ("version", VALUE_TYPE_STRING));
        db->create_table ("meta", "version", meta);

        ustring sql;

        sql .append ("INSERT INTO meta (version) VALUES ('")
            .append (string (MLIB_VERSION))
            .append ("');");

        db->sqlite_exec_simple (sql.c_str());
    }


    Library::Library ()
    {
      using namespace Bmp::DB;
      using namespace std;
      using namespace Glib;
  
      register_taglib_plugins (); 

      for (unsigned int n = 0; n < G_N_ELEMENTS(track_attributes); n++)
        {
          map.insert (std::make_pair (std::string (track_attributes[n].id), track_attributes[n].type));
        }

      db = new Database ("library", BMP_PATH_DATA_DIR, false);

      ValueMap meta;
      meta.insert (make_pair ("version", VALUE_TYPE_STRING));
      db->create_table  ("meta", "version", meta);

      AttributeList attrs;
      attrs.push_back (DB::Attribute (DB::EXACT, std::string("version"), std::string(MLIB_VERSION))); 

      VRow vrows;

      try {
          db->get ("meta", attrs, vrows);
          if (vrows.empty())
          { 
            recreate_db ();
          }
          else
          {
            if (boost::get<ustring>((*(vrows[0].find ("version"))).second) != ustring (MLIB_VERSION)) recreate_db ();
          }
        }
      catch (Bmp::DB::Exception& cxe) { recreate_db (); }

      db->create_table  ("main", string (track_attributes[DATUM_LOCATION].id), map);

      //Create compilations view;
      stringstream sql;
      sql << "SELECT DISTINCT " << track_attributes[DATUM_ALBUM].id 
          << ", " 
          << "count(DISTINCT " << track_attributes[DATUM_ARTIST].id << ") AS nr FROM main GROUP BY "
          << track_attributes[DATUM_ALBUM].id << " HAVING nr > 1"; 
      map_compilations.insert (make_pair (string (track_attributes[DATUM_ALBUM].id), track_attributes[DATUM_ALBUM].type));
      map_compilations.insert (make_pair (string ("nr"), DB::VALUE_TYPE_INT));
      db->create_view_manual ("main", "compilations", sql.str(), map_compilations);

      //Create asin view;
      sql.str("");
      sql << "SELECT DISTINCT " 
          << track_attributes[DATUM_ARTIST].id 
          << ", " 
          << track_attributes[DATUM_ALBUM].id 
          << ", " 
          << track_attributes[DATUM_ASIN].id 
          << " FROM main WHERE length(" 
          << track_attributes[DATUM_ASIN].id 
          << ") AND " 
          << track_attributes[DATUM_ASIN].id 
          << " IS NOT NULL GROUP BY "
          << track_attributes[DATUM_ALBUM].id << ";";

      map_asin.insert (make_pair (string (track_attributes[DATUM_ARTIST].id),
                                  track_attributes[DATUM_ARTIST].type));

      map_asin.insert (make_pair (string (track_attributes[DATUM_ALBUM].id),
                                  track_attributes[DATUM_ALBUM].type));

      map_asin.insert (make_pair (string (track_attributes[DATUM_ASIN].id),
                                  track_attributes[DATUM_ASIN].type));
      db->create_view_manual ("main", "asin", sql.str(), map_asin);

      //Create mbreleases view;
      sql.str("");
      sql << "SELECT DISTINCT " 
          << track_attributes[DATUM_ARTIST].id 
          << ", " 
          << track_attributes[DATUM_ALBUM].id 
          << ", " 
          << track_attributes[DATUM_MB_ALBUM_ID].id 
          << " FROM main WHERE length(" 
          << track_attributes[DATUM_MB_ALBUM_ID].id 
          << ") AND " 
          << track_attributes[DATUM_MB_ALBUM_ID].id 
          << " IS NOT NULL GROUP BY "
          << track_attributes[DATUM_ALBUM].id << ";";

      map_releases.insert (make_pair (string (track_attributes[DATUM_ARTIST].id), track_attributes[DATUM_ARTIST].type));
      map_releases.insert (make_pair (string (track_attributes[DATUM_ALBUM].id), track_attributes[DATUM_ALBUM].type));
      map_releases.insert (make_pair (string (track_attributes[DATUM_MB_ALBUM_ID].id), track_attributes[DATUM_MB_ALBUM_ID].type));
      db->create_view_manual ("main", "mbreleases", sql.str(), map_releases);

      char *sqlstr = 0;
      //Create albums view
      sqlstr = sqlite3_mprintf ("SELECT DISTINCT %q, %q, %q, %q FROM main WHERE %q LIKE 'file://%%' ORDER BY %q;",
                                      track_attributes[DATUM_ARTIST].id,
                                      track_attributes[DATUM_ALBUM].id,
                                      track_attributes[DATUM_ASIN].id,
                                      track_attributes[DATUM_NEW_ITEM].id,
                                      track_attributes[DATUM_LOCATION].id,
                                      track_attributes[DATUM_ARTIST].id);

      map_albums.insert (make_pair (string (track_attributes[DATUM_ARTIST].id),
                                            track_attributes[DATUM_ARTIST].type));

      map_albums.insert (make_pair (string (track_attributes[DATUM_ALBUM].id),
                                            track_attributes[DATUM_ALBUM].type));

      map_albums.insert (make_pair (string (track_attributes[DATUM_ASIN].id),
                                            track_attributes[DATUM_ASIN].type));

      map_albums.insert (make_pair (string (track_attributes[DATUM_NEW_ITEM].id),
                                            track_attributes[DATUM_NEW_ITEM].type));

      db->create_view_manual ("main", "albums", sqlstr, map_albums);
      free (sqlstr);

      //Create artist sort_names view #FIXME: This has to be solved differently persistently
      sqlstr = sqlite3_mprintf ("SELECT DISTINCT %q, %q, %q FROM main WHERE %q NOT NULL AND length(%q) > 0 ORDER BY %q;",
                                      track_attributes[DATUM_ARTIST].id,
                                      track_attributes[DATUM_ALBUM].id,
                                      track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id,
                                      track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id,
                                      track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id,
                                      track_attributes[DATUM_ARTIST].id);

      map_sort_names.insert (make_pair (string (track_attributes[DATUM_ARTIST].id),
                                            track_attributes[DATUM_ARTIST].type));

      map_sort_names.insert (make_pair (string (track_attributes[DATUM_ALBUM].id),
                                            track_attributes[DATUM_ALBUM].type));

      map_sort_names.insert (make_pair (string (track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id),
                                            track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].type));

      db->create_view_manual ("main", "sort_names", sqlstr, map_sort_names);
      free (sqlstr);


#if 0
      //Create pairs table;
      map_transitions.insert (make_pair (string ("track_from"), Bmp::DB::VALUE_TYPE_STRING)); 
      map_transitions.insert (make_pair (string ("track_to"),   Bmp::DB::VALUE_TYPE_STRING)); 
      map_transitions.insert (make_pair (string ("playcount"),  Bmp::DB::VALUE_TYPE_INT)); 
      try
        {
          db->sqlite_exec_simple ("CREATE TABLE IF NOT EXISTS transitions ('track_from' TEXT, 'track_to' TEXT, 'playcount' INTEGER, PRIMARY KEY ('track_from', 'track_to') ON CONFLICT REPLACE);");
        }
      catch (...) {}
      db->insert_map_for_table ("transitions", map_transitions);
#endif

      update_compilations ();
    }

    void
    Library::update_compilations ()
    {
      using namespace DB;
      using namespace std;

      VRow rows;
      db->project ("compilations", track_attributes[DATUM_ALBUM].id, rows);
      compilations.clear ();
      for (VRow::iterator i = rows.begin (), end = rows.end (); i != end; ++i) 
      {
          compilations.insert (boost::get<Glib::ustring>((*i)[track_attributes[DATUM_ALBUM].id]));
      }
    }


    bool
    Library::modify_tracks (const UpdateTracks& tracks, bool retag, bool db_update)
    {
      signal_processing_start_.emit ();
      signal_modify_start_.emit (tracks.size ());

      for (UpdateTracks::const_iterator i_Tracks      = tracks.begin (),
                                        i_TracksEnd   = tracks.end ();
                                        i_Tracks     != i_TracksEnd; ++i_Tracks)
        {
          using namespace Bmp::DB;
          using namespace Glib; 
          using namespace std; 

          UpdateTrack const& track = (*i_Tracks);
          DB::AttributeListSet attrs;

          if (db_update)
            {
              if (track.title)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_TITLE].id,
                                                   track.title.get().raw())); 
              if (track.artist)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_ARTIST].id,
                                                   track.artist.get().raw())); 
              if (track.album)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_ALBUM].id,
                                                   track.album.get().raw())); 
              if (track.genre)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_GENRE].id,
                                                   track.genre.get().raw())); 
              if (track.comment)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_COMMENT].id,
                                                   track.comment.get().raw())); 
              if (track.tracknumber)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_TRACK].id,
                                                   int(track.tracknumber.get()))); 
              if (track.date)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_DATE].id,
                                                   int(track.date.get()))); 
              if (track.rating)
                 attrs.push_back (DB::AttributeSet (track_attributes[DATUM_RATING].id,
                                                    int(track.rating.get()))); 
              if (track.count)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_COUNT].id,
                                                   int(track.count.get()))); 
              if (track.asin)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_ASIN].id,
                                                   track.asin.get().raw())); 
              if (track.mb_album_artist)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_MB_ALBUM_ARTIST].id,
                                                   track.mb_album_artist.get().raw())); 
              if (track.mb_album_artist_id)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_MB_ALBUM_ARTIST_ID].id,
                                                   track.mb_album_artist_id.get().raw())); 
              if (track.mb_album_artist_sort_name)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id,
                                                   track.mb_album_artist_sort_name.get().raw())); 
              if (track.mb_track_id)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_MB_TRACK_ID].id,
                                                   track.mb_track_id.get().raw())); 
              if (track.mb_album_id)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_MB_ALBUM_ID].id,
                                                   track.mb_album_id.get().raw())); 
              if (track.mb_artist_id)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_MB_ARTIST_ID].id,
                                                   track.mb_artist_id.get().raw())); 
              if (track.mb_artist_sort_name)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_MB_ARTIST_SORTNAME].id,
                                                   track.mb_artist_sort_name.get().raw())); 
              if (track.new_item)
                attrs.push_back (DB::AttributeSet (track_attributes[DATUM_NEW_ITEM].id,
                                                   bool(track.new_item.get())));
            }
      
          ustring modify_location = track.location_cur.get();

#ifdef HAVE_HAL
          if ((track.location_new) && hal->is_initialized())
          {
            HALVolume volume = hal->get_volume_for_uri (track.location_new.get());
            string filename1 = filename_from_uri (track.location_cur.get());
            string filename2 = filename_from_uri (track.location_new.get());
            try
              {
                Util::copy_file (filename1, filename2);
                if (Util::compare_files (filename1, filename2))
                {
                  ustring volume_relative_path = filename2.substr (volume.mount_path.length());

                  g_unlink (filename1.c_str());

                  attrs.push_back (DB::AttributeSet (track_attributes[DATUM_VOLUME_RELATIVE_PATH].id, volume_relative_path));
                  attrs.push_back (DB::AttributeSet (track_attributes[DATUM_HAL_VOLUME_UDI].id, volume.volume_udi));
                  attrs.push_back (DB::AttributeSet (track_attributes[DATUM_HAL_DEVICE_UDI].id, volume.device_udi));
                  attrs.push_back (DB::AttributeSet (track_attributes[DATUM_LOCATION].id, filename_to_uri(filename2))); 

                  modify_location = filename_to_uri (filename2);
                }
              }
            catch (std::exception& cxe)
              {
                g_critical ("%s: Unable to copy file to destination. Reason: '%s'. Source: '%s'. Destination: '%s'",
                             G_STRLOC, cxe.what(), filename1.c_str(), filename2.c_str());
                continue;
              }
          }
#else
          if (track.location_new) 
          {
            string filename1 = filename_from_uri (track.location_cur.get());
            string filename2 = filename_from_uri (track.location_new.get());
            try
              {
                Util::copy_file (filename1, filename2);
                if (Util::compare_files (filename1, filename2))
                {
                  g_unlink (filename1.c_str());
                  attrs.push_back (DB::AttributeSet (track_attributes[DATUM_LOCATION].id, Glib::filename_to_uri(filename2))); 
                  modify_location = Glib::filename_to_uri(filename2);
                }
              else
                {
                  g_unlink (filename2.c_str());
                }
              }
            catch (std::exception& cxe)
              {
                g_critical ("%s: Unable to copy file to destination. Reason: '%s'. Source: '%s'. Destination: '%s'",
                            G_STRLOC, cxe.what(), filename1.c_str(), filename2.c_str());
                continue;
              }
          }
#endif

          try {
              if (retag)
              {
                metadata_set_taglib (modify_location, track);
              }
              db->set ("main", track_attributes[DATUM_LOCATION].id, track.location_cur.get(), attrs);
              signal_row_updated_.emit (track.location_cur.get(), track);   
            }
          catch (Bmp::Library::Exception& cxe) {}
          catch (Bmp::DB::Exception& cxe) {}
          
          signal_modify_step_.emit (); 
          while (g_main_context_pending (NULL)) g_main_context_iteration (NULL, TRUE);

      }   

      signal_modify_end_.emit ();
      signal_processing_end_.emit ();
      return true;
    }


    Library::~Library ()
    {
        delete db;
    }

    // Generic "processing" progress #FIXME: What is this good for?  --deadchip
    Library::SignalProcessing&
    Library::signal_processing_start ()
    { 
      return signal_processing_start_;
    }

    Library::SignalProcessing&
    Library::signal_processing_end ()
    { 
      return signal_processing_end_;
    }

    Library::SignalRowUpdated&
    Library::signal_row_updated ()
    {
      return signal_row_updated_;
    }

    // Modification progress
    Library::ModifyStart&
    Library::signal_modify_start()
    { 
      return signal_modify_start_;
    }

    Library::ModifyStep&
    Library::signal_modify_step()
    { 
      return signal_modify_step_;
    }

    Library::ModifyEnd&
    Library::signal_modify_end()
    {
      return signal_modify_end_;
    }

    Album::Album (const Bmp::DB::Row& row)
    {
      using namespace Bmp::DB;
      using namespace Glib; 

      Row::const_iterator i;

      i = row.find (track_attributes[DATUM_ARTIST].id);
      if (!(boost::get<ustring>(i->second).empty()))
        {
          artist = boost::get<ustring> (i->second);
        }

      i = row.find (track_attributes[DATUM_ALBUM].id);
      if (!(boost::get<ustring>(i->second).empty()))
        {
          album = boost::get<ustring> (i->second);
        }

      i = row.find (track_attributes[DATUM_ASIN].id);
      if (!(boost::get<ustring>(i->second).empty()))
        {
          asin = boost::get<ustring> (i->second);
        }

      i = row.find (track_attributes[DATUM_NEW_ITEM].id);
      new_item = boost::get<bool> (i->second);
    }

    // Track 
    Track::Track (const Bmp::DB::Row& row)
    {
      using namespace Bmp::DB;
      using namespace Glib; 

      Row::const_iterator i, end = row.end ();

      i = row.find (track_attributes[DATUM_LOCATION].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          location = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_ARTIST].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          artist = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_ALBUM].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          album = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_TITLE].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          title = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_GENRE].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          genre = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_COMMENT].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          comment = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_TRACK].id);
      if (i != end)
            tracknumber = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_TIME].id);
      if (i != end)
            duration = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_RATING].id);
      if (i != end)
            rating = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_COUNT].id);
      if (i != end)
            count = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_DATE].id);
      if (i != end)
            date = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_MTIME].id);
      if (i != end)
            mtime = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_BITRATE].id);
      if (i != end)
            bitrate = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_SAMPLERATE].id);
      if (i != end)
            samplerate = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_COUNT].id);
      if (i != end)
            samplerate = boost::get<int> (i->second);

      i = row.find (track_attributes[DATUM_ASIN].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          asin = boost::get<ustring> (i->second);

#ifdef HAVE_HAL
      i = row.find (track_attributes[DATUM_HAL_VOLUME_UDI].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          volume_udi = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_HAL_DEVICE_UDI].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          device_udi = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_VOLUME_RELATIVE_PATH].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          vrp = boost::get<ustring> (i->second);
#endif

      i = row.find (track_attributes[DATUM_MB_ARTIST_SORTNAME].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          mb_album_artist_sort_name = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_MB_ALBUM_ARTIST_ID].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          mb_album_artist_id = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_MB_ALBUM_ARTIST_SORTNAME].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          mb_album_artist_sort_name = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_MB_TRACK_ID].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          mb_track_id = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_MB_ALBUM_ID].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          mb_album_id = boost::get<ustring> (i->second);

      i = row.find (track_attributes[DATUM_MB_ARTIST_ID].id);
      if ((i != end) && !(boost::get<ustring>(i->second).empty()))
          mb_artist_id = boost::get<ustring> (i->second);


      i = row.find (track_attributes[DATUM_NEW_ITEM].id);
      if (i != end)
        new_item = boost::get<bool> (i->second);

    }

  } // Library namespace

} // Bmp namespace
