#ifndef BMP_LIBRARY_CLASS_HPP
#define BMP_LIBRARY_CLASS_HPP

#include <set>
#include <gtkmm.h>
#include <glibmm/i18n.h>
#include <boost/optional.hpp>

#ifndef BMP_PLUGIN_BUILD
#  include <taglib/tag.h>
#endif

#include "database.hpp"
#include "util.hpp"

namespace Bmp
{
  class LibraryUI;

  namespace Library
  {
    class Exception : public std::exception
    {
      public:

        enum ExceptionID
        {
          NO_METADATA,
          NO_ASIN,
          NO_RELEASE,
          NO_FILE,
          NO_AUTOPLAY_TRACK,
          FILE_INVALID,
          FILE_NON_READABLE,
          FILE_NON_WRITABLE,
          CONVERSION_ERROR,
          UNABLE_TO_MODIFY,
          UNKNOWN,

          N_IDS
        }; 
        
        enum ExceptionItemType
        {
          ITEM_ITEM,
          ITEM_URI,
        };

        Exception (std::string const& what, ExceptionID id, std::string const& item)
            : what_ (what), id_ (id), item_ (item), type_ (ITEM_ITEM)
        {
//          g_log (BMP_LIBRARY_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "what (\"%s\") id (\"%s\"), item (\"%s\")", what.c_str(), _message_types[id], item_.c_str());
        }

        Exception (std::string const& what, ExceptionID id, const Glib::ustring& uri)
            : what_ (what), id_ (id), uri_ (uri), type_ (ITEM_URI)
        {
//          g_log (BMP_LIBRARY_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "what (\"%s\") id (\"%s\"), uri (\"%s\")", what.c_str(), _message_types[id], uri_.c_str());
        }

        virtual ~Exception () throw()
        {}

        virtual const char *
        what () const throw()
        {
          return what_.c_str();
        }

        const char * item () const
        {
          if (type_ == ITEM_ITEM)
            return item_.c_str();
          else
            return uri_.c_str();
        }

        ExceptionID id () const
        {
          return id_;
        }

        ExceptionItemType type () const
        {
          return type_;
        }

      private:
        
        static char* const _message_types[];

        std::string             what_;
        ExceptionID             id_;
        std::string             item_;
        Glib::ustring           uri_;
        const ExceptionItemType type_;
    };
        
    enum Datum
    {
        DATUM_LOCATION,
        DATUM_ARTIST,
        DATUM_ALBUM,
        DATUM_TITLE,
        DATUM_TRACK,
        DATUM_TIME,
        DATUM_GENRE,
        DATUM_COMMENT,
        DATUM_RATING,
        DATUM_DATE,
        DATUM_MTIME,
        DATUM_BITRATE,
        DATUM_SAMPLERATE,
        DATUM_COUNT,
        DATUM_PLAYDATE,
        DATUM_ASIN,

#ifdef HAVE_HAL
        DATUM_HAL_VOLUME_UDI,
        DATUM_HAL_DEVICE_UDI,
        DATUM_VOLUME_RELATIVE_PATH,       // Path of the file relative to the volume's mount path
#endif

        DATUM_MB_ALBUM_ARTIST,
        DATUM_MB_ALBUM_ARTIST_ID,
        DATUM_MB_ALBUM_ARTIST_SORTNAME,
        DATUM_MB_TRACK_ID,
        DATUM_MB_ALBUM_ID,
        DATUM_MB_ARTIST_ID,
        DATUM_MB_ARTIST_SORTNAME,

        DATUM_NEW_ITEM,

        DATUM_LAST
    };

    struct DatumDefine
    {
          const char    *title;
          const char    *title_plural;
          const char    *description;
          const char    *id;
          DB::ValueType  type;
          bool           filter;
    };

    struct Album
    {
      boost::optional <Glib::ustring>   artist;
      boost::optional <Glib::ustring>   album;
      boost::optional <Glib::ustring>   asin;
      boost::optional <bool>            new_item;

      Album (const Bmp::DB::Row& row);
      Album () {};
      ~Album () {};
    };
 
    struct Track
    {
      boost::optional <Glib::ustring>   location;
      boost::optional <Glib::ustring>   artist;
      boost::optional <Glib::ustring>   album;
      boost::optional <Glib::ustring>   title;
      boost::optional <Glib::ustring>   genre;
      boost::optional <Glib::ustring>   comment;

#ifdef HAVE_HAL
      boost::optional <Glib::ustring>   volume_udi;
      boost::optional <Glib::ustring>   device_udi;
      boost::optional <Glib::ustring>   vrp;
#endif

      boost::optional <Glib::ustring>   asin;

      boost::optional <Glib::ustring>   mb_album_artist;
      boost::optional <Glib::ustring>   mb_album_artist_id;
      boost::optional <Glib::ustring>   mb_album_artist_sort_name;
      boost::optional <Glib::ustring>   mb_track_id;
      boost::optional <Glib::ustring>   mb_album_id;
      boost::optional <Glib::ustring>   mb_artist_id;
      boost::optional <Glib::ustring>   mb_artist_sort_name;

      boost::optional <int>     tracknumber;
      boost::optional <int>     duration;
      boost::optional <int>     rating;
      boost::optional <int>     date;
      boost::optional <int>     mtime;
      boost::optional <int>     bitrate;
      boost::optional <int>     samplerate;
      boost::optional <int>     count;
      boost::optional <int>     playdate;
      boost::optional <bool>    new_item;

      Track (const Bmp::DB::Row& row);
      Track () {};
      ~Track () {};
    };

    struct UpdateTrack
    {
      boost::optional <Glib::ustring>   location_cur;
      boost::optional <Glib::ustring>   location_new;
      boost::optional <Glib::ustring>   artist;
      boost::optional <Glib::ustring>   album;
      boost::optional <Glib::ustring>   title;
      boost::optional <Glib::ustring>   genre;
      boost::optional <Glib::ustring>   comment;
      boost::optional <Glib::ustring>   asin;

      boost::optional <Glib::ustring>   mb_album_artist;
      boost::optional <Glib::ustring>   mb_album_artist_id;
      boost::optional <Glib::ustring>   mb_album_artist_sort_name;
      boost::optional <Glib::ustring>   mb_track_id;
      boost::optional <Glib::ustring>   mb_album_id;
      boost::optional <Glib::ustring>   mb_artist_id;
      boost::optional <Glib::ustring>   mb_artist_sort_name;

      boost::optional <int>     tracknumber;
      boost::optional <int>     date;
      boost::optional <int>     rating;
      boost::optional <int>     mtime;
      boost::optional <int>     count;
      boost::optional <int>     playdate;
      boost::optional <bool>    new_item;

      UpdateTrack& operator=(const Track& track)
      {
        location_cur              = track.location;
        artist                    = track.artist;
        album                     = track.album;
        title                     = track.title;
        genre                     = track.genre;
        comment                   = track.comment;

        asin                      = track.asin;

        mb_album_artist           = track.mb_album_artist;
        mb_album_artist_id        = track.mb_album_artist_id;
        mb_album_artist_sort_name  = track.mb_album_artist_sort_name;
        mb_track_id               = track.mb_track_id;
        mb_album_id               = track.mb_album_id;
        mb_artist_id              = track.mb_artist_id;
        mb_artist_sort_name        = track.mb_artist_sort_name;

        tracknumber               = track.tracknumber;
        date                      = track.date;
        rating                    = track.rating;
        mtime                     = track.mtime;
        playdate                  = track.playdate;
        new_item                  = track.new_item;
      
        return *this;
      }
    };

    typedef std::vector < UpdateTrack > UpdateTracks;

    class Library
      {
        public:

          typedef sigc::signal <void, int> ModifyStart;
          typedef sigc::signal <void>      ModifyStep;
          typedef sigc::signal <void>      ModifyEnd;
          typedef sigc::signal <void>      SignalProcessing;
          typedef sigc::signal <void, Glib::ustring, Bmp::Library::UpdateTrack > SignalRowUpdated; // returns: void, args: pkey, datarow

          SignalProcessing& signal_processing_start ();
          SignalProcessing& signal_processing_end   ();
          SignalRowUpdated& signal_row_updated      ();

          ModifyStart& signal_modify_start ();
          ModifyStep&  signal_modify_step  ();
          ModifyEnd&   signal_modify_end   ();

          enum
          {
              MB_FAILURE,
              MB_OK,
              MB_OFFLINE
          };

          typedef std::set < Glib::ustring > Compilations;

          Library ();
          ~Library ();

          void
          update_compilations ();
          void
          approve_items ();
          void
          approve_items (const Glib::ustring& artist, const Glib::ustring& album, bool sort_name = false);
          void
          drop_items (); 

          void
          increase_play_count   (const Glib::ustring    &uri);

#if 0
          void
          increase_pair_count   (const Glib::ustring    &uri1,
                                 const Glib::ustring    &uri2);
#endif

          bool
          is_compilation        (const Glib::ustring  &album);
          bool
          is_library_track      (const Glib::ustring &uri);
          bool
          is_new                (const Glib::ustring& artist, const Glib::ustring& album);

          // Add
          void
          add_track             (const Glib::ustring& uri);
          void
          add_audiocd           (std::string const& cdindex);

          // Delete
          void
          del                   (const DB::AttributeList  &attrs);

          // Modify
          void
          set_asin_for_album    (const Glib::ustring& artist,
                                 const Glib::ustring& album,
                                 const Glib::ustring& asin);
          bool
          modify_tracks         (const UpdateTracks& tracks, bool retag = true, bool db_update = true);

          // Get
#if 0
          Glib::ustring
          get_autoplay_track    (const Glib::ustring  &pkey);
#endif

          void    
          get_metadata          (const Glib::ustring& pkey,
                                 DB::Row&             row);
          void
          get_albums            (DB::VRow& vector, bool new_albums = false);
 
          GHashTable*
          get_hashtable         (const DB::Row&       row);

          std::string const
          get_metadatum_id      (Datum metadatum) const;

          const DatumDefine 
          get_metadatum_define  (Datum metadatum) const;

          Glib::ustring
          get_titlestring       (Glib::ustring const& uri,
                                 std::string const&   format);

          // Query
          Glib::ustring
          get_release_for_album  (Glib::ustring const& artist,
                                  Glib::ustring const& album);
          Glib::ustring
          get_asin_for_album     (Glib::ustring const& artist,
                                  Glib::ustring const& album);
          Glib::ustring
          get_sort_name_for_album (Glib::ustring const& artist,
                                  Glib::ustring const& album);

#if 0
          Glib::ustring
          get_name_for_sort_name (Glib::ustring const& name);
#endif

          void
          query                 (const DB::AttributeList& attrs,
                                 DB::VRow&                vector);
          void
          project               (std::string const&       project,
                                 DB::VRow&                vector,
                                 const DB::AttributeList& attrs = DB::AttributeList ());

    protected:

          friend class Bmp::LibraryUI;

#ifndef BMP_PLUGIN_BUILD
          void
          metadata_set_taglib_common  (const UpdateTrack& track, TagLib::Tag* tag);
#endif
          void 
          metadata_set_taglib         (const Glib::ustring &uri, const UpdateTrack& track);
          void 
          metadata_get_taglib         (const Glib::ustring &uri, DB::Row &row);
          void
          metadata_get_gst            (const Glib::ustring &uri, DB::Row &row);

          void
          recreate_db ();
          
    private:

          SignalProcessing signal_processing_start_;
          SignalProcessing signal_processing_end_;
          SignalRowUpdated signal_row_updated_;

          ModifyStart  signal_modify_start_;
          ModifyStep   signal_modify_step_;
          ModifyEnd    signal_modify_end_;

          DB::ValueMap  map;
          DB::Database *db;

          DB::ValueMap  map_compilations;
          DB::ValueMap  map_asin;
          DB::ValueMap  map_releases;
#if 0
          DB::ValueMap  map_transitions;
#endif
          DB::ValueMap  map_albums;
          DB::ValueMap  map_sort_names;

          Compilations  compilations;

    }; // class::Library

  } // ns::Library

} // ns::Bmp

#endif // BMP_LIBRARY_CLASS_HPP
