//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glibmm/i18n.h>
#include <glib/gstdio.h>
#include <gtkmm.h>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>
#include <fstream>

#include <boost/format.hpp>
#include <boost/optional.hpp>

#include <mcs/mcs.h>

#include "file_utils.hpp"
#include "ui_util.hpp"
#include "util.hpp"
#include "paths.hpp"
#include "main.hpp"
#include "debug.hpp"
#include "shell.hpp"
#include "stock.hpp"
#include "xds.hpp"
#include "uri++.hpp"

#include "x_library.hpp"
#include "x_system_control.hpp"
#include "x_vfs.hpp"

#ifdef HAVE_HAL
#  include "x_hal.hpp"
#endif

#include "musicbrainz_xml.hpp"
#include "library_ui.hpp"
#include "library_ui_modify.hpp"
#include "library_ui_modify_albums.hpp"
#include "library_ui_relocate_album.hpp"
#include "dialog_filelist.hpp"

#include "banner_image.hpp"

namespace
{
    using namespace Bmp::Library;

    typedef std::vector<Gtk::TreePath> PathList;
    typedef std::vector<Gtk::TreeRowReference> ReferenceList;
    typedef std::vector<Gtk::TreeModel::iterator> IterList;

    const Datum search_data[] =
    {
      DATUM_ARTIST,
      DATUM_ALBUM,
      DATUM_TITLE,
      DATUM_GENRE,
    };

    const int search_default[] =
    {
      0, 1, 2,
    };

    static boost::format progress_text ("%d / %d");
    int n_mod_tracks = 0, n_mod_current = 0;

    void
    on_modify_start (int n_tracks, Gtk::ProgressBar* bar)
    {
      n_mod_tracks = n_tracks;
      n_mod_current = 0;
      bar->show_all ();
      bar->set_fraction (0.0);
      bar->set_text ((progress_text % 0 % n_tracks).str());
    }

    void
    on_modify_step  (Gtk::ProgressBar *bar)
    {
      n_mod_current++;
      bar->set_fraction ((n_mod_current*1.)/(n_mod_tracks*1.));
      bar->set_text ((progress_text % n_mod_current % n_mod_tracks).str());
      while (gtk_events_pending()) gtk_main_iteration ();
    }

    void
    on_modify_end   (Gtk::ProgressBar *bar)
    {
      bar->set_fraction (0.0);
      bar->set_text ("");
      bar->hide ();
      n_mod_tracks = 0;
      n_mod_current = 0;
    }
}

namespace Bmp
{
    class TreeViewXDS : public Gtk::TreeView, public Bmp::XDS
    {
    public:
  
      TreeViewXDS ()
      {
        atom_mlq = gdk_atom_intern ("application/x-media-library-query", TRUE);
      }

      ~TreeViewXDS () {}

      virtual void
      on_drag_end     (const Glib::RefPtr<Gdk::DragContext>& context)
      {
        xds_drag_end (context);
      }

      virtual void
      on_drag_begin   (const Glib::RefPtr<Gdk::DragContext>& context)
      {
        Glib::ustring dir_basename;
        dir_basename.append (artist.get());
        dir_basename.append (" - ");
        dir_basename.append (album.get());
        dir_basename.append (".mlq");

        context->set_icon (pixbuf, 0, 0);       

        xds_drag_begin (context, dir_basename); 
      }

      virtual void
      on_drag_data_get (const Glib::RefPtr<Gdk::DragContext>& context,
                        Gtk::SelectionData&                   selection_data,
                        unsigned int                          info,
                        unsigned int                          time)
      {
        using namespace std;
        using namespace Glib;

        if (info == 0)
          {
            xds_drag_data_get (context, selection_data, info, time);
          }
        else
          {
            string id_artist   = library->get_metadatum_id (DATUM_ARTIST);
            string id_album    = library->get_metadatum_id (DATUM_ALBUM);

            ustring outstr;

            outstr  .append ("mlq:///?")
                    .append (id_artist)
                    .append ("=")
                    .append (artist.get())
                    .append ("&")
                    .append (id_album)
                    .append ("=")
                    .append (album.get());

            Bmp::URI uri (outstr);
            uri.escape ();

            const char *str = strdup (Glib::ustring(uri).c_str());
            GtkSelectionData *sdata = selection_data.gobj();
            gtk_selection_data_set (sdata, atom_mlq, 8, (const unsigned char*)strdup(str), strlen(str)+1); 
          }
      }

      void
      set_album  (const Glib::ustring& album)
      {
        this->album  = album;
      }

      void
      set_artist (const Glib::ustring& artist)
      {
        this->artist = artist;
      }

      void
      clear_artist_album ()
      {
        artist.reset ();
        album.reset ();
      }
    
      void
      set_pixbuf (Glib::RefPtr<Gdk::Pixbuf>& pixbuf)
      {
        this->pixbuf = pixbuf;
      }

      void
      set_source ()
      {
        xds_set ();

        GtkTargetList *target_list;

        target_list = gtk_drag_source_get_target_list (dynamic_cast<Gtk::Widget*>(this)->gobj());
        gtk_target_list_ref (target_list);
        gtk_target_list_add (target_list, atom_mlq, GtkTargetFlags(GTK_TARGET_SAME_APP), 1);
        gtk_drag_source_set_target_list (dynamic_cast<Gtk::Widget*>(this)->gobj(), target_list);
        gtk_target_list_unref (target_list);
      }

      void
      unset_source ()
      {
        xds_unset ();
      }

    protected:

      virtual void
      xds_drag_action     (const Glib::ustring& uri)
      {
          using namespace Bmp::DB;
          using namespace Bmp::Library;
          using namespace Glib; 
          using namespace std; 

          string id_artist = library->get_metadatum_id (DATUM_ARTIST);
          string id_album  = library->get_metadatum_id (DATUM_ALBUM);

          dynamic_cast<Gtk::Widget *>(this)->set_sensitive (false);

          ustring outstr;

          outstr    .append ("mlq:///?")
                    .append (id_artist)
                    .append ("=")
                    .append (artist.get())
                    .append ("&")
                    .append (id_album)
                    .append ("=")
                    .append (album.get());

          Bmp::URI u (outstr);
          u.escape ();

          std::ofstream outfile;
          outfile.open (Glib::filename_from_uri (uri).c_str());
          outfile << "#MLQ\n"; 
          outfile << Glib::ustring (u).c_str(); 
          outfile << "\n";
          outfile.close ();

          dynamic_cast<Gtk::Widget *>(this)->set_sensitive (true);
      }
    
      virtual bool
      xds_acceptable      (const Glib::ustring& uri)
      {
          if (!artist && !album) return false;
          return true;
      }

    private:
    
      boost::optional<Glib::ustring> artist;
      boost::optional<Glib::ustring> album;
      Glib::RefPtr<Gdk::Pixbuf> pixbuf;
      GdkAtom atom_mlq;

    };
}

namespace Bmp
{
      void
      LibraryUI::update_track_view (int id)
      {
        switch (LibraryPage(id))
          {
            case PAGE_ALBUMS: update_track_view_from_albums (); break;
            case PAGE_SEARCH: update_track_view_from_search (); break;
          }
      }

      void
      LibraryUI::notebook_page_switch (GtkNotebookPage *page, unsigned int id)
      {
        update_track_view (id);
      }

#ifdef HAVE_HAL
      void
      LibraryUI::hal_volume_removed (Library::HALVolume volume)
      {
        debug ("library-ui", "HAL Volume Removed: %s", volume.volume_udi.c_str());
        update_track_view (notebook_library->get_current_page());
      }

      void
      LibraryUI::hal_volume_added (Library::HALVolume volume)
      {
        debug ("library-ui", "HAL Volume Added: %s", volume.volume_udi.c_str());
        update_track_view (notebook_library->get_current_page());
      }
#endif

      LibraryUI::LibraryUI (BaseObjectType                        *cobject,
                            const Glib::RefPtr<Gnome::Glade::Xml> &xml)
          : Gtk::Window       (cobject),
            m_ref_xml         (xml),
            m_current_rating  (0),
            m_updated         (false),
            m_busy            (false),

            m_disc            (Gdk::Pixbuf::create_from_file
                              (BMP_IMAGE_DIR G_DIR_SEPARATOR_S "audio.png")->scale_simple (64, 64, Gdk::INTERP_BILINEAR)),

            m_disc_small      (Gdk::Pixbuf::create_from_file
                              (BMP_IMAGE_DIR G_DIR_SEPARATOR_S "audio.png")->scale_simple (32, 32, Gdk::INTERP_BILINEAR)),

            m_viewsize        (ViewSize(mcs->key_get<int>("bmp","library-view-size"))),

            m_init            (true)
      {
          bmp_window_set_icon_list (GTK_WIDGET (gobj ()), "library");

          struct ButtonActionProxyI
          {
            char  *widget;
            char  *action;
          };

          struct ImageStockIdI
          {
            char  *widget;
            char  *stockid;
          };

#define LIBRARY_UI_ACTION_IMPORT_MUSIC       "library-action-import-music"
#define LIBRARY_UI_ACTION_IMPORT_RETAG       "library-action-retag-import"

#define LIBRARY_UI_ACTION_EDIT_ALBUMS        "library-action-edit-album"
#define LIBRARY_UI_ACTION_IMPORT_TRACKS      "library-action-import-tracks"
#define LIBRARY_UI_ACTION_REMOVE_ALBUMS      "library-action-remove-albums"
#define LIBRARY_UI_ACTION_RELOCATE_ALBUM     "library-action-relocate-album"

#define LIBRARY_UI_ACTION_PLAY               "library-action-play"
#define LIBRARY_UI_ACTION_ENQUEUE            "library-action-enqueue"
#define LIBRARY_UI_ACTION_REMOVE             "library-action-remove"

#define LIBRARY_UI_ACTION_APPROVE_ITEMS      "library-action-approve-items"
#define LIBRARY_UI_ACTION_APPROVE_SELECTED   "library-action-approve-selected"
#define LIBRARY_UI_ACTION_DROP_ALL           "library-action-drop-all"
#define LIBRARY_UI_ACTION_SHOW_ONLY_NEW      "library-action-show-only-new"

          // --- UIManager + Actions

          bool enabled = false;

#ifdef HAVE_HAL
          enabled = hal->is_initialized(); 
#else
          enabled = true;
#endif

          m_ref_xml->get_widget ("notebook_albums", notebook_albums);

          m_ui_manager = Gtk::UIManager::create ();

          m_library_actions[IMPORT]   = Gtk::ActionGroup::create ("LibraryActionsImport"); 
          m_library_actions[ALBUM]    = Gtk::ActionGroup::create ("LibraryActionsAlbum"); 
          m_library_actions[PLAYBACK] = Gtk::ActionGroup::create ("LibraryActionsPlayback"); 
          m_library_actions[APPROVE]  = Gtk::ActionGroup::create ("LibraryActionsApprove"); 


          m_library_actions[IMPORT]->add ( Gtk::Action::create (LIBRARY_UI_ACTION_IMPORT_MUSIC,  _("Import Music")),
                                          sigc::mem_fun (this, &Bmp::LibraryUI::import_tracks));
          m_library_actions[IMPORT]->add ( Gtk::Action::create (LIBRARY_UI_ACTION_IMPORT_RETAG,  _("Retag Files and Import")),
                                          sigc::mem_fun (this, &Bmp::LibraryUI::import_tracks_as_album));

          m_library_actions[ALBUM]->add ( Gtk::Action::create (LIBRARY_UI_ACTION_EDIT_ALBUMS,    _("Edit Album Metadata")),
                                          sigc::mem_fun (this, &Bmp::LibraryUI::albums_modify));

#if 0
          m_library_actions[ALBUM]->add ( Gtk::Action::create (LIBRARY_UI_ACTION_IMPORT_TRACKS,  _("Import Tracks Into Album")),
                                          sigc::mem_fun (this, &Bmp::LibraryUI::import_tracks_into_album));
#endif

          m_library_actions[ALBUM]->add ( Gtk::Action::create (LIBRARY_UI_ACTION_REMOVE_ALBUMS,   _("Remove Albums")),
                                          sigc::mem_fun (this, &Bmp::LibraryUI::remove_albums_action_cb));
          m_library_actions[ALBUM]->add ( Gtk::Action::create (LIBRARY_UI_ACTION_RELOCATE_ALBUM, _("Relocate Album")),
                                          sigc::mem_fun (this, &Bmp::LibraryUI::relocate_album));

          m_library_actions[PLAYBACK]->add ( Gtk::Action::create (LIBRARY_UI_ACTION_PLAY,    _("_Play All")),
                                              sigc::mem_fun (this, &Bmp::LibraryUI::play));
          m_library_actions[PLAYBACK]->add ( Gtk::Action::create (LIBRARY_UI_ACTION_ENQUEUE, _("_Enqueue All")),
                                              sigc::mem_fun (this, &Bmp::LibraryUI::enqueue));
          m_library_actions[PLAYBACK]->add ( Gtk::Action::create (LIBRARY_UI_ACTION_REMOVE,  _("Remove Selected")),
                                              sigc::mem_fun (this, &Bmp::LibraryUI::remove));

          m_ref_xml->get_widget ("l_play", l_play);
          m_ref_xml->get_widget ("l_enqueue", l_enqueue);

          Gtk::CheckButton *check_button;
          Gtk::Button      *button;
          m_ref_xml->get_widget ("cb_show_only_new",  check_button);


          m_library_actions[APPROVE]->add ( Gtk::Action::create (LIBRARY_UI_ACTION_APPROVE_ITEMS,     _("Approve All")),
                                            sigc::mem_fun (this, &Bmp::LibraryUI::approve_all));
          m_ref_xml->get_widget ("b_approve", button);
          m_library_actions[APPROVE]->get_action (LIBRARY_UI_ACTION_APPROVE_ITEMS)->connect_proxy (*button);


          m_library_actions[APPROVE]->add ( Gtk::Action::create (LIBRARY_UI_ACTION_APPROVE_SELECTED,  _("Approve Selected")),
                                            sigc::mem_fun (this, &Bmp::LibraryUI::approve_selected));
          m_ref_xml->get_widget ("b_approve_selected", button);
          m_library_actions[APPROVE]->get_action (LIBRARY_UI_ACTION_APPROVE_SELECTED)->connect_proxy (*button);


          m_library_actions[APPROVE]->add ( Gtk::Action::create (LIBRARY_UI_ACTION_DROP_ALL,          _("Drop Remaining")),
                                            sigc::mem_fun (this, &Bmp::LibraryUI::drop_items));
          m_ref_xml->get_widget ("b_drop", button);
          m_library_actions[APPROVE]->get_action (LIBRARY_UI_ACTION_DROP_ALL)->connect_proxy (*button);


          m_library_actions[APPROVE]->add ( Gtk::ToggleAction::create (LIBRARY_UI_ACTION_SHOW_ONLY_NEW, ""));
          Glib::RefPtr<Gtk::ToggleAction> action = Glib::RefPtr<Gtk::ToggleAction>::cast_static
                                                   (m_library_actions[APPROVE]->get_action (LIBRARY_UI_ACTION_SHOW_ONLY_NEW));
          action->signal_toggled().connect (sigc::mem_fun (this, &Bmp::LibraryUI::albums_update));
          m_library_actions[APPROVE]->get_action (LIBRARY_UI_ACTION_SHOW_ONLY_NEW)->connect_proxy (*check_button);


          gtk_label_set_markup (GTK_LABEL (gtk_bin_get_child (GTK_BIN (check_button->gobj()))),       _("<small>Display New Items</small>"));
 

          const ButtonActionProxyI _import[] =
          {
            { "b_import_music",
              LIBRARY_UI_ACTION_IMPORT_MUSIC },

            { "b_import_tracks_as_album",
              LIBRARY_UI_ACTION_IMPORT_RETAG },
          };

          const ButtonActionProxyI _album[] =
          {
            { "b_albums_modify",
              LIBRARY_UI_ACTION_EDIT_ALBUMS    },

/*
            { "b_import_tracks_into_album",
              LIBRARY_UI_ACTION_IMPORT_TRACKS  },
*/

            { "b_remove_album",
              LIBRARY_UI_ACTION_REMOVE_ALBUMS  },

            { "b_relocate_album",
              LIBRARY_UI_ACTION_RELOCATE_ALBUM },
          };

          const ButtonActionProxyI _playback[] =
          {
            { "b_play",
              LIBRARY_UI_ACTION_PLAY    },

            { "b_enqueue",
              LIBRARY_UI_ACTION_ENQUEUE },

            { "b_remove",
              LIBRARY_UI_ACTION_REMOVE  },
          };

          for (unsigned int n = 0; n < G_N_ELEMENTS (_import); ++n)
          {
            Gtk::Button *button;
            m_ref_xml->get_widget (_import[n].widget, button);
            m_library_actions[IMPORT]->get_action (_import[n].action)->connect_proxy (*button);
            m_library_actions[IMPORT]->get_action (_import[n].action)->set_sensitive (enabled);
          }
          m_ui_manager->insert_action_group (m_library_actions[IMPORT]);

          for (unsigned int n = 0; n < G_N_ELEMENTS (_album); ++n)
          {
            Gtk::Button *button;
            m_ref_xml->get_widget (_album[n].widget, button);
            m_library_actions[ALBUM]->get_action (_album[n].action)->connect_proxy (*button);
            m_library_actions[ALBUM]->get_action (_album[n].action)->set_sensitive (enabled);
          }
          m_ui_manager->insert_action_group (m_library_actions[ALBUM]);

          for (unsigned int n = 0; n < G_N_ELEMENTS (_playback); ++n)
          {
            Gtk::Button *button;
            m_ref_xml->get_widget (_playback[n].widget, button);
            m_library_actions[PLAYBACK]->get_action (_playback[n].action)->connect_proxy (*button);
            m_library_actions[PLAYBACK]->get_action (_playback[n].action)->set_sensitive (false);
          }
          m_ui_manager->insert_action_group (m_library_actions[PLAYBACK]);
          m_library_actions[ALBUM]->set_sensitive (false);

          add_accel_group (m_ui_manager->get_accel_group ());

          m_ref_xml->get_widget ("notebook_library", notebook_library);
          conn_notebook = notebook_library->signal_switch_page().connect
                              (sigc::mem_fun (*this, &Bmp::LibraryUI::notebook_page_switch));
                         
          // Header Banner Image
          {
            Glib::RefPtr<Gdk::Pixbuf> left, right, slice;
            left  = Gdk::Pixbuf::create_from_file (BMP_LIBRARY_IMAGE_DIR G_DIR_SEPARATOR_S "header-logo-left.png");
            right = Gdk::Pixbuf::create_from_file (BMP_LIBRARY_IMAGE_DIR G_DIR_SEPARATOR_S "header-logo-right.png");
            slice = Gdk::Pixbuf::create_from_file (BMP_LIBRARY_IMAGE_DIR G_DIR_SEPARATOR_S "header-logo-slice.png");
            Bmp::BannerImage *banner = new Bmp::BannerImage (left, right, slice);
            Gtk::Alignment *a;
            m_ref_xml->get_widget ("alignment8", a);
            a->add (*banner);
            a->show_all ();
            banner->show_all ();
            banner->set_size_request (-1, 78);
          }

          // Throbbers
          {
            dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("i_wait1"))->set
            (BMP_IMAGE_DIR G_DIR_SEPARATOR_S "wait.gif");

            dynamic_cast<Gtk::Image *>(m_ref_xml->get_widget ("i_wait_albums"))->set
            (BMP_IMAGE_DIR G_DIR_SEPARATOR_S "wait.gif");
          }

          const ImageStockIdI _images[] =
          {

            { "i_add",
              BMP_STOCK_ADD_MUSIC    },

            { "i_add_tracks_album",
              BMP_STOCK_ADD_MUSIC    },

            { "i_remove_album",
              BMP_STOCK_REMOVE_MUSIC },

            { "i_remove_tracks",
              BMP_STOCK_DELETE       },

            { "i_import_tracks",
              BMP_STOCK_ADD_MUSIC    },

            { "i_tachikoma1",
              BMP_STOCK_EDIT_ALBUM   },

            { "i_tachikoma2",
              BMP_STOCK_MOVE_ALBUM   },

/*
            { "i_tachikoma3",
              BMP_STOCK_EDIT_ALBUM   },
*/

            { "i_play",
              BMP_STOCK_MEDIA_PLAY   },

            { "i_enqueue",
              BMP_STOCK_ADD          },

            { "i_approve",
              BMP_STOCK_APPROVE      },

            { "i_approve_selected",
              BMP_STOCK_APPROVE      },

            { "i_drop",
              BMP_STOCK_DELETE       },
          };

          for (unsigned int n = 0; n < G_N_ELEMENTS (_images); ++n)
          {
            Gtk::Image *image;
            m_ref_xml->get_widget (_images[n].widget, image);
            image->set (Gtk::StockID (_images[n].stockid), Gtk::ICON_SIZE_SMALL_TOOLBAR);
          }

          m_ref_xml->get_widget ("label_current_album", label_current_album);
          label_current_album->set_text("");

          // Close button 
          dynamic_cast<Gtk::Button *>(m_ref_xml->get_widget ("b_close"))->signal_clicked().connect
                                      (sigc::mem_fun (*this, &Gtk::Widget::hide));
                                                    
          m_ref_xml->get_widget ("cb_view_size", cb_view_size);
          mcs_bind->bind_cbox (cb_view_size, "bmp", "library-view-size"); 
          mcs->subscribe ("LibraryUI", "bmp", "library-view-size",
                  sigc::mem_fun (this, &Bmp::LibraryUI::mcs_library_view_size_changed));

          // Progress for modifications
          m_ref_xml->get_widget ("progress_modify", progress_modify);
          library->signal_modify_start (). connect (sigc::bind (sigc::ptr_fun (&on_modify_start), progress_modify));
          library->signal_modify_step  (). connect (sigc::bind (sigc::ptr_fun (&on_modify_step), progress_modify));
          library->signal_modify_end   (). connect (sigc::bind (sigc::ptr_fun (&on_modify_end), progress_modify));

          // Search widgets
          using namespace Bmp::Library;
          using namespace Gtk;

          cs_store = Gtk::ListStore::create (cr_search);
          for (unsigned int n = 0; n < G_N_ELEMENTS (search_data); ++n) 
            {
              TreeModel::iterator i = cs_store->append ();
              DatumDefine define = library->get_metadatum_define (search_data[n]);
              (*i)[cr_search.name] = _(define.title); 
              (*i)[cr_search.id] = define.id;
            }

          static boost::format f_cbox     ("cbox_%d");
          static boost::format f_cbutton  ("cb_%d");
          static boost::format f_entry    ("entry_%d");
          for (int n = 0; n < N_FILTERS; ++n)
            {
              m_ref_xml->get_widget ((f_cbutton % (n+1)).str(), cbutton[n]);
              m_ref_xml->get_widget ((f_cbox    % (n+1)).str(), cbox[n]);
              m_ref_xml->get_widget ((f_entry   % (n+1)).str(), entry[n]);

              cbutton[n]->signal_toggled ().connect
                      (sigc::bind (sigc::mem_fun (*this, &Bmp::LibraryUI::filter_toggled), FilterID(n)));

              cbox[n]->clear ();
              Gtk::CellRendererText *cell = Gtk::manage ( new Gtk::CellRendererText() );
              cbox[n]->pack_start (*cell);
              cbox[n]->add_attribute (*cell, "text", 0);
              cbox[n]->set_model (cs_store);
              cbox[n]->set_active (search_default[n]);

              entry[n]->signal_changed().connect
                      (sigc::mem_fun (*this, &Bmp::LibraryUI::update_track_view_from_search));
              cbutton[n]->signal_toggled().connect
                      (sigc::mem_fun (*this, &Bmp::LibraryUI::update_track_view_from_search));
              cbox[n]->signal_changed().connect
                      (sigc::mem_fun (*this, &Bmp::LibraryUI::update_track_view_from_search));
            }

          m_ref_xml->get_widget ("cb_rating", cb_rating);
          m_ref_xml->get_widget ("i_rating", i_rating);
          m_ref_xml->get_widget ("evb_rating", evb_rating);

          cb_rating->signal_toggled().connect
                      (sigc::mem_fun (*this, &Bmp::LibraryUI::cb_rating_toggled));
          evb_rating->signal_button_press_event().connect
                      (sigc::mem_fun (*this, &Bmp::LibraryUI::evb_rating_button_press));

          m_ref_xml->get_widget ("cb_played", cb_played);
          m_ref_xml->get_widget ("cbox_playdate", cbox_playdate);

          cb_played->signal_toggled().connect
                        (sigc::mem_fun (*this, &Bmp::LibraryUI::cb_played_toggled));
          cbox_playdate->set_active (0);
          cbox_playdate->signal_changed().connect
                        (sigc::mem_fun (*this, &Bmp::LibraryUI::update_track_view_from_search));

          // Load 'stars' images
          const unsigned int N_RATINGS = 6;

          for (unsigned int n = 0; n < N_RATINGS; ++n)
            {
              static boost::format star ("stars_%d.png");
              rating[n] = Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_RATING_IMAGE_DIR, (star % n).str()));
            }

          i_rating->set (rating[0]);
          i_rating->set_sensitive (false);

          // Watch shutdown-request from SystemControl
          g_signal_connect (G_OBJECT (bmp_system_control),
                            "shutdown-request",
                            G_CALLBACK (Bmp::LibraryUI::shutdown_request),
                            this);

          m_ref_xml->get_widget ("notebook_add",  notebook_add);
          m_ref_xml->get_widget ("view_tracks",   view_tracks);
          m_ref_xml->get_widget ("albums_filter", albums_filter);

          view_albums = new Bmp::TreeViewXDS;
          view_albums->show_all ();
          view_albums->set_headers_visible (false);

          Gtk::ScrolledWindow *sw;
          m_ref_xml->get_widget ("sw_albums", sw);
          sw->add (*view_albums);

          //---Renderer type
          enum Renderer
          {
            R_TEXT,
            R_PIXBUF,
          };

          //Tracks view
          struct { 
              char    *title;
              double   xalign;
              int      width;
              Renderer renderer;
          } headers_tracks[] = {
              { _("Track"),      1.0, 30,   R_TEXT},
              { _("Title"),      0.0, 100,  R_TEXT},
              { _("Rating"),     0.0, 80,   R_PIXBUF},
              { _("Artist"),     0.0, 80,   R_TEXT},
              { _("Album"),      0.0, 80,   R_TEXT},
              { _("Genre"),      0.0, 80,   R_TEXT},
              { _("Bitrate"),    0.0, 30,   R_TEXT},
              { _("Samplerate"), 0.0, 30,   R_TEXT},
          };

          for (unsigned int n = 0; n < G_N_ELEMENTS(headers_tracks); ++n)
          {
              Gtk::TreeViewColumn *column = 0;
              Gtk::CellRenderer *cell_ = 0;

              if (headers_tracks[n].renderer == R_PIXBUF)
              {
                  Gtk::CellRendererPixbuf *cell = Gtk::manage ( new Gtk::CellRendererPixbuf() );
                  cell->property_xalign() = headers_tracks[n].xalign;
                  view_tracks->append_column (headers_tracks[n].title, *cell);
                  column = view_tracks->get_column (n);
#ifndef HAVE_HAL
                  column->set_cell_data_func (*cell, sigc::mem_fun (*this, &Bmp::LibraryUI::rating_cell_data_func));
#endif
                  column->set_resizable   (false);
                  column->set_sort_column (n);
                  column->set_expand      (false);
                  column->property_min_width() = headers_tracks[n].width; 
                  column->property_max_width() = headers_tracks[n].width; 
                  cell_ = cell;
                  cell_rating = cell;
              }
              else 
              {
                  Gtk::CellRendererText *cell = Gtk::manage ( new Gtk::CellRendererText() );

                  cell->property_xalign() = headers_tracks[n].xalign;
                  view_tracks->append_column (headers_tracks[n].title, *cell);
                  column = view_tracks->get_column (n);
                  column->add_attribute     (*cell, "text", n);
                  column->set_resizable     (true);
                  column->set_sort_column   (n);
                  column->set_expand        (true);
                  column->property_min_width() = headers_tracks[n].width;
                  cell_ = cell;
              }

#ifdef HAVE_HAL
            if (hal->is_initialized())
            {
              column->set_cell_data_func (*cell_, sigc::mem_fun (*this, &Bmp::LibraryUI::tracks_set_row_sensitive));
            }
#endif
          }

          store_tracks = Gtk::ListStore::create (tracks);
          view_tracks->set_model (store_tracks);
          view_tracks->get_selection()->set_mode (Gtk::SELECTION_MULTIPLE);
          view_tracks->get_selection()->signal_changed().connect
                (sigc::mem_fun (*this, &Bmp::LibraryUI::tracks_selection_changed));
          view_tracks->signal_row_activated().connect
                (sigc::mem_fun (*this, &Bmp::LibraryUI::activate_default));
          view_tracks->signal_event ().connect
                (sigc::mem_fun (*this, &Bmp::LibraryUI::tracks_event));

#ifdef HAVE_HAL
          if (hal->is_initialized())
            {
              view_tracks->get_selection()->set_select_function
                (sigc::mem_fun (*this, &Bmp::LibraryUI::tracks_row_can_be_selected));
            }
#endif

          //Albums view
         struct { 
              char    *title;
              double   xalign;
              int      width;
              Renderer renderer;
          } headers_albums[] = {
              { _("Cover"),      0.5, 72,  R_PIXBUF },
              { _("Album"),      0.0, 250, R_TEXT   },
          };

          for (unsigned int n = 0; n < G_N_ELEMENTS(headers_albums); ++n)
          {
            switch (headers_albums[n].renderer)
            {
              case R_PIXBUF:
              {
                Gtk::CellRendererPixbuf *cell = Gtk::manage ( new Gtk::CellRendererPixbuf() );

                cell->property_xalign() = headers_albums[n].xalign;
                cell->property_xpad() = 3;
                cell->property_ypad() = 4;

                view_albums->append_column (headers_albums[n].title, *cell);
                Gtk::TreeViewColumn *column = view_albums->get_column (n);

                column->add_attribute (*cell, "pixbuf", n);
                column->set_resizable (false);
                column->set_expand (false);
                column->set_sizing (Gtk::TREE_VIEW_COLUMN_FIXED);
                column->set_fixed_width (m_viewsize?40:72);
                break;
              }

              case R_TEXT:
              {
                cell_album_text = Gtk::manage ( new Gtk::CellRendererText() );

                cell_album_text->property_xalign() = headers_albums[n].xalign;
                cell_album_text->property_size() = (m_viewsize?8:14) * PANGO_SCALE;
                cell_album_text->property_width() = 700;

                view_albums->append_column (headers_albums[n].title, *cell_album_text);
                Gtk::TreeViewColumn *column  = view_albums->get_column (n);

                column->add_attribute (*cell_album_text, "markup", n);
                column->set_resizable (false);
                column->set_expand (false);
                column->set_sizing (Gtk::TREE_VIEW_COLUMN_FIXED);
                column->set_fixed_width(headers_albums[n].width);
                break;
              }
            }
          }

          store_albums = Gtk::ListStore::create (albums);
          store_albums->set_default_sort_func (sigc::mem_fun (*this, &Bmp::LibraryUI::albums_default_sort_func));
          store_albums->set_sort_column (-1, Gtk::SORT_ASCENDING);
          store_albums_filter = Gtk::TreeModelFilter::create (store_albums);
          store_albums_filter->set_visible_func (sigc::mem_fun (*this, &Bmp::LibraryUI::albums_row_is_visible));

          view_albums->set_model (store_albums_filter);

          view_albums->get_selection()->set_mode (Gtk::SELECTION_MULTIPLE);
          view_albums->get_selection()->signal_changed().connect
                (sigc::mem_fun (*this, &Bmp::LibraryUI::update_track_view_from_albums));
          view_albums->get_selection()->set_select_function
                (sigc::mem_fun (*this, &Bmp::LibraryUI::albums_record_selection_change));
          view_albums->signal_event ().connect
                (sigc::mem_fun (*this, &Bmp::LibraryUI::albums_event));

          albums_filter->signal_changed().connect
                (sigc::mem_fun (*this, &Bmp::LibraryUI::albums_refilter));

#ifdef HAVE_HAL
          // Connect Bmp::HAL
          if (hal->is_initialized())
            {
              hal->signal_volume_added().connect
                (sigc::mem_fun (*this, &Bmp::LibraryUI::hal_volume_added));
              hal->signal_volume_removed().connect
                (sigc::mem_fun (*this, &Bmp::LibraryUI::hal_volume_removed));
            }
#endif

          // Connect to the library for blocking the UI to avoid race conditions
          library->signal_processing_start().connect
                (sigc::bind (sigc::mem_fun (*this, &Gtk::Widget::set_sensitive), false));
          library->signal_processing_end().connect
                (sigc::bind (sigc::mem_fun (*this, &Gtk::Widget::set_sensitive), true));
          library->signal_row_updated().connect
                (sigc::mem_fun (*this, &Bmp::LibraryUI::check_update));

          m_init = false;
      }

      void
      LibraryUI::check_update (const Glib::ustring&      location,
                               Bmp::Library::UpdateTrack track)
      {
        RowMapTracks::iterator rmt_Iter = m_row_map_tracks.find (location); 
        if (rmt_Iter != m_row_map_tracks.end())
          {
            if (track.rating)
              {
                Gtk::TreeModel::iterator m_Iter = store_tracks->get_iter (rmt_Iter->second.get_path());
                int r_current = (*m_Iter)[tracks.rating];
                if (r_current != track.rating.get())
                  {
                    (*m_Iter)[tracks.rating] = track.rating.get();
                    if (notebook_library->get_current_page() == PAGE_SEARCH) update_track_view (PAGE_SEARCH);
                  }
              }
          }
      }

      void
      LibraryUI::cb_played_toggled ()
      { 
        cbox_playdate->set_sensitive (cb_played->get_active());
        update_track_view_from_search ();
      }

      void
      LibraryUI::cb_rating_toggled ()
      { 
        i_rating->set_sensitive (cb_rating->get_active());
        evb_rating->set_sensitive (cb_rating->get_active());
        update_track_view_from_search ();
      }

      bool
      LibraryUI::evb_rating_button_press (GdkEventButton * ev)
      {
        int n = (int(ev->x+7)/14);
        i_rating->set (rating[n]);
        m_current_rating = n;
        update_track_view_from_search ();
        return false;
      }

      bool
      LibraryUI::on_key_press_event (GdkEventKey *event)
      {
        if (event->keyval == GDK_Escape)
          {
            Gtk::Widget::hide ();
            return true;
          }

        return Gtk::Widget::on_key_press_event (event);
      }

      void
      LibraryUI::rating_cell_data_func (Gtk::CellRenderer* cell_, const Gtk::TreeModel::iterator& m_Iter)
      {
        using namespace Gtk;
        CellRendererPixbuf *cell = dynamic_cast<CellRendererPixbuf *>(cell_);
        cell->property_pixbuf() = rating[((*m_Iter)[tracks.rating])];
      }

#ifdef HAVE_HAL
      bool
      LibraryUI::tracks_row_can_be_selected (const Glib::RefPtr<Gtk::TreeModel>& model,
                                             const Gtk::TreeModel::Path&         path,
                                             bool                                selected)
      {
        Gtk::TreeModel::iterator cur = model->get_iter (path);
        return ((*cur)[tracks.present]);
      }

      void
      LibraryUI::tracks_set_row_sensitive (Gtk::CellRenderer* cell, const Gtk::TreeModel::iterator& m_Iter)
      {
        cell->property_sensitive() = ((*m_Iter)[tracks.present]); 
        if (cell == cell_rating) rating_cell_data_func (cell, m_Iter); 
      }

      bool
      LibraryUI::all_tracks_present ()
      {
        using namespace Gtk;
        bool all = true;
          
        TreeNodeChildren nodes = store_tracks->children ();

        for (TreeNodeChildren::const_iterator m_Iter = nodes.begin(), end = nodes.end(); m_Iter != end; ++m_Iter)
          {
            if (!((*m_Iter)[tracks.present]))
              {
                all = false;
                break;
              }
          }
        return all;
      }
#endif

      void
      LibraryUI::mcs_library_view_size_changed
      (MCS_CB_DEFAULT_SIGNATURE)
      {
        if (this->m_init) return;
        m_viewsize = ViewSize(mcs->key_get<int>("bmp","library-view-size"));
        cell_album_text->property_size() = (m_viewsize?8:14) * PANGO_SCALE;
        Gtk::TreeViewColumn *column  = view_albums->get_column (0);
        column->set_fixed_width(m_viewsize?40:72);
        view_albums->queue_draw ();
        albums_update ();
      }

      void
      LibraryUI::activate_default (const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn* column)
      {
        Gtk::Window::activate_default ();
      }
      
      LibraryUI::~LibraryUI ()
      {
        conn_notebook.disconnect ();
        delete view_albums; 
        m_ref_xml.clear();
      }

      LibraryUI*
      LibraryUI::create ()
      {
        const std::string path (BMP_GLADE_DIR G_DIR_SEPARATOR_S "library_ui.glade");
        Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);
        LibraryUI *library_ui = 0;
        glade_xml->get_widget_derived ("library_ui", library_ui);
        return library_ui; 
      }

      void
      LibraryUI::filter_toggled (FilterID id)
      {
        bool active = cbutton[int(id)]->get_active();
        cbox[int(id)]->set_sensitive (active);
        entry[int(id)]->set_sensitive (active);
      }

      void
      LibraryUI::fetch_cover (const Gtk::TreePath& path)
      { 
        using namespace Gtk;
        using namespace Gdk;
        using namespace Glib;
        using namespace std;

        TreeModel::iterator m_Iter = store_albums_filter->convert_iter_to_child_iter
                                                      (store_albums_filter->get_iter (path));

        AlbumFlags flags = (*m_Iter)[albums.flags];
        if (flags & A_HAS_COVER) return;

        set_sensitive (false); 
        RefPtr<Pixbuf> cover;

        try {
            ustring asin = library->get_asin_for_album ((*m_Iter)[albums.artist], (*m_Iter)[albums.album]);
            try { 
                Amazon::get_cover_by_asin (asin, cover, false);
                if (cover)
                {
                  library->set_asin_for_album ((*m_Iter)[albums.artist], (*m_Iter)[albums.album], asin);
                  (*m_Iter)[albums.cover] = cover->scale_simple (m_viewsize?32:64, m_viewsize?32:64, INTERP_BILINEAR);
                  (*m_Iter)[albums.flags] = AlbumFlags (flags | A_HAS_COVER);
                }
            } catch (...) {}

          }
        catch (Bmp::Library::Exception& cxe)
          {
            ustring asin;
            try {
                Amazon::get_cover_artist_album ((*m_Iter)[albums.artist], (*m_Iter)[albums.album], cover, asin, false);
                if (cover)
                {
                  library->set_asin_for_album ((*m_Iter)[albums.artist], (*m_Iter)[albums.album], asin);
                  (*m_Iter)[albums.cover] = cover->scale_simple (m_viewsize?32:64, m_viewsize?32:64, INTERP_BILINEAR);
                  (*m_Iter)[albums.flags] = AlbumFlags (flags | A_HAS_COVER);
                }
            } catch (...) {}
          } catch (...) {}
          set_sensitive (true);
        }

        bool
        LibraryUI::tracks_event (GdkEvent * ev)
        {
          using namespace Gtk;

          GdkEventButton *event = 0;

          if (ev->type == GDK_BUTTON_PRESS)
            {
              event = reinterpret_cast<GdkEventButton *>(ev);
              TreeViewColumn *column, *column_rating = view_tracks->get_column (2);
              TreeModel::Path path;
              int cell_x, cell_y;

              if (view_tracks->get_path_at_pos (int(event->x),
                                int(event->y),
                                path,
                                column,
                                cell_x,
                                cell_y)) 
                {
                  if (column == column_rating)
                    {
                      using namespace Gtk;
                      using namespace Library;

                      Gtk::TreeModel::iterator m_Iter = store_tracks->get_iter (path);

                      int n = (int(cell_x+7)/14);

                      UpdateTracks update_tracks;
                      UpdateTrack  update_track;

                      update_track.location_cur = ((*m_Iter)[tracks.location]); 
                      update_track.rating = n;
                      update_tracks.push_back (update_track);
                      try
                        {
                          library->modify_tracks (update_tracks, false, true);
                          (*m_Iter)[tracks.rating] = n; 
                        }
                      catch (...)
                        {
                        }

                      return true;
                    }
                } 
            }
          return false;
      }

      bool
      LibraryUI::albums_event (GdkEvent * ev)
        {
          using namespace Gtk;

          GdkEventButton *event = 0;
          if (ev->type == GDK_BUTTON_PRESS)
            {
              event = reinterpret_cast<GdkEventButton *>(ev);
              TreeViewColumn *column, *column_0 = view_albums->get_column (0);
              TreeModel::Path path;
              int cell_x, cell_y;

              if (view_albums->get_path_at_pos (int(event->x),
                                int(event->y),
                                path,
                                column,
                                cell_x,
                                cell_y)) 
                {
                    TreeModel::iterator cur = store_albums_filter->get_iter (path);

                    AlbumFlags flags = (*cur)[albums.flags];
                    bool has_cover = (flags & A_HAS_COVER); 
 
                  if (column == column_0)
                    {
                      if ((event->type == GDK_BUTTON_PRESS) && !has_cover)
                        {
                          fetch_cover (path);
                          return true;
                        }
                    }
                }
            } 
          return false;
      }

      int 
      LibraryUI::albums_default_sort_func (const Gtk::TreeModel::iterator& iter_a, const Gtk::TreeModel::iterator& iter_b)
      {
        Glib::ustring artist_a = (*iter_a)[albums.artist]; 
        Glib::ustring artist_b = (*iter_b)[albums.artist]; 
        Glib::ustring album_a  = (*iter_a)[albums.album];
        Glib::ustring album_b  = (*iter_b)[albums.album];
        Glib::ustring str_a = artist_a + " " + album_a; 
        Glib::ustring str_b = artist_b + " " + album_b; 
        return str_a.compare(str_b);
      }

      void
      LibraryUI::fill_tracks_from_rows (const Bmp::DB::VRow& vector) 
      {
        using namespace Bmp::DB;
        using namespace Bmp::Library;
        using namespace Gtk;

        m_row_map_tracks.clear ();

        for (VRow::const_iterator v = vector.begin (), end = vector.end (); v != end; ++v) 
        { 
          Track track (*v);
          TreeModel::iterator m_Iter = store_tracks->append ();

          if (track.title) (*m_Iter)[tracks.title] = track.title.get (); 
          if (track.artist) (*m_Iter)[tracks.artist] = track.artist.get (); 
          if (track.album) (*m_Iter)[tracks.album] = track.album.get (); 
          if (track.tracknumber) (*m_Iter)[tracks.track] = track.tracknumber.get (); 
          if (track.genre) (*m_Iter)[tracks.genre] = track.genre.get (); 
          if (track.bitrate) (*m_Iter)[tracks.bitrate] = track.bitrate.get (); 
          if (track.samplerate) (*m_Iter)[tracks.samplerate] = track.samplerate.get (); 
          if (track.location) (*m_Iter)[tracks.location] = track.location.get (); 

          if (track.rating)
            {
              (*m_Iter)[tracks.rating] = track.rating.get ();
            }
          else
            {
              (*m_Iter)[tracks.rating] = 0; 
            }

#ifdef HAVE_HAL
          if (hal->is_initialized())
             (*m_Iter)[tracks.present] = ((track.volume_udi && hal->volume_is_present (track.volume_udi.get()) || (!track.volume_udi)));
          else
             (*m_Iter)[tracks.present] = true; 
#endif
          m_row_map_tracks.insert (std::make_pair (track.location.get(), Gtk::TreeRowReference (store_tracks, Gtk::TreePath (m_Iter))));
        }
      }

      //////////////////////////// Selections
      void
      LibraryUI::tracks_selection_changed ()
      {
        bool selected = (view_tracks->get_selection ()->count_selected_rows() > 0);

        m_library_actions[PLAYBACK]->get_action (LIBRARY_UI_ACTION_REMOVE)->set_sensitive (selected);

        l_play->set_markup_with_mnemonic ((selected) ? _("_Play Selected")     : _("_Play All"));
        l_enqueue->set_markup_with_mnemonic ((selected) ? _("_Enqueue Selected")  : _("_Enqueue All"));
      }

      bool
      LibraryUI::albums_record_selection_change  (const Glib::RefPtr<Gtk::TreeModel>& model,
                                                  const Gtk::TreeModel::Path&         path,
                                                  bool                                selected)
      {
        using namespace Gtk;

        TreeModel::iterator m_Iter = model->get_iter (path); 

        if (!selected)
          {
            m_selection_list_albums.push_back (std::make_pair ((*m_Iter)[albums.rowuid],
                                               Gtk::TreeRowReference (model, path))); 


          }
        else
          {
            for (AlbumsRowSelectionList::iterator pair  = m_selection_list_albums.begin(),
                                                  end   = m_selection_list_albums.end(); pair != end; ++pair)
            {
              AlbumRowSelectionPair const& selection_pair = (*pair);
              if (selection_pair.first == ((*m_Iter)[albums.rowuid]))
                {
                  m_selection_list_albums.erase (pair);
                  break;
                }
            }
          }
        return true;
      }

      void
      LibraryUI::update_track_view_from_search ()
      {
        using namespace Bmp::DB;
        using namespace Bmp::Library;
        using namespace Gtk;

        label_current_album->set_text ("");
        store_tracks->clear ();

        AttributeList attrs; 
        VRow vector;
        for (unsigned int n = 0; n < 3; ++n)
        {
            if (cbutton[n]->get_active ())
            {
              Datum datum = search_data[cbox[n]->get_active_row_number()]; 
              std::string id = library->get_metadatum_id (datum);
              if (!entry[n]->get_text().empty())
                {
                  attrs.push_back (Attribute (DB::FUZZY, id, entry[n]->get_text())); 
                }
            }
        }

        if (cb_rating->get_active() && m_current_rating > 0)
          {
            attrs.push_back (Attribute (DB::GREATER_THAN_OR_EQUAL,
                  library->get_metadatum_id (Bmp::Library::DATUM_RATING), int(m_current_rating))); 
          }

        if (cb_played->get_active())
          {
            const int day_s = 86400;

            int current = time (NULL);
            int match   = current;

            enum
            {
              R_DAY,
              R_WEEK,
              R_MONTH,
              R_3_MONTHS,
              R_6_MONTHS,
              R_YEAR,
            };

            switch (cbox_playdate->get_active_row_number()) 
              {
                case R_DAY:       match = (current % day_s) - day_s;
                case R_WEEK:      match = current - (7          * day_s); break;
                case R_MONTH:     match = current - ((4*7)      * day_s); break;
                case R_3_MONTHS:  match = current - ((3*(4*7))  * day_s); break;
                case R_6_MONTHS:  match = current - ((6*(4*7))  * day_s); break;
                case R_YEAR:      match = current - ((12*(4*7)) * day_s); break;
              }
            attrs.push_back (Attribute (DB::GREATER_THAN_OR_EQUAL, library->get_metadatum_id (Bmp::Library::DATUM_PLAYDATE), int(match))); 
          }

        if (!attrs.empty())
          {
            library->query (attrs, vector);
            fill_tracks_from_rows (vector);

            if (!store_tracks->children ().empty())
              {
                m_library_actions[PLAYBACK]->get_action (LIBRARY_UI_ACTION_PLAY)->set_sensitive (true);
                m_library_actions[PLAYBACK]->get_action (LIBRARY_UI_ACTION_ENQUEUE)->set_sensitive (true);
              }
          }
      }

      void
      LibraryUI::update_track_view_from_albums ()
      {
        using namespace Bmp::DB;
        using namespace Bmp::Library;
        using namespace Gtk;
        using namespace Glib;
        using namespace std;

        label_current_album->set_text ("");
        store_tracks->clear ();

        if (m_busy) return;

        m_library_actions[ALBUM]    ->  set_sensitive (false);
        m_library_actions[ALBUM]    ->  get_action (LIBRARY_UI_ACTION_REMOVE_ALBUMS)    ->  set_sensitive (false);
        m_library_actions[PLAYBACK] ->  get_action (LIBRARY_UI_ACTION_PLAY)             ->  set_sensitive (false);
        m_library_actions[PLAYBACK] ->  get_action (LIBRARY_UI_ACTION_ENQUEUE)          ->  set_sensitive (false);
        m_library_actions[APPROVE]  ->  get_action (LIBRARY_UI_ACTION_APPROVE_SELECTED) ->  set_sensitive (false);

        if (!m_selection_list_albums.empty())
          {
            if (!view_albums->get_selection()->count_selected_rows())
              {
                m_selection_list_albums.clear();
                return;
              }
          }
        else return;

        RefPtr<Gtk::ToggleAction> action = RefPtr<Gtk::ToggleAction>::cast_static
                          (m_library_actions[APPROVE]->get_action (LIBRARY_UI_ACTION_SHOW_ONLY_NEW));
        bool show_only_new = action->get_active (); 

        m_library_actions[ALBUM]    ->  get_action (LIBRARY_UI_ACTION_REMOVE_ALBUMS)  ->  set_sensitive (true);
        m_library_actions[PLAYBACK] ->  get_action (LIBRARY_UI_ACTION_PLAY)           ->  set_sensitive (true);
        m_library_actions[PLAYBACK] ->  get_action (LIBRARY_UI_ACTION_ENQUEUE)        ->  set_sensitive (true);
        m_library_actions[APPROVE]  ->  get_action (LIBRARY_UI_ACTION_APPROVE_ITEMS)  ->  set_sensitive (show_only_new);

        if (m_selection_list_albums.size () == 1)
          {
            Gtk::TreeModel::Path path = m_selection_list_albums.begin()->second.get_path();

            //FIXME: We should possibly allow for removal of several albums in a run
            m_library_actions[APPROVE]->get_action (LIBRARY_UI_ACTION_APPROVE_SELECTED)->set_sensitive (show_only_new);

#ifdef HAVE_HAL
            m_library_actions[ALBUM]->get_action (LIBRARY_UI_ACTION_RELOCATE_ALBUM) ->  set_sensitive (hal->is_initialized());
#if 0
            m_library_actions[ALBUM]->get_action (LIBRARY_UI_ACTION_IMPORT_TRACKS)  ->  set_sensitive (hal->is_initialized());
#endif

#else
            m_library_actions[ALBUM]->get_action (LIBRARY_UI_ACTION_RELOCATE_ALBUM) ->  set_sensitive (true);
#if 0
            m_library_actions[ALBUM]->get_action (LIBRARY_UI_ACTION_IMPORT_TRACKS)  ->  set_sensitive (true);
#endif

#endif

            //FIXME: This is a little awkward as we do an iteration below anyway
            //FIXME: We should possibly allow for fetching of several albums in a run
            TreeModel::iterator cur = store_albums_filter->get_iter (path);

            AlbumFlags flags = (*cur)[albums.flags];
            bool has_cover = (flags & A_HAS_COVER); 

            if (has_cover)
              {
                RefPtr<Gdk::Pixbuf> cover = (*cur)[albums.cover]; 
                view_albums->set_pixbuf (cover);
              }
            else
              {
                view_albums->set_pixbuf (m_viewsize?m_disc_small:m_disc);
              }

            view_albums->set_artist ((*cur)[albums.artist]);
            view_albums->set_album  ((*cur)[albums.album]);
            view_albums->set_source ();

            ustring label_text;

            label_text    .append ("<b>")
                          .append (Markup::escape_text ((*cur)[albums.artist]))
                          .append (":</b> <i>")
                          .append (Markup::escape_text ((*cur)[albums.album]))
                          .append ("</i>");

            label_current_album->set_markup (label_text);
          }
        else
          {
            m_library_actions[APPROVE]->get_action (LIBRARY_UI_ACTION_APPROVE_SELECTED)->set_sensitive (false);
            store_tracks->set_sort_column (Gtk::TreeSortable::DEFAULT_UNSORTED_COLUMN_ID, Gtk::SORT_ASCENDING);
            view_albums->clear_artist_album ();
            view_albums->unset_source ();
            label_current_album->set_markup (_("(Multiple Albums)"));
          }

        string id_artist            = library->get_metadatum_id (DATUM_ARTIST);
        string id_artist_sort_name   = library->get_metadatum_id (DATUM_MB_ALBUM_ARTIST_SORTNAME);
        string id_album             = library->get_metadatum_id (DATUM_ALBUM);
        string id_new_item          = library->get_metadatum_id (DATUM_NEW_ITEM);

        for (AlbumsRowSelectionList::const_iterator pair  = m_selection_list_albums.begin(),
                                                    end   = m_selection_list_albums.end();
                                                    pair != end; ++pair)
          {
            AlbumRowSelectionPair const& selection_pair = (*pair);

            TreeModel::Path path = selection_pair.second.get_path();  
            TreeModel::iterator cur = store_albums_filter->get_iter (path);

            ustring key_A;
            ustring key_B;
  
            AlbumFlags flags = (*cur)[albums.flags];
            bool has_artist_sort = (flags & A_HAS_SORTNAME); 

            key_A = (*cur)[albums.artist];
            key_B = (*cur)[albums.album];

            // Library query
            AttributeList attrs; 
            VRow vector;
            attrs.push_back (Attribute (DB::EXACT, has_artist_sort ? id_artist_sort_name : id_artist,   key_A));
            attrs.push_back (Attribute (DB::EXACT, id_album, key_B));
            attrs.push_back (Attribute (DB::EXACT, id_new_item, show_only_new));
            library->query (attrs, vector);
  
            // Fill tracklist
            fill_tracks_from_rows (vector);

#ifdef HAVE_HAL
            if (hal->is_initialized() && !all_tracks_present())
              {
                m_library_actions[ALBUM]->set_sensitive (false);
              }
            else
              {
                m_library_actions[ALBUM]->set_sensitive (true);
              }
#else // !HAVE_HAL
                m_library_actions[ALBUM]->set_sensitive (true);
#endif

          }
      }

      //////////////////////////// Album Modifications 
      void
      LibraryUI::drop_items ()
      {
        Gtk::MessageDialog dialog (*this,
                                  _("This will remove <b>all</b> remaining new items. Do you really wish to proceed?\n"
                                    "<small>To remove just one or a few albums, use the regular 'Remove Album' option.</small>"),
                                   true, Gtk::MESSAGE_WARNING, Gtk::BUTTONS_YES_NO, true);

        if (dialog.run() == GTK_RESPONSE_YES)
          {
            library->drop_items ();
            albums_update ();
          }
      }

      void
      LibraryUI::approve_selected ()
      {
        using namespace Bmp::DB;
        using namespace Bmp::Library;
        using namespace Gtk;
        using namespace Glib;
        using namespace std;

        TreeSelection::ListHandle_Path paths = view_albums->get_selection()->get_selected_rows ();
        TreeModel::iterator m_Iter = store_albums_filter->convert_iter_to_child_iter (store_albums_filter->get_iter (*(paths.begin()))); 

        AlbumFlags flags = (*m_Iter)[albums.flags];
        bool has_artist_sort = (flags & A_HAS_SORTNAME); 

        ustring key_A = (*m_Iter)[albums.artist];
        ustring key_B = (*m_Iter)[albums.album];

        library->approve_items (key_A, key_B, has_artist_sort);
        store_albums->erase (m_Iter);
      }

      void
      LibraryUI::approve_all ()
      {
        library->approve_items ();
        albums_update ();
      }

      void
      LibraryUI::remove_albums_action_cb ()
      {
        remove_albums (false);
      }

      void
      LibraryUI::remove_selected_album (Glib::ustring const& key_A, Glib::ustring const& key_B, Gtk::TreeModel::iterator const& m_Iter)
      {
        RowMapAlbums::iterator rm_Iter = m_row_map_albums.find (RowKeyAlbum (key_A, key_B));
        if (rm_Iter != m_row_map_albums.end ())
          {
            m_row_map_albums.erase (rm_Iter);
            store_albums->erase (m_Iter);
          }
      }

      void
      LibraryUI::remove_albums (bool force)
      {
        using namespace Bmp::DB;
        using namespace Bmp::Library;
        using namespace Glib;
        using namespace Gtk;
        using namespace std;

        PathList paths = view_albums->get_selection()->get_selected_rows ();
        ReferenceList list;

        for (PathList::const_iterator path  = paths.begin (),
                                   end   = paths.end ();
                                   path != end; ++path)
        {
          list.push_back (Gtk::TreeRowReference (store_albums, store_albums_filter->convert_path_to_child_path (*path)));
        }

        for (ReferenceList::const_iterator ref  = list.begin (),
                                           end  = list.end ();
                                           ref != end; ++ref)
        {
            TreeModel::iterator m_Iter = store_albums->get_iter (ref->get_path());
            AttributeList attrs;

            string id_artist            = library->get_metadatum_id (DATUM_ARTIST);
            string id_artist_sort_name  = library->get_metadatum_id (DATUM_MB_ALBUM_ARTIST_SORTNAME);
            string id_album             = library->get_metadatum_id (DATUM_ALBUM);

            AlbumFlags flags = (*m_Iter)[albums.flags];
            bool has_artist_sort = (flags & A_HAS_SORTNAME); 

            ustring key_A = (*m_Iter)[albums.artist];
            ustring key_B = (*m_Iter)[albums.album];

            attrs.push_back (Attribute (DB::EXACT, has_artist_sort ? id_artist_sort_name : id_artist, key_A));
            attrs.push_back (Attribute (DB::EXACT, id_album, key_B));

            Bmp::DB::VRow vector;
  
            library->query (attrs, vector);

            if (force && !vector.empty()) continue;
      
            if (!force)
              {
                ustring message;
                message.append (_("Do you really want to remove the album"));
                message.append (" '").append (key_B).append("' ");
                message.append (_("from the library?"));

                Gtk::MessageDialog dialog (*this, message, false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_YES_NO, true);
                dialog.set_title (_("Remove Album - BMP"));

                if (dialog.run () == Gtk::RESPONSE_YES)
                  {
                    library->del (attrs);
                    remove_selected_album (key_A, key_B, m_Iter);
                  }
              }
            else
              {
                library->del (attrs);
                remove_selected_album (key_A, key_B, m_Iter);
              }
        }
      }
  
      void
      LibraryUI::remove ()
      {
        using namespace Bmp::DB;
        using namespace Bmp::Library;
        using namespace Glib;
        using namespace Gtk;
        using namespace std;

        bool physically_delete_files = mcs->key_get<bool>("bmp", "physically-delete-files");
        if (physically_delete_files)
        {
          Gtk::MessageDialog dialog (*this,
                                     _("<b>You are about to delete the selected files also <i>physically</i> from the filesystem.\n"
                                       "If you do NOT wish that, please press cancel now.</b>\n\n<small>NOTE: You can change this "
                                       "option in Preferences, section 'Library'.</small>"),
                                      true, Gtk::MESSAGE_WARNING, Gtk::BUTTONS_OK_CANCEL, true);
          if (dialog.run() != Gtk::RESPONSE_OK)
          {
            dialog.hide ();
            return;
          }
        }

        IterList iters;

        TreeSelection::ListHandle_Path paths = view_tracks->get_selection()->get_selected_rows ();
        for (TreeSelection::ListHandle_Path::const_iterator iter = paths.begin (); iter != paths.end (); ++iter)
          {
            iters.push_back (store_tracks->get_iter (*iter));
          }

        for (IterList::const_iterator cur = iters.begin (), end = iters.end (); cur != end; ++cur)
          {
            ustring key = (*(*cur))[tracks.location];
            AttributeList attrs;
            attrs.push_back (Attribute (DB::EXACT, library->get_metadatum_id (DATUM_LOCATION), string(key)));
            library->del (attrs);
            store_tracks->erase (*cur);
            if (physically_delete_files)
              {
                std::string filename = filename_from_uri (key);
                g_unlink (filename.c_str());    
              }
          }

        remove_albums (true);
      }

      //////////////////////////// Play/Enqueue
      void
      LibraryUI::play ()
      {
        using namespace Glib;
        using namespace Gtk;

        RefPtr<TreeSelection> sel = view_tracks->get_selection();

        if (!sel->count_selected_rows ()) sel->select_all(); 

        TreeSelection::ListHandle_Path paths = sel->get_selected_rows ();
        char **uri_list = (char**)calloc (paths.size()+1, sizeof(char*));
        unsigned int counter = 0;
        for (TreeSelection::ListHandle_Path::const_iterator iter = paths.begin (); iter != paths.end (); ++iter)
          {
            TreeModel::iterator cur = store_tracks->get_iter (*iter);
#ifdef HAVE_HAL
            if ((*cur)[tracks.present])
              {
#endif
                ustring str = (*cur)[tracks.location];
                uri_list[counter++] = strdup(str.c_str ());
#ifdef HAVE_HAL
              }
#endif
          }
        uri_list[paths.size()] = 0;

        int out;
        bmp_system_control_add_uri_list (bmp_system_control,
                         (const char**)uri_list,
                         -1,
                         true, 
                         true, 
                         0, 
                         &out,
                         NULL);
        g_strfreev (uri_list);
        sel->unselect_all ();
      }

      void
      LibraryUI::enqueue ()
      {
        using namespace Glib;
        using namespace Gtk;

        RefPtr<TreeSelection> sel = view_tracks->get_selection();
        if (!sel->count_selected_rows ()) sel->select_all(); 

        TreeSelection::ListHandle_Path paths = sel->get_selected_rows ();
        char **uri_list = (char**)calloc (paths.size()+1, sizeof(char*));
        unsigned int counter = 0;
        for (TreeSelection::ListHandle_Path::const_iterator iter = paths.begin (); iter != paths.end (); ++iter)
          {
            TreeModel::iterator cur = store_tracks->get_iter (*iter);
#ifdef HAVE_HAL
            if ((*cur)[tracks.present])
              {
#endif
                ustring str = (*cur)[tracks.location];
                uri_list[counter++] = strdup (str.c_str ());
#ifdef HAVE_HAL
              }
#endif
          }
        uri_list[paths.size()] = 0;

        int out;
        bmp_system_control_add_uri_list (bmp_system_control,
                         (const char**)uri_list,
                         -1,
                         false, 
                         false, 
                         -1, 
                         &out,
                         NULL);
        g_strfreev (uri_list);
        sel->unselect_all ();
      }

      //////////////////////////// Adding/Modifications 
      void
      LibraryUI::import_tracks_as_album ()
      {
        using namespace Bmp::DB;
        using namespace Bmp::Library;
        using namespace Gtk;
        using namespace Gdk;
        using namespace Glib;
        using namespace std;
        using namespace Bmp::VFS;

        set_sensitive (false);

        Gtk::FileChooserDialog dialog (*this, _("Select Tracks to Import as Album - BMP"), Gtk::FILE_CHOOSER_ACTION_OPEN);
        dialog.set_select_multiple (true);
        dialog.add_button (Gtk::StockID (Gtk::Stock::CANCEL), Gtk::RESPONSE_CANCEL);
        dialog.add_button (Gtk::StockID (Gtk::Stock::OPEN), Gtk::RESPONSE_OK);
        dialog.set_default_response (Gtk::RESPONSE_CANCEL);
        dialog.set_current_folder (mcs->key_get<std::string>("bmp", "file-chooser-path"));

        if (dialog.run () == Gtk::RESPONSE_OK)
          {
            dialog.hide ();
            mcs->key_set<std::string>("bmp", "file-chooser-path", dialog.get_current_folder());

            SListHandle<ustring> handle = dialog.get_uris (); 
            UriList urilist (handle.begin(), handle.end());

            if (urilist.empty()) 
              {
                set_sensitive (true);
                return;
              }

            // Check for writability
            Util::FileList list, non_writable;
            for (UriList::const_iterator iter  = urilist.begin (), end = urilist.end ();
                                         iter != end; ++iter)
              {
                list.push_back (filename_from_uri(*iter));
              }

            Util::files_writable (list, non_writable);
            if (!non_writable.empty ())
              {
                DialogFilelist *dialog = Bmp::DialogFilelist::create ();
                dialog->run (non_writable);
                delete dialog;
                set_sensitive (true);
                return;
              }

            // Check for taggability
            if (!Util::files_taggable (list))
              {
                Gtk::MessageDialog dialog (*this,
                                           _("Unable to modify: Some of the selected/specified files are currently not taggable with BMP"),
                                           false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
                dialog.run ();
                set_sensitive (true);
                return;
              }

            MusicBrainzXML::MusicBrainzReleases mb_releases; 
            LibraryUIModifyAlbums *dialog = LibraryUIModifyAlbums::create ();
            if (dialog->run (mb_releases) != Gtk::RESPONSE_OK)
              {
                delete dialog;
                set_sensitive (true);
                return;
              }

            MusicBrainzXML::MusicBrainzRelease selected = dialog->selected;
            delete dialog;

            m_busy = true;
            int n_mod_tracks = urilist.size();
            int n_mod_current = 0;
            progress_modify->show_all ();
            progress_modify->set_fraction (0.0);
            progress_modify->set_text ((progress_text % 0 % n_mod_tracks).str());

            for (UriList::const_iterator  iter  = urilist.begin (), end = urilist.end ();
                                          iter != end; ++iter)
              {
                UpdateTrack update_track;

                update_track.mb_artist_id     = selected.mb_artist_id;
                update_track.artist           = selected.mb_artist_name;
                update_track.mb_album_id      = selected.mb_release_id;
                update_track.album            = selected.mb_release_name;

                if (!selected.asin.empty()) update_track.asin = selected.asin;
            
                try
                  { 
                      library->metadata_set_taglib ((*iter), update_track); 
                      library->add_track (*iter);
                  }
                catch (Bmp::Library::Exception& cxe) {}

                  n_mod_current++;
                  progress_modify->set_fraction ((n_mod_current*1.)/(n_mod_tracks*1.));
                  progress_modify->set_text ((progress_text % n_mod_current % n_mod_tracks).str());

                while (gtk_events_pending()) gtk_main_iteration ();
              }

            progress_modify->set_fraction (0.0);
            progress_modify->set_text ("");
            progress_modify->hide ();

            library->update_compilations ();
            m_busy = false;
            set_sensitive (true);
            albums_update ();
          }
        else
          {
            set_sensitive (true);
          } 
      }

      bool
      LibraryUI::import_tracks_as_album_X (const Bmp::VFS::UriList& urilist)
      {
        using namespace Bmp::DB;
        using namespace Bmp::Library;
        using namespace Gtk;
        using namespace Gdk;
        using namespace Glib;
        using namespace std;
        using namespace Bmp::VFS;

        show ();
        set_sensitive (false);

        if (urilist.empty()) 
          {
            set_sensitive (true);
            return false;
          }

        // Check for writability
        Util::FileList list, non_writable;
        for (UriList::const_iterator iter  = urilist.begin (), end = urilist.end ();
                                     iter != end; ++iter)
          {
            list.push_back (filename_from_uri(*iter));
          }

        Util::files_writable (list, non_writable);
        if (!non_writable.empty ())
          {
            DialogFilelist *dialog = Bmp::DialogFilelist::create ();
            dialog->run (non_writable);
            delete dialog;
            set_sensitive (true);
            return false;
          }

        // Check for taggability
        if (!Util::files_taggable (list))
          {
            Gtk::MessageDialog dialog (*this,
                                       _("Unable to modify: Some of the selected/specified files are currently not taggable with BMP"),
                                       false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
            dialog.run ();
            set_sensitive (true);
            return false;
          }

        MusicBrainzXML::MusicBrainzReleases mb_releases; 
        LibraryUIModifyAlbums *dialog = LibraryUIModifyAlbums::create ();
        if (dialog->run (mb_releases) != Gtk::RESPONSE_OK)
          {
            delete dialog;
            set_sensitive (true);
            return false;
          }

        MusicBrainzXML::MusicBrainzRelease selected = dialog->selected;
        delete dialog;

        m_busy = true;
        int n_mod_tracks = urilist.size();
        int n_mod_current = 0;
        progress_modify->show_all ();
        progress_modify->set_fraction (0.0);
        progress_modify->set_text ((progress_text % 0 % n_mod_tracks).str());

        for (UriList::const_iterator  iter  = urilist.begin (), end = urilist.end ();
                                      iter != end; ++iter)
          {
            UpdateTrack update_track;

            update_track.mb_artist_id     = selected.mb_artist_id;
            update_track.artist           = selected.mb_artist_name;
            update_track.mb_album_id      = selected.mb_release_id;
            update_track.album            = selected.mb_release_name;

            if (!selected.asin.empty()) update_track.asin = selected.asin;
        
            try {
                library->metadata_set_taglib ((*iter), update_track); 
                library->add_track (*iter);
              }
            catch (Bmp::Library::Exception& cxe) {} //FIXME: 

            n_mod_current++;
            progress_modify->set_fraction ((n_mod_current*1.)/(n_mod_tracks*1.));
            progress_modify->set_text ((progress_text % n_mod_current % n_mod_tracks).str());

            while (gtk_events_pending()) gtk_main_iteration ();
          }

        progress_modify->set_fraction (0.0);
        progress_modify->set_text ("");
        progress_modify->hide ();

        library->update_compilations ();

        m_busy = false;
        set_sensitive (true);

        if (!Glib::RefPtr<Gtk::ToggleAction>::cast_static(m_library_actions[APPROVE]->get_action (LIBRARY_UI_ACTION_SHOW_ONLY_NEW))->get_active ())
          {
            Glib::RefPtr<Gtk::ToggleAction>::cast_static(m_library_actions[APPROVE]->get_action (LIBRARY_UI_ACTION_SHOW_ONLY_NEW))->set_active ();
          }
        else
          {
            albums_update ();
          }
        return true;
      }

#if 0
      void
      LibraryUI::import_tracks_into_album ()
      { 
        using namespace Bmp::DB;
        using namespace Bmp::Library;
        using namespace Gtk;
        using namespace Gdk;
        using namespace Glib;
        using namespace std;
        using namespace Bmp::VFS;

        Gtk::FileChooserDialog dialog (*this, _("Select Tracks to Import into selected Album - BMP"), Gtk::FILE_CHOOSER_ACTION_OPEN);
        dialog.set_select_multiple (true);
        dialog.add_button (Gtk::StockID (Gtk::Stock::CANCEL), Gtk::RESPONSE_CANCEL);
        dialog.add_button (Gtk::StockID (Gtk::Stock::OPEN), Gtk::RESPONSE_OK);
        dialog.set_default_response (Gtk::RESPONSE_CANCEL);
        dialog.set_current_folder (mcs->key_get<std::string>("bmp", "file-chooser-path"));

        set_sensitive (false);
  
        if (dialog.run () == Gtk::RESPONSE_OK)
          {
            dialog.hide ();

            SListHandle<ustring> handle = dialog.get_uris (); 
            UriList urilist (handle.begin(), handle.end());

            // Check for writability
            Util::FileList list, non_writable;
            for (UriList::const_iterator iter  = urilist.begin (),
                                              end   = urilist.end ();
                                              iter != end; ++iter)
              {
                list.push_back (filename_from_uri(*iter));
              }

            Util::files_writable (list, non_writable);
            if (!non_writable.empty ())
              {
                DialogFilelist *dialog = Bmp::DialogFilelist::create ();
                dialog->run (non_writable);
                delete dialog;
                set_sensitive (true);
                return;
              }

            // Check for taggability
            if (!Util::files_taggable (list))
              {
                Gtk::MessageDialog dialog (*this,
                                           _("Unable to modify: Some of the selected/specified files are currently not taggable with BMP"),
                                           false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
                dialog.run ();
                set_sensitive (true);
                return;
              }

            mcs->key_set<std::string>("bmp", "file-chooser-path", dialog.get_current_folder());
            for (UriList::const_iterator iter  = urilist.begin (),
                                         end   = urilist.end ();
                                         iter != end; ++iter)
              {
                library->add_track (*iter);
                while (gtk_events_pending ()) gtk_main_iteration ();
              }

              PathList paths = view_albums->get_selection()->get_selected_rows ();
              Gtk::TreePath path = paths[0];
              TreeModel::iterator cur = store_albums_filter->convert_iter_to_child_iter (store_albums_filter->get_iter (path));

              ustring artist = (*cur)[albums.artist];
              ustring album  = (*cur)[albums.album];

              bool is_new = library->is_new (artist, album);

              boost::optional<ustring> asin;
              try { asin = library->get_asin_for_album (artist, album); } catch (Bmp::Library::Exception& cxe) {}

              UpdateTracks update_tracks;
              for (UriList::const_iterator  iter  = urilist.begin (),
                                            end   = urilist.end ();
                                            iter != end; ++iter)
                {
                  UpdateTrack update_track;

                  update_track.location_cur   = (*iter); 
                  update_track.artist         = artist;
                  update_track.album          = album;
                  update_track.new_item       = is_new;

                  if (asin) update_track.asin = asin;

                  update_tracks.push_back (update_track); 
                }

              m_busy = true;
              library->modify_tracks (update_tracks);
              library->update_compilations ();
              m_busy = false;

              update_track_view_from_albums ();
          }
        set_sensitive (true);
      }
#endif
 
      void
      LibraryUI::relocate_album ()
      {
        using namespace Bmp::DB;
        using namespace Bmp::Library;
        using namespace Gtk;
        using namespace Gdk;
        using namespace Glib;
        using namespace std;

        TreeSelection::ListHandle_Path paths = view_albums->get_selection()->get_selected_rows ();
        if (paths.size () == 1)
          {
            TreeModel::iterator cur = store_albums_filter->get_iter (*(paths.begin())); 
            LibraryUIMoveAlbum *dialog = LibraryUIMoveAlbum::create ();
            set_sensitive (false);

            int response = dialog->run ((*cur)[albums.artist], (*cur)[albums.album]);
            dialog->hide ();

            if (response == Gtk::RESPONSE_OK)
              {
                std::string path_new;
                ustring uri = dialog->fcb_move->get_current_folder_uri ();

                if (dialog->cb_append_albumname->get_active())
                  {
                    std::string path = filename_from_uri (uri);
                    std::string append = filename_from_utf8 ((*cur)[albums.album]);
                    path_new  = build_filename (path, append);

                    g_mkdir (path_new.c_str(), mode_t(S_IRWXU)); 
                  }
                else
                  {
                    path_new  = filename_from_uri (uri);
                  }

                UpdateTracks update_tracks;
                for (VRow::const_iterator i = dialog->vector.begin (), end = dialog->vector.end(); i != end; ++i)
                  {
                    try
                      {
                        Row const& row = *i;
                        Track track (row);

                        std::string filename = filename_from_uri (track.location.get());
                        std::string basename = path_get_basename (filename);
                        std::string pathname = path_get_dirname  (filename);

                        if (pathname != path_new) 
                          {
                            UpdateTrack update_track;
                            update_track.location_cur = track.location.get(); 
                            update_track.location_new = filename_to_uri (build_filename (path_new, basename));
                            update_tracks.push_back (update_track); 
                          }
                      }
                    catch (ConvertError& cxe)
                      {
                        Gtk::MessageDialog message_dialog (*this,
                          _("Unable to proceed: A conversion error occured during processing"),
                          false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
                        message_dialog.run ();
                        set_sensitive (true);
                        delete dialog;
                        return;
                      }
                  }
                library->modify_tracks (update_tracks, false);
                update_track_view_from_albums ();
              }
            set_sensitive (true);
            delete dialog;
          }
      }

      void
      LibraryUI::albums_modify ()
      {
        using namespace std;
        using namespace Gtk;
        using namespace Gdk;
        using namespace Glib;
        using namespace Bmp::DB;
        using namespace Bmp::Library;
        using namespace Bmp::MusicBrainzXML;


        Util::FileList list;
        TreeModel::Children nodes = store_tracks->children ();
        for (TreeModel::Children::const_iterator node  = nodes.begin (),
                                                 end   = nodes.end (); 
                                                 node != end; ++node)
          {
            ustring uri = (*node)[tracks.location];
            list.push_back (filename_from_uri (uri));
          }

        Util::FileList non_writable;
        Util::files_writable (list, non_writable);

        if (!non_writable.empty ())
          {
            DialogFilelist *dialog = Bmp::DialogFilelist::create ();
            dialog->run (non_writable);
            delete dialog;
            return;
          }

        if (!Util::files_taggable (list))
          {
            MessageDialog dialog (*this, 
                  _("Unable to modify: Some of the selected/specified files are currently not taggable with BMP"),
                  false, MESSAGE_ERROR, BUTTONS_OK, true);
            dialog.run ();
            return;
          }

        set_sensitive (false);

        MusicBrainzReleases releases; 
        PathList paths = view_albums->get_selection()->get_selected_rows ();
        for (PathList::const_iterator path  = paths.begin(),
                                      end   = paths.end();
                                      path != end; ++path)
          {
            TreeModel::iterator m_Iter = store_albums_filter->convert_iter_to_child_iter (store_albums_filter->get_iter (*path)); 
            ustring artist = (*m_Iter)[albums.artist];
            ustring album  = (*m_Iter)[albums.album];
            get_releases (artist, album, releases);
          }

        Bmp::VFS::UriList locations;
        TreeNodeChildren track_nodes = store_tracks->children ();
        for (TreeNodeChildren::const_iterator m_Iter = nodes.begin(), end = nodes.end(); m_Iter != end; ++m_Iter)
          {
            locations.push_back ((*m_Iter)[tracks.location]);
          }

        LibraryUIModify *modify = LibraryUIModify::create ();

        if ( modify->run (releases, locations) == GTK_RESPONSE_OK )
          {
            modify->hide();

            if (modify->modify_tracks.update_tracks.size() < list.size())
              {
                boost::format msg (_("The number of tracks in the selected release is only <b>'%s'</b>, while you have selected <b>'%s'</b> "
                                     "from your collection. Are you sure you want to proceed?\n\n<small>Note: This might be actually valid "
                                     "if you e.g. have duplicate tracks, but you might want to check again if this is the correct release.</small>"));

                MessageDialog message_dialog (*this,
                      (msg % modify->modify_tracks.update_tracks.size() % list.size()).str(), true, MESSAGE_QUESTION, BUTTONS_YES_NO, true);

                if ( message_dialog.run () == RESPONSE_NO )
                  {
                    set_sensitive (true);
                    delete modify;
                    return;
                  }
              }
            else if ( modify->modify_tracks.update_tracks.size() > list.size() )
              {
                boost::format msg (_("The number of tracks in the selected release is <b>'%s'</b>, while you have selected only <b>'%s'</b> from "
                                     "your collection. Are you sure you want to proceed?\n\n<small>Note: This might be actually valid if you e.g. "
                                     "don't have all tracks from this album currently.</small>"));

                MessageDialog message_dialog (*this,
                      (msg % modify->modify_tracks.update_tracks.size() % list.size()).str(), true, MESSAGE_QUESTION, BUTTONS_YES_NO, true);

                if ( message_dialog.run () == RESPONSE_NO )
                  {
                    set_sensitive (true);
                    delete modify;
                    return;
                  }
              }

            m_busy = true;
            library->modify_tracks (modify->modify_tracks.update_tracks);

            boost::optional<ustring> sort_name;
            sort_name = modify->modify_tracks.mb_album_artist_sort_name;

            ReferenceList list;
            for (PathList::const_iterator path  = paths.begin (),
                                          end   = paths.end ();
                                          path != end; ++path)
              {
                list.push_back (TreeRowReference (store_albums, store_albums_filter->convert_path_to_child_path (*path)));
              }

            for (ReferenceList::const_iterator ref   = list.begin (),
                                               end   = list.end ();
                                               ref  != end; ++ref)
              {
                TreeModel::iterator m_Iter = store_albums->get_iter (ref->get_path());
                ustring key_A = (*m_Iter)[albums.artist];
                ustring key_B = (*m_Iter)[albums.album];
                remove_selected_album (key_A, key_B, m_Iter);
              }

            // Before we insert, make sure we remove entries that possibly match our new entry, or we'll end up with dupes    
            ustring key_A = sort_name ? sort_name.get() : modify->modify_tracks.mb_album_artist;
            ustring key_B = modify->modify_tracks.mb_album_name;
            RowMapAlbums::iterator rm_Iter = m_row_map_albums.find (RowKeyAlbum (key_A, key_B));
            if (rm_Iter != m_row_map_albums.end ())
              {
                TreeModel::RowReference& ref = rm_Iter->second;
                TreeModel::iterator m_Iter = store_albums->get_iter (ref.get_path ());
                remove_selected_album (key_A, key_B, m_Iter);
              }

            TreeModel::iterator m_Iter = store_albums->append ();
            (*m_Iter)[albums.flags] = AlbumFlags (0);
            AlbumFlags flags = (*m_Iter)[albums.flags];

            RefPtr<Pixbuf> cover;
            if (modify->modify_tracks.asin)
              {
                try {
                    Amazon::get_cover_by_asin (modify->modify_tracks.asin.get(), cover, false);
                    (*m_Iter)[albums.cover] = cover->scale_simple (m_viewsize?32:64, m_viewsize?32:64, INTERP_BILINEAR);
                    (*m_Iter)[albums.flags] = AlbumFlags (flags | A_HAS_COVER);
                  }
                catch (...)
                  {
                    (*m_Iter)[albums.cover] = m_viewsize?m_disc_small:m_disc;
                    (*m_Iter)[albums.flags] = AlbumFlags (flags & ~A_HAS_COVER);
                  }
              }
            else
              {
                (*m_Iter)[albums.cover] = m_viewsize?m_disc_small:m_disc;
                (*m_Iter)[albums.flags] = AlbumFlags (flags & ~A_HAS_COVER);
              }

            ustring title;
            title   .append ("<b>")
                    .append (Markup::escape_text ((sort_name ? sort_name.get() : modify->modify_tracks.mb_album_artist)))
                    .append ("</b>\n")
                    .append ("<i>")
                    .append (Markup::escape_text (modify->modify_tracks.mb_album_name))
                    .append ("</i>");


            (*m_Iter)[albums.title]  = title;
            (*m_Iter)[albums.album]  = modify->modify_tracks.mb_album_name;
            (*m_Iter)[albums.artist] = (sort_name ? sort_name.get() : modify->modify_tracks.mb_album_artist);
            (*m_Iter)[albums.flags]  = AlbumFlags (flags | (sort_name ? A_HAS_SORTNAME : 0));

            m_row_map_albums.insert (std::make_pair (RowKeyAlbum (sort_name ? sort_name.get() : modify->modify_tracks.mb_album_artist,
                                                                                                modify->modify_tracks.mb_album_name),
                                                     TreeModel::RowReference (store_albums, TreePath (m_Iter))));

            m_busy = false;

            store_albums_filter->clear_cache ();
            store_albums_filter->refilter ();
            update_track_view_from_albums ();

            view_albums->get_selection()->select (store_albums_filter->convert_child_iter_to_iter (m_Iter));
          }

        delete modify;
        set_sensitive (true);
      }

      void
      LibraryUI::import_tracks ()
      {
        using namespace Gtk;
        using namespace Glib;

        FileChooserDialog dialog (*this,
                                  _("Select Path To Add - BMP"),
                                  FILE_CHOOSER_ACTION_SELECT_FOLDER);

        dialog.add_button (Stock::CANCEL, RESPONSE_CANCEL);
        dialog.add_button (Stock::ADD, RESPONSE_OK);
        dialog.set_current_folder (mcs->key_get<std::string>("bmp", "file-chooser-path"));
        dialog.set_default_response (RESPONSE_CANCEL);
        dialog.set_default_size (750, 570);
        dialog.set_position (Gtk::WIN_POS_CENTER);

        set_sensitive (false);

        if (dialog.run () == RESPONSE_OK)
          {
            mcs->key_set<std::string>("bmp", "file-chooser-path", dialog.get_current_folder());
            notebook_add->set_current_page (1);
            Glib::ustring folder = dialog.get_filename ();
            dialog.hide ();


            Util::FileList list;
            Util::collect_path (folder, list);
            m_busy = true; 
            for (Util::FileList::const_iterator iter  = list.begin (),
                                                end   = list.end ();
                                                iter != end; ++iter)
              {
                try
                  {
                    debug ("library-ui", "Adding URI '%s' to Library", (*iter).c_str()); 
                    library->add_track (Glib::filename_to_uri (*iter));
                    while (gtk_events_pending ()) gtk_main_iteration ();
                  }
                catch (Glib::ConvertError& cxe) {} // FIXME:
              }

            library->update_compilations ();
            m_busy = false; 

            notebook_add->set_current_page (0);
            set_sensitive (true);

            if (!Glib::RefPtr<Gtk::ToggleAction>::cast_static(m_library_actions[APPROVE]->get_action (LIBRARY_UI_ACTION_SHOW_ONLY_NEW))->get_active ())
              {
                Glib::RefPtr<Gtk::ToggleAction>::cast_static(m_library_actions[APPROVE]->get_action (LIBRARY_UI_ACTION_SHOW_ONLY_NEW))->set_active ();
              }
            else
              {
                albums_update ();
              }
          }
        else
          {
            set_sensitive (true);
          }
      }

      void
      LibraryUI::albums_refilter ()
      {
        store_albums_filter->refilter ();
      }

      bool
      LibraryUI::albums_row_is_visible (const Gtk::TreeModel::const_iterator& iter)
      {
        using namespace Glib;
        using namespace Bmp::Util;

        ustring artist ((*iter)[albums.artist]);  
        ustring album  ((*iter)[albums.album]);   
        ustring needle (albums_filter->get_text ());

        return (needle.empty() || match_keys (artist + " " + album, needle));
      }

      void
      LibraryUI::albums_update ()
      {
          using namespace Bmp::DB;
          using namespace Bmp::Library;
          using namespace Gtk; 
          using namespace Gdk; 
          using namespace Glib; 
          using namespace boost;
          using namespace std;

          if (this->m_init) return;

          notebook_albums->set_current_page (1);

          view_albums->unset_model ();

          store_albums->clear ();
          m_row_map_albums.clear ();

          m_library_actions[IMPORT]->set_sensitive (false);
          m_library_actions[APPROVE]->set_sensitive (false);
          cb_view_size->set_sensitive (false);
          albums_filter->set_sensitive (false);

          unsigned int uid = 0;
          bool show_only_new = 
              Glib::RefPtr<Gtk::ToggleAction>::cast_static (m_library_actions[APPROVE]->get_action (LIBRARY_UI_ACTION_SHOW_ONLY_NEW))->get_active();

          VRow vector;
          library->get_albums (vector, show_only_new);

          for (VRow::iterator i = vector.begin(), e = vector.end(); i != e; ++i)
            {
              Bmp::Library::Album album (*i);

              if (album.artist && album.album)
                {
                  boost::optional<ustring> sort_name;
                  try { sort_name = library->get_sort_name_for_album (album.artist.get(), album.album.get()); } catch (...) {}

                  RowKeyAlbum key (sort_name ? sort_name.get() : album.artist .get(),
                                                               album.album  .get());

                  RowMapAlbums::iterator mapping = m_row_map_albums.find (key); 
                  if (mapping != m_row_map_albums.end())
                    {
                      continue;
                    }

                  TreeModel::iterator cur = cur = store_albums->append ();
                  (*cur)[albums.flags] = A_NONE;

                  AlbumFlags flags = (*cur)[albums.flags];

                  ustring title;
                  title .append ("<b>")
                        .append (Markup::escape_text (sort_name ? sort_name.get() : album.artist.get()))
                        .append ("</b>\n")
                        .append ("<i>") 
                        .append (Markup::escape_text (album.album.get()))
                        .append ("</i>");

                  TreePath path (cur);
                  RefPtr<Pixbuf> cover;
              
                  if (album.asin)
                    {
                      try {
                          Amazon::get_cover_by_asin (album.asin.get(), cover, true);
                          (*cur)[albums.cover] = cover->scale_simple (m_viewsize?32:64, m_viewsize?32:64, INTERP_BILINEAR);
                          (*cur)[albums.flags] = AlbumFlags (flags | A_HAS_COVER);
                      } catch (...) {
                          (*cur)[albums.cover] = m_viewsize?m_disc_small:m_disc;
                          (*cur)[albums.flags] = AlbumFlags (flags & ~A_HAS_COVER);
                      }
                    }
                  else
                    {
                      (*cur)[albums.cover] = m_viewsize?m_disc_small:m_disc;
                      (*cur)[albums.flags] = AlbumFlags (flags & ~A_HAS_COVER);
                    }

                  (*cur)[albums.artist]            = sort_name ? sort_name.get() : album.artist.get();
                  (*cur)[albums.album]             = album.album.get();
                  (*cur)[albums.title]             = title;
                  (*cur)[albums.rowuid]            = uid++;

                  if (sort_name)
                    {
                      flags = AlbumFlags (flags | A_HAS_SORTNAME);
                    }
                  else
                    {
                      flags = AlbumFlags (flags &~ A_HAS_SORTNAME);
                    }

                  (*cur)[albums.flags] = flags; 

                  m_row_map_albums.insert (make_pair (key,TreeModel::RowReference (store_albums, TreePath (cur))));

                  while (gtk_events_pending ()) gtk_main_iteration ();
                }
            }

          cb_view_size->set_sensitive (true);
          albums_filter->set_sensitive (true);
          m_library_actions[APPROVE]->set_sensitive (true);
          m_library_actions[APPROVE]->get_action (LIBRARY_UI_ACTION_APPROVE_ITEMS)->set_sensitive (show_only_new);
          m_library_actions[APPROVE]->get_action (LIBRARY_UI_ACTION_DROP_ALL)->set_sensitive (show_only_new);
          m_library_actions[APPROVE]->get_action (LIBRARY_UI_ACTION_APPROVE_SELECTED)->set_sensitive (false);

#ifdef HAVE_HAL
          m_library_actions[IMPORT]->set_sensitive (hal->is_initialized());
#else
          m_library_actions[IMPORT]->set_sensitive (true);
#endif

          notebook_albums->set_current_page (0);
          view_albums->set_model (store_albums_filter);
      }
}
