//  BMPx - The Dumb Music Player
//  Copyright (C) 2005-2006 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non GPL-compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <revision.h>
#include <build.h>

#include "main.hpp"

#include <glibmm.h>
#include <glib-object.h>
#include <glib/gi18n.h>
#include <glib/gprintf.h>
#include <glib/gstdio.h>

#include <gtkmm.h>
#include <gdkmm/wrap_init.h>
#include <gtkmm/wrap_init.h>

#ifdef HAVE_MOOD
#  include <moo/driver.hh>
#endif

#include <mcs/mcs.h>

#include <goa/libgoa.h>
#include <gst/gst.h>
#include <cstdlib>
#include <cstring>

#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#ifdef HAVE_SM
#  include "sm.hpp"
#endif

#ifdef HAVE_STARTUP_NOTIFICATION
#    define SN_API_NOT_YET_FROZEN
#    include <libsn/sn-launchee.h>
#    include <gdk/gdkx.h>
namespace
{
  SnLauncheeContext *sn_context = 0;
  SnDisplay *sn_display = 0;
}
#endif // HAVE_STARTUP_NOTIFICATION

#include "play.hpp"
#include "lastfm.hpp"
#include "signals.hpp"
#include "monitor.hpp"
#include "logger.hpp"

#ifdef HAVE_HAL
# include "hal.hpp"
#endif

#include "plugin.h"
#include "sanity.h"
#include "util.hpp"
#include "file_utils.hpp"
#include "vfs.hpp"

#include "playlist.hpp"
#include "library.hpp"
#include "paths.hpp"
#include "loader.hpp"
#include "debug.hpp"
#include "bookmarks.hpp"

#include <dbus/dbus-gtype-specialized.h>

#ifdef ENABLE_TESTING_CODE 
#include "playlist++.hpp"
#include "service_core.hpp"
#endif

BmpPlaylist                 *bmp_playlist = 0;
BmpSystemControl            *bmp_system_control = 0;

// C++ objects go here
Bmp::FileMonitor::Monitor   *monitor = 0;
Bmp::Play                   *play = 0;
Bmp::Logger                 *logger = 0;
Bmp::LastFM::Service        *lastfm = 0;
Bmp::LastFM::Radio          *lastfm_radio = 0;
Bmp::Library::Library       *library = 0;
Bmp::VFS::VFS               *vfs = 0;
Bmp::Streams::Bookmarks     *bookmarks = 0;
Mcs::Mcs                    *mcs = 0;

#ifdef ENABLE_TESTING_CODE
Bmp::ServiceCore            *core = 0;
#endif

#ifdef HAVE_HAL
Bmp::Library::HAL           *hal = 0;
#endif

#ifdef HAVE_MOOD
Moo::Driver                 *mood = 0;
#endif

#ifdef HAVE_SM
namespace { char *session_id = 0; }
#endif

namespace { bool show_version  = false; }

namespace
{
   bool run_headless    = false;
   bool forced_headless = false;
}

char *bmp_paths[BMP_N_PATHS] = { 0 };
bool  startup = false;

namespace
{
  bool detach = false;
  bool log_stdout = false;
  bool no_resume = false;
  bool no_remote = false;

  GOptionEntry bmp_options[] =
    {
      {"version", 'v', 0, G_OPTION_ARG_NONE, &show_version,
       N_("Display version"), 0},
      {"no-log", 0, 0, G_OPTION_ARG_NONE, &log_stdout,
       N_("Disable logfile, log messages to shell"), 0},
      {"no-ui", 0, 0, G_OPTION_ARG_NONE, &run_headless,
       N_("Disable UI"), 0},
      {"detach", 0, 0, G_OPTION_ARG_NONE, &detach,
       N_("Detach from the tty"), 0},
      {"detach", 0, 0, G_OPTION_ARG_NONE, &detach,
       N_("Detach from the tty"), 0},
      {"no-remote", 0, 0, G_OPTION_ARG_NONE, &no_remote,
       N_("Disable DBUS remote interface"), 0},
#ifdef HAVE_SM
      {"sm-client-id", 0, 0, G_OPTION_ARG_STRING, &session_id,
       N_("Specify session management ID"), N_("ID")},
#endif // HAVE_SM
      { 0 }
    };

  static const char *features[] =
    {

#ifdef HAVE_SM
      N_("X11 Session Management"),
#endif

      N_("D-BUS support"),

#ifdef HAVE_HAL
      N_("HAL support"),
#endif

#ifdef HAVE_ALSA
      N_("ALSA Support (Linux)"),
#endif

#ifdef HAVE_SUN
      N_("SUN Audio support")
#endif

    };

  void
  print_version ()
  {
    g_print ("%s %s",
             PACKAGE,
             VERSION);

    if (std::strlen (RV_REVISION))
      g_print (" (%s-R%s)",
               RV_LAST_CHANGED_DATE,
               RV_REVISION);

    g_print (_("\nCopyright (c) 2005-2006 BMP Project <http://www.beep-media-player.org>\n\n"));

    g_print (_("Built on %s for %s with:\n"),
             BUILD_DATE,
             BUILD_ARCH);

    for (unsigned int i = 0; i < G_N_ELEMENTS (features); i++)
      g_print ("* %s\n", _(features[i]));

    g_print ("\n");
  }

  int
  create_path (const char *name)
  {
    const mode_t mode755 = S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;

    if (Glib::file_test (name, Glib::FILE_TEST_IS_DIR))
      return 0;

    int mkdir_rv = g_mkdir_with_parents (name, mode755);

    if (mkdir_rv)
      {
        int errsv = errno;
        g_critical ("%s: Unable to create %s: %s", G_STRFUNC, name, strerror(errsv));
      }

    return mkdir_rv;
  }

  void
  create_paths ()
  {
    const mode_t mode755 = S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
    int          sanity  = 0;

    std::string cache_dir = Glib::build_filename (g_get_user_cache_dir (), BMP_XDG_BASEDIR);

    if (!g_file_test (cache_dir.c_str (), G_FILE_TEST_IS_DIR))
      {
        sanity |= g_mkdir_with_parents (cache_dir.c_str (), mode755);
        if (sanity != 0)
          {
            int errsv = errno;
            g_critical ("%s: Unable to create %s: %s", G_STRFUNC, cache_dir.c_str (), strerror(errsv));
          }
      }

    sanity |= create_path (BMP_PATH_USER_DIR);
    sanity |= create_path (BMP_PATH_DATA_DIR);

    sanity |= create_path (BMP_PATH_LOGS_DIR);
    sanity |= create_path (BMP_PATH_SKIN_DIR);

    sanity |= create_path (BMP_PATH_SKINS_THUMB_DIR);
    sanity |= create_path (BMP_PATH_COVER_THUMB_DIR);
    sanity |= create_path (BMP_PATH_LYRIC_CACHE_DIR);

    if (sanity)
      {
        g_error (_("%s: Unable to create one or more user directories used by BMP (please see log above). Aborting startup; please check permissions in your home directory"), G_STRFUNC);
        exit (EXIT_FAILURE);
      }
  }

  void
  free_paths ()
  {
    for (int i = 0; i < BMP_N_PATHS; i++)
      {
        g_free (bmp_paths[i]);
        bmp_paths[i] = 0;
      }
  }

#define USER_PATH(path) \
  g_build_filename (BMP_PATH_USER_DIR, path, NULL);

#define DATA_PATH(path) \
  g_build_filename (BMP_PATH_DATA_DIR, path, NULL);

#define XDG_CACHE_PATH(path) \
  g_build_filename (g_get_user_cache_dir(), BMP_XDG_BASEDIR, path, NULL);

  void
  init_paths ()
  {
    std::string culprit = Glib::build_filename (Glib::get_home_dir(), ".config");

    if (Glib::file_test (culprit, Glib::FILE_TEST_EXISTS) && 
        Glib::file_test (culprit, Glib::FILE_TEST_IS_REGULAR))
      {
        std::string destination = Glib::build_filename (Glib::get_home_dir(), ".config-moved-by-bmpx");

        try
          {
            Bmp::Util::copy_file (culprit.c_str(), destination.c_str());  
          }
        catch (std::exception& cxe)
          {
            g_critical ("%s: Unable to copy file to destination. Reason: '%s'. Source: '%s'. Destination: '%s'", G_STRLOC, cxe.what(), culprit.c_str(), destination.c_str());
          }

        g_unlink (culprit.c_str());
        g_message ("The file %s has been moved to %s to accomodate for the XDG user home config dir", culprit.c_str(), destination.c_str());
      }

    BMP_PATH_USER_DIR =
      g_build_filename (g_get_user_config_dir (), BMP_XDG_BASEDIR, NULL);
    BMP_PATH_DATA_DIR =
      g_build_filename (g_get_user_data_dir (),   BMP_XDG_BASEDIR, NULL);

    BMP_PATH_LOGS_DIR =
      USER_PATH(BMP_BASENAME_LOGS_DIR);
    BMP_PATH_SKIN_DIR =
      DATA_PATH(BMP_BASENAME_SKIN_DIR);

    BMP_PATH_SKINS_THUMB_DIR =
      XDG_CACHE_PATH(BMP_BASENAME_SKINS_THUMB_DIR);
    BMP_PATH_COVER_THUMB_DIR =
      XDG_CACHE_PATH(BMP_BASENAME_COVER_THUMB_DIR);
    BMP_PATH_LYRIC_CACHE_DIR =
      XDG_CACHE_PATH(BMP_BASENAME_LYRIC_CACHE_DIR);

    create_paths ();
  }

  void
  setup_i18n ()
  {
    gtk_set_locale ();
    bindtextdomain (PACKAGE, LOCALE_DIR);
    bind_textdomain_codeset (PACKAGE, "UTF-8");
    textdomain (PACKAGE);
  }

#ifdef HAVE_STARTUP_NOTIFICATION
  void
  sn_error_trap_push(SnDisplay *display, Display *xdisplay)
  {
    gdk_error_trap_push();
  }

  void
  sn_error_trap_pop(SnDisplay *display, Display *xdisplay)
  {
    gdk_error_trap_pop();
  }

  void
  startup_notification_complete()
  {
    Display *xdisplay;

    xdisplay = GDK_DISPLAY();
    sn_display = sn_display_new(xdisplay,
                                sn_error_trap_push,
                                sn_error_trap_pop);
    sn_context =
      sn_launchee_context_new_from_environment(sn_display,
                                               DefaultScreen(xdisplay));

    if (sn_context != 0)
      {
        sn_launchee_context_complete(sn_context);
        sn_launchee_context_unref(sn_context);

        sn_display_unref(sn_display);
      }
  }
#endif // HAVE_STARTUP_NOTIFICATION

  void
  shutdown_func_gtk (gpointer data)
  {
    gtk_main_quit ();
  }

  void
  startup_func_gtk (gpointer data)
  {
    gtk_main ();
  }

  void
  register_keys (void)
  {
    using namespace Mcs;

    mcs->register_domain ("bmp");
    mcs->register_domain ("audio");
    mcs->register_domain ("tracklist");
    mcs->register_domain ("lastfm");

    mcs->register_domain ("playlist-window");

    mcs->register_key ("playlist-window", "visible", false);
    mcs->register_key ("playlist-window", "width", 0); //FIXME:
    mcs->register_key ("playlist-window", "height",0); //FIXME:
    mcs->register_key ("playlist-window", "pos_x", 0);
    mcs->register_key ("playlist-window", "pos_y", 0);

    mcs->register_domain ("main-window");

    mcs->register_key ("main-window", "width", 0); //FIXME:
    mcs->register_key ("main-window", "height", 0); //FIXME:
    mcs->register_key ("main-window", "pos_x", 0);
    mcs->register_key ("main-window", "pos_y", 0);

    mcs->register_key ("bmp", "no-ui", false);
    mcs->register_key ("bmp", "ui-esc-trayconify", false);
    mcs->register_key ("bmp", "keep-above", false);
    mcs->register_key ("bmp", "file-chooser-close-on-open", true);
    mcs->register_key ("bmp", "file-chooser-close-on-add", false);
    mcs->register_key ("bmp", "file-chooser-path", std::string (g_get_home_dir ()));
    mcs->register_key ("bmp", "file-chooser-path-relocate-album", std::string (g_get_home_dir ()));
    mcs->register_key ("bmp", "font", std::string ("Sans 12"));
    mcs->register_key ("bmp", "icon-theme", std::string ("red"));
    mcs->register_key ("bmp", "skin", std::string (DATA_DIR G_DIR_SEPARATOR_S BMP_BASENAME_SKIN_DIR G_DIR_SEPARATOR_S "default"));
    mcs->register_key ("bmp", "use-custom-cursors", true);
    mcs->register_key ("bmp", "display-tooltips", true);
    mcs->register_key ("bmp", "display-tracklist-numbers", true);
    mcs->register_key ("bmp", "library-view-size", 0);

    mcs->register_key  ("tracklist","column_0_visible", true);
    mcs->register_key  ("tracklist","column_1_visible", true);
    mcs->register_key  ("tracklist","column_2_visible", true);
    mcs->register_key  ("tracklist","column_3_visible", true);
    mcs->register_key  ("tracklist","column_4_visible", false);
    mcs->register_key  ("tracklist","column_5_visible", false);
    mcs->register_key  ("tracklist","column_6_visible", true);

    mcs->register_key  ("tracklist","column_0_width", 0.2);
    mcs->register_key  ("tracklist","column_1_width", 0.2);
    mcs->register_key  ("tracklist","column_2_width", 0.2);
    mcs->register_key  ("tracklist","column_3_width", 0.2);
    mcs->register_key  ("tracklist","column_4_width", 0.0);
    mcs->register_key  ("tracklist","column_5_width", 0.0);
    mcs->register_key  ("tracklist","column_6_width", 0.2);

#ifdef HAVE_LIBNOTIFY
    mcs->register_key ("bmp", "display-notifications", true);
#endif

    mcs->register_key ("bmp", "playlist-show-dnd-menu", false);

    mcs->register_key ("bmp", "no-remote", false);

    mcs->register_key ("bmp", "physically-delete-files", false);
    mcs->register_key ("bmp", "time-remaining", false);
    mcs->register_key ("bmp", "volume", 50);
    mcs->register_key ("bmp", "follow-current-track", false);
    mcs->register_key ("bmp", "resume-on-startup", false);
    mcs->register_key ("bmp", "resume-on-startup-track", 0);
    mcs->register_key ("bmp", "resume-on-startup-time", 0);
    mcs->register_key ("bmp", "shuffle", false);
    mcs->register_key ("bmp", "repeat", false);
    mcs->register_key ("bmp", "send-statistics", true);
    mcs->register_key ("bmp", "enable-autoplay", false);

#ifdef HAVE_MOOD
    mcs->register_key ("bmp", "enable-mood", false);
#endif

    mcs->register_key ("audio", "cdrom-device", std::string ("/dev/cdrom"));

    mcs->register_key ("audio", "sink", std::string (DEFAULT_SINK));

#ifdef HAVE_ALSA
    mcs->register_key ("audio", "alsa-buffer-time",200000);
    mcs->register_key ("audio", "device-alsa", std::string (DEFAULT_DEVICE_ALSA));
#endif

#ifdef HAVE_SUN
    mcs->register_key ("audio", "sun-buffer-time", 200000);
    mcs->register_key ("audio", "device-sun", std::string (DEFAULT_DEVICE_SUN));
#endif

    mcs->register_key ("audio", "oss-buffer-time",  200000);
    mcs->register_key ("audio", "device-oss", std::string (DEFAULT_DEVICE_OSS));

    mcs->register_key ("audio", "esd-buffer-time",  200000);
    mcs->register_key ("audio", "device-esd", std::string (DEFAULT_DEVICE_ESD));

    mcs->register_key ("lastfm", "queue-enable", false);
    mcs->register_key ("lastfm", "radio-enable", false);
    mcs->register_key ("lastfm", "submit-enable", false);

    mcs->register_key ("lastfm", "username", std::string (""));
    mcs->register_key ("lastfm", "password", std::string (""));

#ifdef HAVE_MOOD
    mcs->register_domain ("mood");
    mcs->register_key ("mood", "hostname", std::string ("localhost"));
    mcs->register_key ("mood", "port", std::string ("2240"));
    mcs->register_key ("mood", "password", std::string (""));
    mcs->register_key ("mood", "enqueue-finished", bool(false)); 
#endif
  }
} // anonymous namespace

int
main (int    argc,
      char** argv)
{
  GError              *error = 0;
  char                *sanity_error = 0;
  GOptionContext      *bmp_option_context = 0;

  sanity_error = sanity_check_libs ();
  if (sanity_error)
    {
      g_printerr ("%s\n", sanity_error);
      g_free (sanity_error);

      std::exit (EXIT_FAILURE);
    }

  g_type_init ();
  Glib::thread_init (0);
  Glib::init ();
  Bmp::debug_init ();
  setup_i18n ();
  init_paths ();

  //Attempt to run X11-less if we are --daemon or --no-ui
  bmp_option_context = g_option_context_new (_(" - Run BMPx"));
  g_option_context_add_main_entries (bmp_option_context, bmp_options, PACKAGE);

  g_option_context_add_group (bmp_option_context, gtk_get_option_group (false));
  g_option_context_add_group (bmp_option_context, gst_init_get_option_group ());

  if (!g_option_context_parse (bmp_option_context, &argc, &argv, &error))
    {
      g_printerr (_("Invalid argument: %s\n"), error->message);
      g_error_free (error);
      std::exit (EXIT_FAILURE);
    }

  startup = true;

  if (!no_remote)
    {
      dbus_g_type_specialized_init ();
    }

  if (detach)
    {
      if (!Bmp::Util::daemonize())
      {
        g_warning (_("Unable to detach, proceeding with normal startup."));
      }
    }

  if ((!run_headless) && (!g_getenv("DISPLAY")))
    {
      run_headless = true;
      forced_headless = true;
    }

#if 0
  Bmp::Playlist *pltest = new Bmp::Playlist();
  GParamSpec *pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (pltest->gobj()), "current");
  g_message ("VType: %s", g_type_name (pspec->value_type));
  g_message ("OType: %s", g_type_name (pspec->owner_type));
#endif

  if (show_version)
    {
      print_version ();
      goto main_early_exit;
    }

  if (!log_stdout)
    {
      logger = new Bmp::Logger;
    }

  if (!run_headless)
    {
      gtk_init (&argc, &argv);
      Gdk::wrap_init();
      Gtk::wrap_init();
    }

  try {
      mcs = new Mcs::Mcs (Glib::build_filename (BMP_PATH_USER_DIR, "config.xml"), "bmp", 0.20);
      register_keys ();
      mcs->load (Mcs::Mcs::VERSION_IGNORE);
    }
  catch (Mcs::Mcs::Exceptions& cxe)
    {
      if (cxe == Mcs::Mcs::PARSE_ERROR)
      {
        g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, _("Unable to parse configuration file"));
      }
    }

  if (no_resume)
    {
      mcs->key_set<bool>( "bmp","resume-on-startup", false);
    }

  Bmp::Plugins::init ();
  Signals::handlers_init ();

  /* FIXME: gtk_init_with_args() (above) will remove all recognized
     arguments, so the session manager will run BMPx with incomplete
     options when it restarts. */

  if (!run_headless)
    {
      Session::start (argc, argv, session_id);
#ifdef HAVE_STARTUP_NOTIFICATION
      gtk_window_set_auto_startup_notification (false);
#endif
    }

  bmp_playlist = bmp_playlist_new ();
  bmp_system_control = bmp_system_control_new (no_remote);

#ifdef ENABLE_TESTING_CODE
  //core = new Bmp::ServiceCore();
#endif

  if (!no_remote)
    {
      bmp_system_control_initialize_dbus_connection (bmp_system_control);
    }

  if (bmp_system_control)
    {
      vfs             = new Bmp::VFS::VFS;
      monitor         = new Bmp::FileMonitor::Monitor;
#ifdef HAVE_HAL
      hal             = new Bmp::Library::HAL;
#endif
      play            = new Bmp::Play;
      library         = new Bmp::Library::Library;
      bookmarks       = new Bmp::Streams::Bookmarks;
      lastfm          = new Bmp::LastFM::Service;
      lastfm_radio    = new Bmp::LastFM::Radio;

#ifdef HAVE_MOOD
      mood            = new Moo::Driver (mcs->key_get<std::string>("mood", "hostname"),
                                         mcs->key_get<std::string>("mood", "port"),
                                         mcs->key_get<std::string>("mood", "password"));
#endif

      bmp_system_control_init_finalize (bmp_system_control);
      bmp_playlist_load_playlist (bmp_playlist);
      bmp_playlist_connect_library (bmp_playlist);

      if (!run_headless)
        {
          bmp_system_control_ui_start (bmp_system_control, 0, 0);
        }
      else
        {
          if (!forced_headless)
            {
              g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Not starting UI on request.");
            }
          else
            {
              g_printerr (_("No X11 server found (DISPLAY environment variable not set): Not starting UI\n"));
            }
        }

      if (!run_headless)
        {
#ifdef HAVE_STARTUP_NOTIFICATION
          startup_notification_complete ();
#endif
        }

      //Initialize with GTK+2
      bmp_system_control_register_startup_func  (bmp_system_control,
                                                 (SystemStartupFunc) startup_func_gtk,
                                                 0);

      bmp_system_control_register_shutdown_func (bmp_system_control,
                                                 (SystemShutdownFunc) shutdown_func_gtk,
                                                 0);

      startup = false;
      bmp_system_control_application_run (bmp_system_control);
    }
  else
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "BmpSystemControl not present, cannot proceed with startup sequence.");
    }

  if (!run_headless)
    {
      Session::end ();
    }

  g_object_unref (bmp_playlist);

  Bmp::Plugins::shutdown ();

  delete monitor;
  delete play;
  delete lastfm;
  delete vfs;
  delete bookmarks;
  delete library;

#ifdef HAVE_HAL
  delete hal;
#endif

#ifdef HAVE_MOOD
  mood->disconnect ();
  delete mood;
#endif

  g_object_unref (bmp_system_control);

  delete mcs;
  delete logger;

 main_early_exit:

  free_paths ();
  std::exit (EXIT_SUCCESS);
}
