//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 2
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <gtkmm.h>
#include <dbus/dbus-glib.h>

#include "notify.hpp"
#include "amazon.hpp"

#include "debug.hpp"
#include "paths.hpp"

#include <libnotify/notify.h>

namespace Bmp
{
    Notify::Notify (GtkWidget *attach) : attach (attach), visible (false)
    {
      // Init libnotify
      notify_init (PACKAGE);

      // Make sure notify-daemon is running
      DBusGConnection *bus;
      DBusGProxy      *obj;
      GError          *error = NULL;
      char            *introspect_data;

      bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
      if (!bus)
        {
          debug ("notify", "Unable to get session bus");
          throw UNABLE_TO_DBUS_CONNECT;
        }

      obj = dbus_g_proxy_new_for_name (bus,
                                       "org.freedesktop.Notifications", 
                                       "/org/freedesktop/Notifications", 
                                       DBUS_INTERFACE_INTROSPECTABLE);

      if (!dbus_g_proxy_call (obj, "Introspect", &error,
                          G_TYPE_INVALID,
                          G_TYPE_STRING, &introspect_data, G_TYPE_INVALID))
        {
          debug ("notify", "Unable to introspect");
          throw UNABLE_TO_INTROSPECT_INTERFACE;
        }

      free (introspect_data);

      dbus_g_connection_unref (bus);     
      g_object_unref (obj);

      // Load default icon
      const std::string filename = BMP_THEME_ICON_DIR G_DIR_SEPARATOR_S "red" G_DIR_SEPARATOR_S "icon_player_32.png";
      default_icon = Gdk::Pixbuf::create_from_file (filename);

      n = notify_notification_new ("(none)",
			                             "(none)", 
                                   "(none)",
                                   NULL); 

      if (attach) notify_notification_attach_to_widget (n, attach);

      notify_notification_set_icon_from_pixbuf (n, default_icon->gobj());
      g_signal_connect (G_OBJECT (n), "closed", G_CALLBACK (Notify::notification_closed), this);
    }

    Notify::~Notify ()
    {
      g_object_unref (n);
      notify_uninit ();
    }

    void
    Notify::notification_closed (NotifyNotification *n, Bmp::Notify *notify)
    {
      notify->visible = false;	
      g_object_unref (n);

      notify->n = notify_notification_new ("(none)",
				                                   "(none)", 
                                           "(none)",
                                            NULL); 

      if (notify->attach) notify_notification_attach_to_widget (notify->n, notify->attach);

      notify_notification_set_icon_from_pixbuf (notify->n, notify->default_icon->gobj());
      g_signal_connect (G_OBJECT (notify->n), "closed", G_CALLBACK (Notify::notification_closed), notify);
    }

    bool 
    Notify::set_notification_actual         (Glib::ustring const& summary,
                                             Glib::ustring const& message,
                                             unsigned int         timeout,
                                             Glib::ustring&       uri)
    { 
      GError *error = 0;
      notify_notification_update (n, (!summary.empty() ? (summary).c_str () : ""),
				  (!message.empty() ? (message).c_str () : ""),
				  NULL);
      notify_notification_set_timeout (n, timeout * 1000);

      try {
        Glib::RefPtr<Gdk::Pixbuf> cover;
        Bmp::Amazon::get_cover (uri, cover, true);
        notify_notification_set_icon_from_pixbuf (n, cover->gobj()); 
      } catch (...) {
        notify_notification_set_icon_from_pixbuf (n, default_icon->gobj()); 
      }
      notify_notification_show (n, &error);
      visible = true;
      return false;
    }

    bool 
    Notify::set_notification_actual_default (Glib::ustring const& summary,
                                             Glib::ustring const& message,
                                             unsigned int         timeout)
    { 
      GError *error = 0;
      notify_notification_update (n, (!summary.empty() ? (summary).c_str () : ""),
				  (!message.empty() ? (message).c_str () : ""),
				  NULL);
      notify_notification_set_timeout (n, timeout * 1000);
      notify_notification_set_icon_from_pixbuf (n, default_icon->gobj()); 
      notify_notification_show (n, &error);
      visible = true;
      return false;
    }

    void
    Notify::set_notification (Glib::ustring const& summary,
			                        Glib::ustring const& message,
			                        unsigned int	       timeout) // We multiply it by 1k so we can just pass seconds here
    {
      Glib::signal_idle().connect (sigc::bind (sigc::mem_fun (*this, &Bmp::Notify::set_notification_actual_default),
                                               summary, message, timeout));
    }

    void
    Notify::set_notification (Glib::ustring const& summary,
                              Glib::ustring const& message,
			                        unsigned int	       timeout, // We multiply it by 1k so we can just pass seconds here
			                        Glib::ustring&       uri) 
    {
      Glib::signal_idle().connect (sigc::bind (sigc::mem_fun (*this, &Bmp::Notify::set_notification_actual),
                                               summary, message, timeout, uri));

    }

    void
    Notify::attach_to_widget (GtkWidget *widget)
    {
      attach = widget;
    }

} // ns::Bmp
