//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtkmm.h>
#include <glibmm.h>
#include <iostream>
#include <sstream>

#include <musicbrainz/musicbrainz.h>
#include <boost/format.hpp>
#include <boost/optional.hpp>
#include <chroma/chroma-list.h>
#include <chroma/chroma-list-selection.h>
#include <widgets/bmp_tray_icon.h>
#include <widgets/bmp_status_icon.h>

#include <goa/libgoa.h>

#ifdef HAVE_LIBNOTIFY
#  include <libnotify/notify.h>
#  include "notify.hpp"
#endif

#include "paths.hpp"
#include "main.hpp"

#include "x_library.hpp"
#include "x_play.hpp"
#include "x_playlist.hpp"
#include "x_system_control.hpp"

#include "filebrowser.hpp"
#include "trackinfo.hpp"
#include "preferences.hpp"
#include "streams.hpp"
#include "library_ui.hpp"
#include "about_dialog.hpp"
#include "export.hpp"
#include "network.hpp"
#include "lastfm_recommend.hpp"

#ifdef HAVE_GAIM
#  include "gaim_dbus.hpp"
#  include "gaim_send_file.hpp"
#endif

#ifdef HAVE_MOOD
#  include "x_mood.hpp"
#  include "mood_dialog.hpp"
#endif

#include "ui_util.hpp"
#include "winamp2_skin.hpp"
#include "winamp2_main.hpp"
#include "winamp2_playlist.hpp"

#include "shell.hpp"
#include "stock.hpp"
#include "action_defines.hpp"

namespace Bmp
{
  Glib::RefPtr<Gtk::UIManager>     ui_manager;
  Glib::RefPtr<Gtk::ActionGroup>   actions;
  Glib::RefPtr<Gtk::ActionGroup>   actions_toggle; 

  Bmp::LibraryUI    *library_ui = 0;
  BmpWinamp2Skin    *winamp2_skin = 0;
  BmpWindowMain     *w_main = 0;
  BmpWindowPlaylist *w_playlist = 0;
  Mcs::Bind         *mcs_bind = 0;

#ifdef HAVE_LIBNOTIFY
  Notify            *notify = 0;
#endif

#ifdef HAVE_GAIM
  Gaim::DBUS        *gaim_dbus = 0;
#endif

  bool skin_changing = false;
}

namespace
{
  bool running = false;
  int  msg_domain_id = 0;

  enum DnDTypes
  {
    TRAY_DROP_URLENCODED,
    TRAY_DROP_QUERY,
    TRAY_DROP_STRING,
    TRAY_DROP_TEXT_PLAIN,
  };

  //Drag data format listing for gtk_drag_dest_set()
  const GtkTargetEntry tray_drop_types[] = 
  {
        {"text/uri-list",		                  0, TRAY_DROP_URLENCODED},
        {"application/x-media-library-query", 0, TRAY_DROP_QUERY  },
        {"STRING",                            0, TRAY_DROP_STRING  },
        {"text/plain",                        0, TRAY_DROP_TEXT_PLAIN  },
  };

  // Data
  boost::optional<int> x_main;
  boost::optional<int> y_main;
  boost::optional<int> x_pl;
  boost::optional<int> y_pl;
  boost::optional<int> w_pl;
  boost::optional<int> h_pl;

  // Objects  
  Bmp::FileBrowser              *browser          = 0;
  Bmp::TrackInfo                *trackinfo        = 0;
  Bmp::Preferences              *preferences      = 0;
  Bmp::Streams::Dialog          *streams          = 0;
  Bmp::AboutDialog              *about            = 0;
  Bmp::ExportDialog             *export_dialog    = 0;
#ifdef HAVE_MOOD
  Bmp::MooDialog                *moo_dialog       = 0;
#endif
  Bmp::LastFM::RecommendDialog  *lastfm_recommend = 0;

  BmpStatusIcon           *status_icon;
  GdkPixbuf               *status_icon_image_default;
  GdkPixbuf               *status_icon_image_paused;
  GdkPixbuf               *status_icon_image_playing;

  /*------------------------------------------------------------------------------------------------*/
  // Tray Icon Callbacks
  void
  status_icon_click (BmpStatusIcon *icon, gpointer data)
  {
    Glib::RefPtr<Gtk::ToggleAction> action = Glib::RefPtr<Gtk::ToggleAction>::cast_static(Bmp::actions_toggle->get_action (BMP_TOGGLE_ACTION_SHOW_UI));
    action->set_active (!action->get_active());
  }

  void
  status_icon_scroll_up (BmpStatusIcon *icon, gpointer data)
  {
    int volume;
    volume = mcs->key_get<int>("bmp", "volume"); 
    volume += 4;
    volume = (volume > 100) ? 100 : volume;
    mcs->key_set<int>("bmp", "volume", volume);
  }

  void
  status_icon_scroll_down (BmpStatusIcon *icon, gpointer data)
  {
    int volume;
    volume = mcs->key_get<int>("bmp", "volume"); 
    volume -= 4;
    volume = (volume < 0) ? 0 : volume;
    mcs->key_set<int>("bmp", "volume", volume);
  }

  void
  status_icon_popup_menu (BmpStatusIcon *icon, guint arg1, guint arg2, gpointer data)
  {
    GtkWidget *menu;
    menu = ui_manager_get_popup (Bmp::ui_manager->gobj(), "/popup-tray/menu-tray");
    gtk_widget_realize (menu);
    gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, arg1, arg2);
  }

  void
  tray_icon_drag_data_received (GtkWidget         *widget,
                                GdkDragContext    *context,
                                int                x,
                                int                y,
                                GtkSelectionData  *selection_data,
                                guint              info,
                                guint              time_,
                                gpointer           user_data)
  {
    char     **strv;
    char      *data;

    g_return_if_fail(selection_data != NULL);

    if (!selection_data->data)
      {
          gtk_drag_finish (context, FALSE, FALSE, time_);
          return;
      }

    data = g_strdup((char*)selection_data->data);
    if (info == TRAY_DROP_QUERY)
      {
        strv = g_new0 (char*,2);
        strv[0] = g_strdup (data);
        strv[1] = NULL;
      }
    else
      {
        g_strstrip (data);
        strv = g_uri_list_extract_uris (data);
      }

    gtk_drag_finish (context, TRUE, FALSE, time_);

    int out = 0;
    bmp_system_control_add_uri_list (bmp_system_control,
                (const char**)strv,
                0,
                TRUE,
                TRUE,
                0,
                &out,
                NULL);

    g_strfreev (strv);
  }

  /*------------------------------------------------------------------------------------------------*/
  // SystemControl Callbacks

  void
  on_play_set_title (char *title)
  {
#ifdef HAVE_LIBNOTIFY
    if (Bmp::notify && mcs->key_get<bool>("bmp","display-notifications")) Bmp::notify->set_notification ("BMP", Glib::Markup::escape_text(title), 10);
      
#endif
    bmp_status_icon_set_text (status_icon, Glib::Markup::escape_text (title).c_str());
  }
 
  void
  on_system_control_track_change (BmpSystemControl *control)
  {
    GtkListStore            *playlist = bmp_playlist_get_playlist (bmp_playlist);
    GtkTreeRowReference     *reference;
    GtkTreePath             *path;
    GtkTreeIter              iter;
    char                    *artist, *album, *title, *uri;

    reference = static_cast<GtkTreeRowReference *> (g_object_get_data (G_OBJECT(playlist), "current"));
    if (!reference) 
      {
        bmp_status_icon_set_text (status_icon, ""); 
        return;
      }

    path = gtk_tree_row_reference_get_path (reference);
    if (!path) 
      {
        bmp_status_icon_set_text (status_icon, ""); 
        return;
      }

    gtk_tree_model_get_iter (GTK_TREE_MODEL (playlist), &iter, path);
    gtk_tree_path_free (path);

    gtk_tree_model_get (GTK_TREE_MODEL (playlist),
            &iter,
            COLUMN_ARTIST, &artist,
            COLUMN_ALBUM, &album,
            COLUMN_TITLE, &title,
            COLUMN_URI, &uri, -1);

    Glib::ustring status_icon_text;
    status_icon_text    .append ("<b>") 
                        .append (Glib::Markup::escape_text(artist ? artist : "")) 
                        .append ("</b>\n") 
                        .append ("<i>")
                        .append (Glib::Markup::escape_text(album ? album : ""))
                        .append ("</i>\n")
                        .append (Glib::Markup::escape_text (title ? title : ""));

#ifdef HAVE_LIBNOTIFY
    if (Bmp::notify && mcs->key_get<bool>("bmp","display-notifications"))
      {
        try {
          Glib::ustring u (uri);
          Bmp::notify->set_notification ("BMP", status_icon_text, 10, u);
        } catch (...) {
          Bmp::notify->set_notification ("BMP", status_icon_text, 10);
        }
      }
#endif

    free (artist);
    free (album);
    free (title);
    free (uri);
    bmp_status_icon_set_text (status_icon, status_icon_text.c_str());
  }

  void
  set_ui_playstatus (Bmp::Playstatus playstatus)
  {
    using namespace Bmp;

    const char *stock_id;

      if (playstatus == PLAYSTATUS_PLAYING)
        {
            bmp_status_icon_set_from_pixbuf (status_icon,
                                             status_icon_image_playing);
        }
      else
      if (playstatus == PLAYSTATUS_PAUSED)
        {
            bmp_status_icon_set_from_pixbuf (status_icon,
                                             status_icon_image_paused);
        }
      else
        {
           bmp_status_icon_set_from_pixbuf  (status_icon,
                                             status_icon_image_default);
        }

      switch (playstatus)
        {
          case PLAYSTATUS_STOPPED: stock_id = GTK_STOCK_MEDIA_STOP; break;
          case PLAYSTATUS_PAUSED:  stock_id = GTK_STOCK_MEDIA_PAUSE; break;
          case PLAYSTATUS_PLAYING: stock_id = GTK_STOCK_MEDIA_PLAY; break;
          default: stock_id = GTK_STOCK_MEDIA_STOP; break;
        }

      bmp_status_icon_set_icon (status_icon, stock_id);

      if (playstatus == PLAYSTATUS_STOPPED)
        {
          bmp_status_icon_set_text (status_icon, "");
        }
  }

  void
  on_system_control_set_playstatus (BmpSystemControl *system_control,
                                    Bmp::Playstatus   status)
  {
    set_ui_playstatus (status);
  }

  void
  set_ui_title    (const char  *title)
  {
    char *window_title;

    if (!title)
        window_title = g_strdup_printf ("BMP");
    else
        window_title = g_strdup_printf ("(%s) - BMP", title);

    gtk_window_set_title (GTK_WINDOW (Bmp::w_main->window), window_title);
    free (window_title);
  }

  void
  on_system_control_notify_title (BmpSystemControl *control)
  {
    char *title;
    bmp_system_control_get_current_title (control, &title, NULL);
    set_ui_title (title);
    free (title);
  }

  void
  on_system_control_app_busy (BmpSystemControl *control)
  {
    bmp_window_set_busy (Bmp::w_playlist->window);
    bmp_window_set_busy (Bmp::w_main->window);
  }

  void
  on_system_control_app_idle (BmpSystemControl *control)
  {
    bmp_window_set_idle (Bmp::w_playlist->window);
    bmp_window_set_idle (Bmp::w_main->window);
  }


  /*------------------------------------------------------------------------------------------------*/
  // MCS Callbacks

  void
  on_bmp_skin_changed (MCS_CB_DEFAULT_SIGNATURE)
  {
    std::string path = boost::get<std::string> (value);

    Bmp::skin_changing = true;

    if (!bmp_winamp2_skin_load_skin (Bmp::winamp2_skin, path))
      {
        mcs->key_unset ("bmp", "skin");
      }

    bmp_window_main_configure (Bmp::w_main);
    bmp_window_playlist_configure (Bmp::w_playlist);

    Bmp::skin_changing = false;

    gtk_widget_queue_draw (GTK_WIDGET(Bmp::w_main->window));
    gtk_widget_queue_draw (GTK_WIDGET(Bmp::w_playlist->window));
  }

  /*------------------------------------------------------------------------------------------------*/

  // Callbacks
  void
  action_play_files (GtkAction *action, gpointer data)
  {
    browser->run (Bmp::FileBrowser::PLAY);
  }

  void
  action_add_files  (GtkAction *action, gpointer data)
  {
    browser->run (Bmp::FileBrowser::ADD);
  }

  /*------------------------------------------------------------------------------------------------*/

  void
  action_about      (GtkAction *action, gpointer data)
  {
    about->present (); 
  }

  void
  action_library    (GtkAction *action, gpointer data)
  {
    Bmp::library_ui->present (); 
  }

#ifdef HAVE_MOOD
  void
  action_moo_dialog (GtkAction *action, gpointer data)
  {
    moo_dialog->present (); 
  }
#endif

  void
  action_streams    (GtkAction *action, gpointer data)
  {
    streams->present (); 
  }

  void
  action_preferences
            (GtkAction *action, gpointer data)
  {
    preferences->present (); 
  }

  /*------------------------------------------------------------------------------------------------*/

  void
  action_iconify    (GtkAction *action, gpointer data)
  {
    if (mcs->key_get<bool>("bmp", "ui-esc-trayconify"))
    {
        Glib::RefPtr<Gtk::ToggleAction>::cast_static(Bmp::actions_toggle->get_action (BMP_TOGGLE_ACTION_SHOW_UI))->set_active (false);
    }
    else
    {
        gtk_window_iconify (GTK_WINDOW (Bmp::w_main->window));
    }
  }

  void
  action_quit       (GtkAction *action, gpointer data)
  {
    bmp_system_control_quit (bmp_system_control, NULL);
  }

  /*------------------------------------------------------------------------------------------------*/

  void
  action_remove     (GtkAction *action, gpointer data)
  {
    chroma_list_selection_remove_selected (CHROMA_LIST_SELECTION (wTracklist));
    bmp_system_control_update (bmp_system_control);
  }

  void
  action_remove_all  (GtkAction *action, gpointer data)
  {
    bmp_playlist_remove_all (bmp_playlist);
    bmp_system_control_clear_playback_history (bmp_system_control);
    g_object_set (G_OBJECT (bmp_system_control), "current_index", int(-1), NULL);
    bmp_system_control_update (bmp_system_control);
  }

  void
  action_select_all (GtkAction *action, gpointer data)
  {
    chroma_list_selection_select_all (CHROMA_LIST_SELECTION (wTracklist));
  }

  void
  action_clear_playback_history
              (GtkAction *action, gpointer data)
  {
    bmp_system_control_clear_playback_history (bmp_system_control);
    gtk_action_set_sensitive (action, FALSE);
  }

  /*------------------------------------------------------------------------------------------------*/

  void
  action_export_tracklist
            (GtkAction *action, gpointer data)
  {
      int response = export_dialog->run ();
      switch (response)
      {
    case GTK_RESPONSE_CANCEL: break;

    case GTK_RESPONSE_OK:
      {
        bmp_system_control_export_tracklist (bmp_system_control,
                         export_dialog->container.c_str(),
                         export_dialog->all_files, 
                         export_dialog->entry_filename->get_text().c_str(), 
                         NULL);

      }
      }
      export_dialog->hide ();
  }

  void
  action_trackinfo  (GtkAction *action, gpointer data)
  {
    GtkListStore   *list;
    GtkTreeIter     iter;
    GtkTreePath    *path;
    char           *uri;

    path = chroma_list_selection_get_selected (CHROMA_LIST_SELECTION (wTracklist));
    if (!path) return;

    list = bmp_playlist_get_playlist (bmp_playlist);
    gtk_tree_model_get_iter (GTK_TREE_MODEL(list), &iter, path);
    gtk_tree_model_get (GTK_TREE_MODEL(list), &iter, COLUMN_URI, &uri, -1);
    gtk_tree_path_free (path);

    trackinfo->show (uri);
    free (uri);
  }

#ifdef HAVE_GAIM
  void
  action_send_file (GtkAction *action, gpointer data)
  {
    GtkTreePath *path = chroma_list_selection_get_selected (CHROMA_LIST_SELECTION (wTracklist)); 
    GtkListStore *store = bmp_playlist_get_playlist (bmp_playlist);
    GtkTreeIter iter;
    char *uri;
    std::string filename;
    gtk_tree_model_get_iter (GTK_TREE_MODEL (store), &iter, path);
    gtk_tree_path_free (path);
    gtk_tree_model_get (GTK_TREE_MODEL (store), &iter, COLUMN_URI, &uri, -1);

    try
      {
        Bmp::URI u (uri);

        if (u.get_protocol () != Bmp::URI::PROTOCOL_FILE)
          {
            free (uri);
            Gtk::MessageDialog dialog (_("You can only send locally accessible files trough Gaim!"), false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
            dialog.set_title (_("Send File: Error - BMP"));
            dialog.run ();
            return;
          }
      }
    catch (...) {} // FIXME: 

    filename = Glib::filename_from_uri (uri);

    try
      {
        Bmp::Gaim::SendFile *dialog = Bmp::Gaim::SendFile::create ();
        int response = dialog->run (Bmp::gaim_dbus->get_online_buddies(), uri);
        free (uri);

        if (response == GTK_RESPONSE_OK)
          {
            Bmp::gaim_dbus->send_file ( dialog->get_current_buddy () , filename );
          }
        delete dialog;
      }
    catch (Bmp::Gaim::DBUS::Exception& cxe)
      {
          Glib::ustring message (_("An Error occured while communicating with Gaim: "));
          message.append (cxe.what());
          Gtk::MessageDialog dialog (message, false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
          dialog.set_title (_("Send File: Error - BMP"));
          dialog.run ();
      }
    catch (...) //FIXME: What else could happen from our side..?!
      {
      }
  }
#endif
 
  void
  action_trackinfo_clear
             (GtkAction *action, gpointer data)
  {
    trackinfo->clear ();
  }

  void
  action_trackinfo_if_visible
             (GtkAction *action, gpointer data)
  {
    if (trackinfo->is_visible ())
    {
      Bmp::actions->get_action (BMP_ACTION_FILEINFO)->activate ();
    }
  }

  void
  action_recommend
             (GtkAction *action, gpointer data)
  {
    GtkTreePath  *path  = chroma_list_selection_get_selected (CHROMA_LIST_SELECTION (wTracklist)); 
    GtkListStore *store = bmp_playlist_get_playlist (bmp_playlist);
    GtkTreeIter   iter;
    gchar        *uri;
    gtk_tree_model_get_iter (GTK_TREE_MODEL (store), &iter, path);
    gtk_tree_path_free (path);
    gtk_tree_model_get (GTK_TREE_MODEL (store), &iter, COLUMN_URI, &uri, -1);
    lastfm_recommend->run (uri);
    g_free (uri);
  }

  /*------------------------------------------------------------------------------------------------*/

  bool
  get_cdinfo (std::string& cdindex, int& n_tracks)
  {
    MusicBrainz o;
    bool ret;

    o.UseUTF8 (true); 
    o.SetDevice (mcs->key_get<std::string>("audio", "cdrom-device"));

    ret = o.Query (MBQ_GetCDTOC);
    if (!ret) return false;

    cdindex  = o.Data (MBE_TOCGetCDIndexId);
    n_tracks = o.DataInt (MBE_TOCGetLastTrack);
    return true;
  }

  void
  action_add_cd (GtkAction *action, gpointer data)
  {
    int         n_tracks;
    std::string         cdindex;
    static boost::format cdda ("cdda://%s/%d");

    if (!get_cdinfo(cdindex, n_tracks)) return;

    char **uris = g_new0 (char*, n_tracks+1);
    try 
      {
        library->add_audiocd (cdindex);

        for (int n = 1; n <= n_tracks; n++)
          {
            uris[n-1] = strdup((cdda % cdindex % n).str().c_str()); 
          }

        uris[n_tracks] = 0;

        int out=0;
        bmp_system_control_add_uri_list (bmp_system_control,
                         (const gchar**)uris,
                         -1,
                         false,
                         false,
                         0,
                         &out,
                         NULL);
        g_strfreev (uris);
      }
    catch (Bmp::Library::Exception& cxe) {}
  }

  /*------------------------------------------------------------------------------------------------*/

  void
  action_prev   (GtkAction *action, gpointer data)
  {
    bmp_system_control_go_prev (bmp_system_control, NULL);
  }

  void 
  action_next   (GtkAction *action, gpointer data)
  {
    bmp_system_control_go_next (bmp_system_control, NULL);
  }

  void 
  action_play   (GtkAction *action, gpointer data)
  {
    bmp_system_control_play (bmp_system_control, NULL);
  }

  void 
  action_stop   (GtkAction *action, gpointer data)
  {
    bmp_system_control_stop (bmp_system_control, NULL);
  }
  
  void
  action_pause  (GtkAction *action, gpointer data)
  {
    bmp_system_control_pause (bmp_system_control, NULL);
  }

  void
  action_dnd_play
                (GtkAction *action, gpointer data)
  {
    bmp_window_playlist_dnd_txn_play (Bmp::w_playlist);
  }

  void
  action_dnd_append
                (GtkAction *action, gpointer data)
  {
    bmp_window_playlist_dnd_txn_append (Bmp::w_playlist);
  }

  void
  action_dnd_cancel
                (GtkAction *action, gpointer data)
  {
    bmp_window_playlist_dnd_txn_cancel (Bmp::w_playlist);
  }

  /*------------------------------------------------------------------------------------------------*/

  void
  action_toggle_show_ui
        (GtkToggleAction *action,
                    gpointer data)
  {
    static bool visible = false;

    if (gtk_toggle_action_get_active (action))
      {
        if (x_main && y_main)
          {
            gtk_window_move  (GTK_WINDOW (Bmp::w_main->window), x_main.get(), y_main.get());
          }
        gtk_window_present (GTK_WINDOW (Bmp::w_main->window));

        if (visible)
          {
            if (x_pl && y_pl)
              {
                gtk_window_move  (GTK_WINDOW (Bmp::w_playlist->window), x_pl.get(), y_pl.get());
              }
            gtk_window_present (GTK_WINDOW (Bmp::w_playlist->window));
          }

        gtk_window_set_skip_taskbar_hint
                                        (GTK_WINDOW (Bmp::w_main->window),
                                             false);
        gtk_window_set_skip_pager_hint  (GTK_WINDOW (Bmp::w_main->window),
                                             false);

        gtk_window_set_skip_taskbar_hint
                                        (GTK_WINDOW (Bmp::w_playlist->window),
                                             false);
        gtk_window_set_skip_pager_hint  (GTK_WINDOW (Bmp::w_playlist->window),
                                             false);
      }
    else
      {
        int _x_main, _y_main;
        int _x_pl,   _y_pl;

            gtk_window_get_position (GTK_WINDOW (Bmp::w_main->window),
                     &_x_main,
                     &_y_main);

            gtk_window_get_position (GTK_WINDOW (Bmp::w_playlist->window),
                     &_x_pl,
                     &_y_pl);

        x_pl = _x_pl;
        y_pl = _y_pl;

        x_main = _x_main;
        y_main = _y_main;

        gtk_widget_hide (GTK_WIDGET (Bmp::w_main->window));

    
        visible = GTK_WIDGET_VISIBLE (Bmp::w_playlist->window);
        if (visible)
          {
            gtk_widget_hide (Bmp::w_playlist->window);
          }

        gtk_window_set_skip_taskbar_hint
                                        (GTK_WINDOW (Bmp::w_main->window),
                                             true);
        gtk_window_set_skip_pager_hint  (GTK_WINDOW (Bmp::w_main->window),
                                             true);

        gtk_window_set_skip_taskbar_hint
                                        (GTK_WINDOW (Bmp::w_playlist->window),
                                             true);
        gtk_window_set_skip_pager_hint  (GTK_WINDOW (Bmp::w_playlist->window),
                                             true);
      }
  }

  void
  action_toggle_stop_after_current
        (GtkToggleAction *action,
                    gpointer data)
  {
    bmp_system_control_stop_after_current_set (bmp_system_control, gtk_toggle_action_get_active (action), NULL);
  }


  void
  action_toggle_shuffle
        (GtkToggleAction *action,
                    gpointer data)
  {
    bmp_system_control_shuffle_set (bmp_system_control, gtk_toggle_action_get_active (action), NULL);
  }

  void
  action_toggle_repeat
        (GtkToggleAction *action,
                    gpointer data)
  {
    bmp_system_control_repeat_set (bmp_system_control, gtk_toggle_action_get_active (action), NULL);
  }

  void
  action_toggle_keep_above
        (GtkToggleAction *action,
                    gpointer data)
  {
    bool keep_above = gtk_toggle_action_get_active (action);

    gtk_window_set_keep_above (GTK_WINDOW (Bmp::w_main->window), keep_above);
    gtk_window_set_keep_above (GTK_WINDOW (Bmp::w_playlist->window), keep_above);
  }

  void
  action_toggle_pl
        (GtkToggleAction *action,
                    gpointer data)
  {
    if (!gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)))
      {
        int x, y, w, h;
        gtk_window_get_position (GTK_WINDOW (Bmp::w_playlist->window), &x, &y);
        gtk_window_get_size (GTK_WINDOW (Bmp::w_playlist->window), &w, &h);
        g_object_set_data (G_OBJECT (Bmp::w_playlist->window), "x", GINT_TO_POINTER (x));
        g_object_set_data (G_OBJECT (Bmp::w_playlist->window), "y", GINT_TO_POINTER (y));
        g_object_set_data (G_OBJECT (Bmp::w_playlist->window), "w", GINT_TO_POINTER (w));
        g_object_set_data (G_OBJECT (Bmp::w_playlist->window), "h", GINT_TO_POINTER (h));
        gtk_widget_hide (Bmp::w_playlist->window);
      }
    else
      {
        int x = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (Bmp::w_playlist->window), "x"));
        int y = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (Bmp::w_playlist->window), "y"));
        int w = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (Bmp::w_playlist->window), "w"));
        int h = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (Bmp::w_playlist->window), "h"));

    gtk_window_resize (GTK_WINDOW (Bmp::w_playlist->window), w, h);

        if ((x > 0) && (y > 0))
          {
            gtk_window_move (GTK_WINDOW (Bmp::w_playlist->window), x, y);
          }
        gtk_widget_show_all (Bmp::w_playlist->window);
      }
  }

  /*------------------------------------------------------------------------------------------------*/

  // ActionEntries
  GtkActionEntry bmp_action_entries[] = {

      { "dummy", NULL, "dummy" },

      { BMP_ACTION_DND_CANCEL,
        GTK_STOCK_CANCEL,
        N_("Cancel"),
        NULL, 
        N_("Cancel"),
        G_CALLBACK (action_dnd_cancel)},

      { BMP_ACTION_DND_PLAY,
        GTK_STOCK_MEDIA_PLAY,
        N_("Play"),
        NULL, 
        N_("Play"),
        G_CALLBACK (action_dnd_play)},

      { BMP_ACTION_DND_APPEND,
        GTK_STOCK_ADD,
        N_("Append"),
        NULL, 
        N_("Append"),
        G_CALLBACK (action_dnd_append)},

      { BMP_ACTION_OPEN,
        GTK_STOCK_MEDIA_PLAY,
        N_("Play Files..."),
        "<ctrl>o",
        N_("Play Files..."),
        G_CALLBACK (action_play_files)},

      { BMP_ACTION_ADD,
        GTK_STOCK_ADD,
        N_("Add Files..."),
        "<ctrl>f",
        NULL,
        G_CALLBACK (action_add_files)},

      { BMP_ACTION_ADD_CD,
        GTK_STOCK_CDROM,
        N_("Add CD..."),
        "<ctrl>d",
        NULL,
        G_CALLBACK (action_add_cd)},



      { BMP_ACTION_PREV,
        GTK_STOCK_MEDIA_PREVIOUS,
        N_("Previous"),
        NULL,
        N_("Previous"),
        G_CALLBACK (action_prev)},

      { BMP_ACTION_PLAY,
        GTK_STOCK_MEDIA_PLAY,
        N_("Play"),
        NULL,
        N_("Play"),
        G_CALLBACK (action_play)},

      { BMP_ACTION_PAUSE,
        GTK_STOCK_MEDIA_PAUSE,
        N_("Pause"),
        "space",
        N_("Pause"),
        G_CALLBACK (action_pause)},

      { BMP_ACTION_STOP,
        GTK_STOCK_MEDIA_STOP,
        N_("Stop"),
        NULL,
        N_("Stop"),
        G_CALLBACK (action_stop)},

      { BMP_ACTION_NEXT,
        GTK_STOCK_MEDIA_NEXT,
        N_("Next"),
        NULL,
        N_("Next"),
        G_CALLBACK (action_next)},



      { BMP_ACTION_ICONIFY,
        NULL,
        N_("Minimize BMP"),
        "<alt>m",
        N_("Minimize"),
        G_CALLBACK (action_iconify)},

      { BMP_ACTION_QUIT,
        GTK_STOCK_QUIT,
        N_("Quit BMP"),
        "<control>w",
        N_("Quit BMP"),
        G_CALLBACK (action_quit)},

      { BMP_ACTION_QUIT_TITLE,
        GTK_STOCK_QUIT,
        "", "",
        "",
        G_CALLBACK (action_quit)},



      { BMP_ACTION_ABOUT,
        BMP_STOCK_ABOUT,
        N_("About BMP"),
        NULL,
        N_("About BMP"),
        G_CALLBACK (action_about)},

      { BMP_ACTION_PREFERENCES,
        GTK_STOCK_PREFERENCES,
        N_("Preferences"),
        "<ctrl>p",
        N_("Preferences"),
        G_CALLBACK (action_preferences)},

      { BMP_ACTION_LIBRARY,
        BMP_STOCK_LIBRARY,
        N_("Music Library"),
        "<ctrl>l",
        N_("Music Library"),
        G_CALLBACK (action_library)},

#ifdef HAVE_MOOD
      { BMP_ACTION_MUSEEK,
        BMP_STOCK_MUSEEK,
        N_("Downloads"),
        "<ctrl>d",
        N_("Downloads"),
        G_CALLBACK (action_moo_dialog)},
#endif

      { BMP_ACTION_STREAMS,
        BMP_STOCK_STREAMS,
        N_("Radio Streams"),
        "<ctrl>s",
        N_("Radio Streams"),
        G_CALLBACK (action_streams)},

      { BMP_ACTION_EXPORT,
        GTK_STOCK_SAVE,
        N_("Export Tracklist"),
        "<ctrl>e",
        NULL,
        G_CALLBACK (action_export_tracklist)},

#ifdef HAVE_GAIM
      { BMP_ACTION_SEND_TRACK,
        BMP_STOCK_GAIM,
        N_("Send File to GAIM Buddy..."),
        NULL, 
        NULL,
        G_CALLBACK (action_send_file)},
#endif

      { BMP_ACTION_LASTFM_RECOMMEND,
        BMP_STOCK_LASTFM,
        N_("Recommend to Last.FM Friend..."),
        NULL, 
        NULL,
        G_CALLBACK (action_recommend)},

      { BMP_ACTION_FILEINFO,
        GTK_STOCK_INFO,
        N_("Track Information"),
        "<ctrl>i",
        NULL,
        G_CALLBACK (action_trackinfo)},

      { BMP_ACTION_FILEINFO_CLEAR,
        GTK_STOCK_DELETE,
        N_("Clear Fileinfo"), 
        NULL, 
        NULL,
        G_CALLBACK (action_trackinfo_clear)},

      { BMP_ACTION_FILEINFO_IF_VISIBLE,
        GTK_STOCK_DELETE,
        N_("Clear Fileinfo If Visible"), 
        NULL, 
        NULL,
        G_CALLBACK (action_trackinfo_if_visible)},

    #if 0
      { BMP_ACTION_ADD_TO_NEW,
        GTK_STOCK_ADD,
        N_("Add Files to New..."),
        NULL,   
        NULL,
        G_CALLBACK (NULL)},
    #endif

      { BMP_ACTION_DEL,
        GTK_STOCK_DELETE,
        N_("Remove Selected"),
        "<ctrl>x",
        NULL,
        G_CALLBACK (action_remove)},

    #if 0
      { BMP_ACTION_CROP,
        GTK_STOCK_CUT,
        N_("Remove Unselected"),
        NULL, 
        NULL,
        G_CALLBACK (NULL)},
    #endif

      { BMP_ACTION_DEL_ALL,
        GTK_STOCK_CLEAR,
        N_("Remove All"),
        NULL,
        NULL,
        G_CALLBACK (action_remove_all)},

      { BMP_ACTION_CLEAR_PLAYBACK_HISTORY,
        GTK_STOCK_UNDO,
        N_("Clear Playback History"),
        "<ctrl>h",
        NULL,
        G_CALLBACK (action_clear_playback_history)},

    #if 0
      { BMP_ACTION_JTT,
        GTK_STOCK_JUMP_TO,
        N_("Jump To Track"),
        "<ctrl>j",
        NULL,
        G_CALLBACK (NULL)},
    #endif

      { BMP_ACTION_SELECT_ALL,
        NULL,
        N_("Select All"),
        "<control>a",
        NULL,
        G_CALLBACK (action_select_all) },

    #if 0
      { BMP_ACTION_SELECT_NONE,
        NULL,
        N_("Select None"),
        "<control><shift>a",
        NULL,
        G_CALLBACK (NULL) },

      { BMP_ACTION_SELECT_INVERT,
        NULL,
        N_("Invert Selection"),
        NULL,
        NULL,
        G_CALLBACK (NULL)},

      { BMP_ACTION_SORT_VISIBLE,
        NULL,
        N_("Sort by Visible Title"),
        "<shift>s",
        NULL,
        G_CALLBACK (NULL)},

      { BMP_ACTION_SORT_ARTIST,
        NULL,
        N_("Sort by Artist"),
        NULL,
        NULL,
        G_CALLBACK (NULL)},

      { BMP_ACTION_SORT_ALBUM,
        NULL,
        N_("Sort by Album"),
        NULL,
        NULL,
        G_CALLBACK (NULL)},

      { BMP_ACTION_SORT_TITLE,
        NULL,
        N_("Sort by Title"),
        NULL,
        NULL,
        G_CALLBACK (NULL)},

      { BMP_ACTION_SORT_GENRE,
        NULL,
        N_("Sort by Genre"),
        NULL,
        NULL,
        G_CALLBACK (NULL)},

      { BMP_ACTION_SORT_TRACK,
        NULL,
        N_("Sort by Tracknumber"),
        NULL,
        NULL,
        G_CALLBACK (NULL)},

      { BMP_ACTION_SORT_LOCATION,
        NULL,
        N_("Sort by Location"),
        NULL,
        NULL,
        G_CALLBACK (NULL)}
    #endif

  };

  GtkToggleActionEntry bmp_toggle_action_entries[] = {

      { BMP_TOGGLE_ACTION_FOLLOW_CURRENT,
        NULL,
        N_("Follow Currently Played Track"),
        NULL,
        NULL,
        G_CALLBACK (NULL),
        FALSE },

      { BMP_TOGGLE_ACTION_STOP_AFTER_CURRENT,
        NULL,
        N_("Stop After Current Track"),
        NULL,
        NULL,
        G_CALLBACK (action_toggle_stop_after_current),
        FALSE },

      { BMP_TOGGLE_ACTION_SHUFFLE,
        NULL,
        N_("_Shuffle"),
        NULL,
        N_("Switch between random and continuous playback"),
        G_CALLBACK (action_toggle_shuffle),
        FALSE },

      { BMP_TOGGLE_ACTION_REPEAT,
        NULL,
        N_("_Repeat"),
        NULL,
        N_("Whether to repeat the tracklist or not"),
        G_CALLBACK (action_toggle_repeat),
        FALSE },

      { BMP_TOGGLE_ACTION_PL,
        NULL,
        N_("Show Playlist Window"),
        NULL,
        N_("Toggles playlist window visibility"),
        G_CALLBACK (action_toggle_pl),
        FALSE },

      { BMP_TOGGLE_ACTION_SCROLL,
        NULL,
        N_("Scroll Textbox"),
        NULL,
        N_("Toggles textbox scroll"),
        G_CALLBACK (NULL),
        TRUE },

      { BMP_TOGGLE_ACTION_SHOW_UI,
        NULL,
        N_("Show UI"),
        NULL,
        N_("Toggle BMP UI Visiblity"),
        G_CALLBACK (action_toggle_show_ui),
        TRUE },

      { BMP_TOGGLE_ACTION_KEEP_ABOVE,
        NULL,
        N_("Keep BMP on top"),
        NULL,
        N_("Keep BMP on top"),
        G_CALLBACK (action_toggle_keep_above),
        FALSE },
  };

} // ns::<unnamed>

namespace Bmp
{
  void
  raise_windows ()
  {
    Glib::RefPtr<Gtk::ToggleAction>::cast_static (actions_toggle->get_action (BMP_TOGGLE_ACTION_SHOW_UI))->set_active (true);
    if (Glib::RefPtr<Gtk::ToggleAction>::cast_static (actions_toggle->get_action (BMP_TOGGLE_ACTION_PL))->get_active ())
      {
        gtk_window_present (GTK_WINDOW (Bmp::w_playlist->window));
      } 

    gtk_window_present (GTK_WINDOW (Bmp::w_main->window));
  }

  void
  display_windows ()
  {
    int x,y,w,h;
    bool visible;

    w = mcs->key_get<int>("playlist-window", "width");
    h = mcs->key_get<int>("playlist-window", "height");
    x = mcs->key_get<int>("playlist-window", "pos_x");
    y = mcs->key_get<int>("playlist-window", "pos_y");

    if ((x < 0) || (y < 0)) 
      {
        x = gdk_screen_get_width  (gdk_screen_get_default());
        y = gdk_screen_get_height (gdk_screen_get_default());
        x /= 2;
        y /= 2;
        x -= BMP_WINDOW_MAIN_WIDTH;
        y -= BMP_WINDOW_MAIN_HEIGHT;
      }

    g_object_set_data (G_OBJECT   (Bmp::w_playlist->window), "x", GINT_TO_POINTER (x));
    g_object_set_data (G_OBJECT   (Bmp::w_playlist->window), "y", GINT_TO_POINTER (y));
    g_object_set_data (G_OBJECT   (Bmp::w_playlist->window), "w", GINT_TO_POINTER (w));
    g_object_set_data (G_OBJECT   (Bmp::w_playlist->window), "h", GINT_TO_POINTER (h));

    visible = mcs->key_get<bool>("playlist-window", "visible");

    if (visible)
    {
        Glib::RefPtr<Gtk::ToggleAction>::cast_static(actions_toggle->get_action (BMP_TOGGLE_ACTION_PL))->set_active (true);
    }

    x = mcs->key_get<int>("main-window", "pos_x");
    y = mcs->key_get<int>("main-window", "pos_y");

    if ((x < 0) || (y < 0)) 
    {
        x = gdk_screen_get_width (gdk_screen_get_default());
        y = gdk_screen_get_height (gdk_screen_get_default());
        x /= 2;
        y /= 2;
        x -= BMP_WINDOW_MAIN_WIDTH/2;
        y -= BMP_WINDOW_MAIN_HEIGHT/2;

    }

    gtk_window_move   (GTK_WINDOW (Bmp::w_main->window), x, y);
    g_object_set_data (G_OBJECT   (Bmp::w_main->window), "x", GINT_TO_POINTER (x));
    g_object_set_data (G_OBJECT   (Bmp::w_main->window), "y", GINT_TO_POINTER (y));

    gtk_widget_show_all (Bmp::w_main->window);
  }

  void
  shell_init () 
  {
      if (running) return;

      GtkUIManager    *_ui_manager = gtk_ui_manager_new ();
      GtkActionGroup  *_actions;
      GtkActionGroup  *_actions_toggle;

      Bmp::register_stock_icons ();

      _actions = gtk_action_group_new (BMP_ACTION_GROUP);
      _actions_toggle = gtk_action_group_new (BMP_ACTION_GROUP_TOGGLE);

      gtk_action_group_set_translate_func (_actions,
                bmp_menu_translate,
                NULL,
                NULL);

      gtk_action_group_set_translate_func (_actions_toggle,
                bmp_menu_translate,
                NULL,
                NULL);
      
      gtk_ui_manager_insert_action_group (_ui_manager, _actions, 0);
      gtk_ui_manager_insert_action_group (_ui_manager, _actions_toggle, 0);
      gtk_action_group_set_sensitive (_actions, true);
      gtk_action_group_set_sensitive (_actions_toggle, true);

      gtk_action_group_add_actions    (_actions,
                       bmp_action_entries,
                       G_N_ELEMENTS (bmp_action_entries),
                       0);

      gtk_action_group_add_toggle_actions (_actions_toggle,
                       bmp_toggle_action_entries,
                       G_N_ELEMENTS (bmp_toggle_action_entries),
                       0);

      gtk_ui_manager_add_ui_from_file (_ui_manager, BMP_UI_DIR G_DIR_SEPARATOR_S "trayicon.ui",       NULL);
      gtk_ui_manager_add_ui_from_file (_ui_manager, BMP_UI_DIR G_DIR_SEPARATOR_S "tracklist.ui",      NULL);
      gtk_ui_manager_add_ui_from_file (_ui_manager, BMP_UI_DIR G_DIR_SEPARATOR_S "tracklist-dnd.ui",  NULL);
      gtk_ui_manager_add_ui_from_file (_ui_manager, BMP_UI_DIR G_DIR_SEPARATOR_S "main.ui",           NULL);

#ifdef HAVE_GAIM
      {
        guint merge_id = gtk_ui_manager_new_merge_id (_ui_manager);

        gtk_ui_manager_add_ui (_ui_manager,
                             merge_id, 
                             "/popup-tracklist/menu-tracklist/action-recommend-item",
                             BMP_ACTION_SEND_TRACK,
                             BMP_ACTION_SEND_TRACK,
                             GTK_UI_MANAGER_MENUITEM,
                             FALSE);

        gtk_ui_manager_add_ui (_ui_manager,
                             merge_id, 
                             "/popup-tracklist/menu-tracklist/" BMP_ACTION_SEND_TRACK,
                             "separator-send-track",
                             NULL,
                             GTK_UI_MANAGER_SEPARATOR,
                             FALSE);
      }
#endif

#ifdef HAVE_MOOD
      {
        guint merge_id = gtk_ui_manager_new_merge_id (_ui_manager);
        gtk_ui_manager_add_ui (_ui_manager,
                             merge_id, 
                             "/popup/menu-main/placeholder-museek",
                             BMP_ACTION_MUSEEK,
                             BMP_ACTION_MUSEEK,
                             GTK_UI_MANAGER_AUTO,
                             FALSE);

        gtk_ui_manager_add_ui (_ui_manager,
                             merge_id, 
                             "/popup-tracklist/menu-add/placeholder-museek",
                             BMP_ACTION_MUSEEK,
                             BMP_ACTION_MUSEEK,
                             GTK_UI_MANAGER_AUTO,
                             FALSE);

        gtk_ui_manager_add_ui (_ui_manager,
                             merge_id, 
                             "/popup-tray/menu-tray/placeholder-museek",
                             BMP_ACTION_MUSEEK,
                             BMP_ACTION_MUSEEK,
                             GTK_UI_MANAGER_AUTO,
                             FALSE);
      }
#endif

      ui_manager = Glib::wrap (_ui_manager);
      actions = Glib::wrap (_actions);
      actions_toggle = Glib::wrap (_actions_toggle);

      // Set up internet dependent actions 
      if (!Bmp::Network::is_connected ())
      {
        actions->get_action (BMP_ACTION_STREAMS)->set_sensitive (false);
      }

      // Init C++ objects needed for Winamp2.x skinning engine
      mcs_bind = new Mcs::Bind (mcs);

      // Acquire message domain id
      msg_domain_id = bmp_system_control_message_domain_register (bmp_system_control, "shell", "BMP");

      // Init Winamp2.x skinning system
      winamp2_skin  = bmp_winamp2_skin_new ();
      if (!bmp_winamp2_skin_load_skin (winamp2_skin, mcs->key_get<std::string>("bmp","skin")))
      {
        mcs->key_unset ("bmp", "skin");
        if (!bmp_winamp2_skin_load_skin (winamp2_skin, mcs->key_get<std::string>("bmp","skin")))
        {
          bmp_system_control_message_dispatch (bmp_system_control, msg_domain_id,
                               GTK_MESSAGE_ERROR, _("No skin could be loaded, check your BMP installation please!"));
          g_object_unref (winamp2_skin);
          delete mcs_bind;
          bmp_system_control_quit (bmp_system_control, NULL);
          exit (1); // Just to make sure in worst case; we can't proceed anyway
        }
      }
      w_main  = bmp_window_main_new ();
      w_playlist  = bmp_window_playlist_new ();
      mcs->subscribe ("Shell", "bmp", "skin", sigc::ptr_fun (&on_bmp_skin_changed));

      // Init C++ objects, the rest 
      browser = Bmp::FileBrowser::create ();
      trackinfo = Bmp::TrackInfo::create ();

      if (Bmp::Network::is_connected () )
        {
          streams = Bmp::Streams::Dialog::create ();
        }

      library_ui = Bmp::LibraryUI::create ();
      export_dialog = Bmp::ExportDialog::create ();

#ifdef HAVE_MOOD
      moo_dialog = Bmp::MooDialog::create();
#endif

      about = new Bmp::AboutDialog;

      // Tray icon
      status_icon_image_default =
        gdk_pixbuf_new_from_file (BMP_TRAY_ICON_DIR G_DIR_SEPARATOR_S "tray-icon-default.png", NULL);
      status_icon_image_paused =
        gdk_pixbuf_new_from_file (BMP_TRAY_ICON_DIR G_DIR_SEPARATOR_S "tray-icon-paused.png", NULL);
      status_icon_image_playing =
        gdk_pixbuf_new_from_file (BMP_TRAY_ICON_DIR G_DIR_SEPARATOR_S "tray-icon-playing.png", NULL);
      status_icon =
        bmp_status_icon_new_from_pixbuf (status_icon_image_default);

      bmp_status_icon_set_tooltip (status_icon, "", GTK_STOCK_MEDIA_STOP);
      bmp_status_icon_set_visible (status_icon, TRUE);

      g_object_connect (G_OBJECT (status_icon),

                        "signal::click",
                        G_CALLBACK(status_icon_click),
                        NULL,

                        "signal::popup-menu",
                        G_CALLBACK(status_icon_popup_menu),
                        NULL,

                        "signal::scroll-up",
                        G_CALLBACK(status_icon_scroll_up),
                        NULL,

                        "signal::scroll-down",
                        G_CALLBACK(status_icon_scroll_down),

                        NULL, NULL);

      GtkWidget *tray_icon = bmp_status_icon_get_tray_icon (status_icon);
      gtk_drag_dest_set (GTK_WIDGET (tray_icon),
		                     GtkDestDefaults (GTK_DEST_DEFAULT_MOTION | GTK_DEST_DEFAULT_DROP),
                         tray_drop_types, 4,
                         (GdkDragAction) (GDK_ACTION_MOVE | GDK_ACTION_COPY | GDK_ACTION_LINK | GDK_ACTION_ASK));
      g_object_connect (G_OBJECT (tray_icon),
                "signal::drag-data-received",
                G_CALLBACK(tray_icon_drag_data_received),
                NULL, NULL);

#ifdef HAVE_LIBNOTIFY
      try {
        notify = new Bmp::Notify (tray_icon);
      } catch (Bmp::NotifyException& cxe) {
        bmp_system_control_message_dispatch (bmp_system_control,
                                             msg_domain_id,
                                             GTK_MESSAGE_ERROR,
                                             _("BMP was unable to start and/or connect to notification-daemon."
                                               "Notification popups will not be available."));
        notify = 0;
      }
#endif

#ifdef HAVE_GAIM
      try {
        gaim_dbus = new Bmp::Gaim::DBUS ();
      } catch (Bmp::Gaim::DBUS::Exception& cxe) {
        gaim_dbus = 0;
        actions->get_action (BMP_ACTION_SEND_TRACK)->set_sensitive (false);
      }
#endif

      lastfm_recommend = Bmp::LastFM::RecommendDialog::create ();

      // SystemControl callbacks
      g_object_connect (bmp_system_control,

                        "signal::set-playstatus",
                        G_CALLBACK (on_system_control_set_playstatus),
                        NULL,

                       "signal::track-change",
                        G_CALLBACK (on_system_control_track_change),
                        NULL,

                        "signal::app-busy",
                        G_CALLBACK (on_system_control_app_busy),
                        NULL,

                        "signal::app-idle",
                        G_CALLBACK (on_system_control_app_idle),
                        NULL,

                        "signal::notify::current-title",
                        G_CALLBACK (on_system_control_notify_title),
                        NULL, NULL);

    display_windows ();
    preferences = Bmp::Preferences::create ();

    // MCS Bindings
    Glib::RefPtr<Gtk::ToggleAction> action;
    action = Glib::RefPtr<Gtk::ToggleAction>::cast_static
                           (actions_toggle->get_action (BMP_TOGGLE_ACTION_PL));
    mcs_bind->bind_toggle_action (action, "playlist-window", "visible");

    action = Glib::RefPtr<Gtk::ToggleAction>::cast_static
                           (actions_toggle->get_action (BMP_TOGGLE_ACTION_KEEP_ABOVE));
    mcs_bind->bind_toggle_action (action, "bmp", "keep-above");

    action = Glib::RefPtr<Gtk::ToggleAction>::cast_static
                           (actions_toggle->get_action (BMP_TOGGLE_ACTION_REPEAT));
    mcs_bind->bind_toggle_action (action, "bmp", "repeat");

    action = Glib::RefPtr<Gtk::ToggleAction>::cast_static
                           (actions_toggle->get_action (BMP_TOGGLE_ACTION_SHUFFLE));
    mcs_bind->bind_toggle_action (action, "bmp", "shuffle");


    set_ui_playstatus (Bmp::Playstatus (play->property_status().get_value()));

    char *title; 
    g_object_get (G_OBJECT (bmp_system_control), "current_title", &title, NULL);
    set_ui_title (title);
    free (title);

    play->signal_title().connect (sigc::ptr_fun (&on_play_set_title));

    running = true;
    g_message ("Running: Shell");
  }

  void
  shell_stop ()
  {
    int _x_main, _y_main;

    gtk_window_get_position (GTK_WINDOW (Bmp::w_main->window),
                 &_x_main,
                 &_y_main);

    mcs->key_set<int>("main-window", "pos_x", _x_main);
    mcs->key_set<int>("main-window", "pos_y", _y_main);

    int x, y, w, h;

    if (GTK_WIDGET_VISIBLE (Bmp::w_playlist->window))
      {
          gtk_window_get_position (GTK_WINDOW (Bmp::w_playlist->window),
                                    &x,
                                    &y);

          gtk_window_get_size     (GTK_WINDOW (Bmp::w_playlist->window),
                                    &w,
                                    &h);
      }
    else 
      {
        x = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (Bmp::w_playlist->window), "x"));
        y = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (Bmp::w_playlist->window), "y"));
        w = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (Bmp::w_playlist->window), "w"));
        h = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (Bmp::w_playlist->window), "h"));
      }

    mcs->key_set<int>("playlist-window", "width",  w);
    mcs->key_set<int>("playlist-window", "height", h);
    mcs->key_set<int>("playlist-window", "pos_x",  x);
    mcs->key_set<int>("playlist-window", "pos_y",  y);

    // Dispose C Objects
    g_object_unref (Bmp::w_main);
    g_object_unref (Bmp::w_playlist);
    g_object_unref (Bmp::winamp2_skin);

    // Delete C++ objects 
    delete  mcs_bind;
    delete  library_ui;
    delete  export_dialog;
    delete  about;
#ifdef HAVE_LIBNOTIFY
    delete  notify;
#endif
    delete  browser;
    delete  trackinfo;
    delete  streams;
    delete  preferences;

#ifdef HAVE_MOOD
    delete  moo_dialog;
#endif

    running = false;

    g_message ("Stopped: Shell");
  }

  bool
  shell_running ()
  {
    return running;
  }
}
