//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 2
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glibmm.h>
#include <gtkmm.h>
#include <libglademm.h>
#include <glib/gstdio.h>

#include <cstring>
#include <iostream>
#include <sstream>
#include <string>

#include <boost/format.hpp>
#include <boost/optional.hpp>

#include "util.hpp"
#include "uri++.hpp"
#include "database.hpp"
#include "library.hpp"
#include "network.hpp"
#include "paths.hpp"
#include "amazon.hpp"
#include "lyrics.hpp"
#include "file_utils.hpp"
#include "ui_util.hpp"

#include "trackinfo.hpp"
#include "x_library.hpp"

#include "main.hpp"
#include "xds.hpp"

namespace
{
    boost::format progress_text ("%d / %d"); 
}

namespace Bmp
{
    class EventBoxXDS : public Gtk::EventBox, public Bmp::XDS
    {
    public:
  
      EventBoxXDS () {}
      ~EventBoxXDS () {}

      virtual void
      on_drag_end (const Glib::RefPtr<Gdk::DragContext>& context)
      {
        xds_drag_end (context);
      }

      virtual void
      on_drag_begin  (const Glib::RefPtr<Gdk::DragContext>& context)
      {
        Glib::ustring dir_basename;
        dir_basename.append (artist.get());
        dir_basename.append (" - ");
        dir_basename.append (album.get());

        context->set_icon (pixbuf, 0, 0);       

        xds_drag_begin (context, dir_basename); 
      }

      virtual void
      on_drag_data_get (const Glib::RefPtr<Gdk::DragContext>& context, Gtk::SelectionData& selection_data, guint info, guint time)
      {
        xds_drag_data_get (context, selection_data, info, time);
      }

      void
      set_album  (Glib::ustring const& album)
      {
        this->album  = album;
      }

      void
      set_artist (Glib::ustring const& artist)
      {
        this->artist = artist;
      }

      void
      clear_artist_album ()
      {
        artist.reset ();
        album.reset ();
      }
    
      void
      set_pixbuf (Glib::RefPtr<Gdk::Pixbuf>& pixbuf)
      {
        this->pixbuf = pixbuf;
      }


      DragStart&
      signal_drag_start ()
      {
        return signal_drag_start_;
      }

      DragStep&
      signal_drag_step ()
      {
        return signal_drag_step_;
      }

      DragEnd&
      signal_drag_end ()
      {
        return signal_drag_end_;
      }

    protected:

      virtual void
      xds_drag_action     (Glib::ustring const& uri)
      {
        using namespace Bmp::DB;
        using namespace Bmp::Library;

        std::string id_artist = library->get_metadatum_id (DATUM_ARTIST);
        std::string id_album  = library->get_metadatum_id (DATUM_ALBUM);
        std::string id_location   = library->get_metadatum_id (DATUM_LOCATION);

        AttributeList attrs;
        VRow vector;
        attrs.push_back (Attribute (DB::EXACT, id_artist, artist.get()));
        attrs.push_back (Attribute (DB::EXACT, id_album,  album.get()));
        library->query (attrs, vector);
        if (vector.empty()) return;


        std::string destination_path = Glib::filename_from_uri (uri);
        if (g_mkdir (destination_path.c_str(), 0700)) return;
      
        signal_drag_start_.emit (vector.size());
        for (VRow::iterator vIter = vector.begin (), vEnd = vector.end (); vIter != vEnd; ++vIter)
          {
            Row& row = (*vIter); 
            std::string original = Glib::filename_from_uri (boost::get<Glib::ustring>(row[id_location]));
            std::string basename = Glib::path_get_basename (original);
            std::string filename = Glib::build_filename (destination_path, basename); 

            signal_drag_step_.emit ();
            try
              {
                Util::copy_file (original.c_str(), filename.c_str());
              }
            catch (std::exception& cxe)
              {
                g_critical ("%s: Unable to copy file to destination. Reason: '%s'. Source: '%s'. Destination: '%s'", G_STRLOC, cxe.what(), original.c_str(), filename.c_str());
              }
          }

        signal_drag_end_.emit ();
      }
    
      virtual bool
      xds_acceptable      (Glib::ustring const& uri)
      {
          if (!artist && !album) return false;
          return true;
      }

    private:
    
      DragStart signal_drag_start_;
      DragStep  signal_drag_step_;
      DragEnd   signal_drag_end_;
  
      boost::optional<Glib::ustring> artist;
      boost::optional<Glib::ustring> album;

      Glib::RefPtr<Gdk::Pixbuf> pixbuf;
  
    };
}

namespace Bmp
{

  TrackInfo::~TrackInfo ()
  {
    delete ebox;
  }

  TrackInfo *
  TrackInfo::create ()
  {
      const std::string path (BMP_GLADE_DIR G_DIR_SEPARATOR_S "trackinfo.glade");

      Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);

      TrackInfo *track_info = 0;
      glade_xml->get_widget_derived ("trackinfo", track_info);

      return track_info;
  }

  TrackInfo::TrackInfo (BaseObjectType                        *cobject,
                        const Glib::RefPtr<Gnome::Glade::Xml> &xml)
      : Gtk::Window (cobject),
        path_image_default (BMP_IMAGE_DIR G_DIR_SEPARATOR_S "audio.png"),
        ref_xml (xml),
        ok      (0),
        image   (0),
        lyrics  (0)
  {

      bmp_window_set_icon_list (GTK_WIDGET (this->gobj()), "player");

      ref_xml->get_widget ("ok", ok);
      ok->signal_clicked ().connect (sigc::mem_fun (this, &Bmp::TrackInfo::ok_clicked));

      // Load 'stars' images
      for (unsigned int n = 0; n < N_RATINGS; ++n)
        {
          static boost::format star ("stars_%d.png");
          rating[n] = Gdk::Pixbuf::create_from_file (Glib::build_filename (BMP_RATING_IMAGE_DIR, (star % n ).str())); 
        }

      image = new Gtk::Image;
      ebox  = new Bmp::EventBoxXDS;

      image->show_all ();
      image->set (path_image_default);
      ebox->add (*image);

      ebox->signal_drag_start().connect (sigc::mem_fun (*this, &Bmp::TrackInfo::on_drag_start));
      ebox->signal_drag_step().connect (sigc::mem_fun (*this, &Bmp::TrackInfo::on_drag_step));
      ebox->signal_drag_end().connect (sigc::mem_fun (*this, &Bmp::TrackInfo::on_drag_end));

      Gtk::Alignment *a;
      ref_xml->get_widget ("alignment19", a);
      a->add (*ebox);
      a->show_all ();

      ref_xml->get_widget ("lyrics", lyrics);
      ref_xml->get_widget ("copy_progress", copy_progress);

      ref_xml->get_widget ("entry_file",    entry_file);
      ref_xml->get_widget ("entry_artist",  entry_artist);
      ref_xml->get_widget ("entry_album",   entry_album);
      ref_xml->get_widget ("entry_title",   entry_title);
      ref_xml->get_widget ("entry_genre",   entry_genre);
      ref_xml->get_widget ("entry_comment", entry_comment);

      ref_xml->get_widget ("label_track",   label_track);
      ref_xml->get_widget ("label_date",    label_date);
      ref_xml->get_widget ("label_bitrate", label_bitrate);
      ref_xml->get_widget ("label_count",   label_count);

      ref_xml->get_widget ("i_rating", i_rating);
      ref_xml->get_widget ("evb_rating", evb_rating);

      ref_xml->get_widget ("table_mutable_info", table_mutable_info);

      evb_rating->signal_button_press_event().connect (sigc::mem_fun (*this, &Bmp::TrackInfo::rating_button_press));

      ref_xml->get_widget ("b_lyrics", b_lyrics);
      b_lyrics->signal_clicked().connect (sigc::mem_fun (*this, &Bmp::TrackInfo::get_lyrics));
      b_lyrics->set_sensitive (Bmp::Network::is_connected());
      library->signal_row_updated().connect (sigc::mem_fun (*this, &Bmp::TrackInfo::check_update));
  }

  bool
  TrackInfo::rating_button_press (GdkEventButton *event)
  {
    using namespace Library;

    int n = (int(event->x+7)/14);

    UpdateTracks update_tracks;
    UpdateTrack update_track;

    update_track.location_cur = location;
    update_track.rating = n;
    update_tracks.push_back (update_track);
    library->modify_tracks (update_tracks, false, true); 
    return false;
  }

  void
  TrackInfo::check_update (Glib::ustring const&         pkey,
                           Bmp::Library::UpdateTrack    track)
  {
    if (location && (location.get() == pkey) && this->is_visible())
    {
      location.reset ();
      if (track.location_new)
        {
          this->show (track.location_new.get());
        }
      else
        {
          this->show (track.location_cur.get());
        }
    }
  }
 
  void
  TrackInfo::ok_clicked ()
  {
      hide ();
  }

  void
  TrackInfo::get_lyrics ()
  {
      set_sensitive (false);

      Glib::ustring text;
      if (Bmp::Lyrictracker::get_lyrics (track, text))
      {
        Glib::RefPtr<Gtk::TextBuffer> buffer = lyrics->get_buffer ();
        buffer->set_text (text);
        lyrics->set_sensitive (true);
        b_lyrics->set_sensitive (false);
      }

      set_sensitive (true);
  }

  void
  TrackInfo::clear ()
  {
      entry_file->set_text ("");
      entry_file->set_sensitive (false);

      entry_artist->set_text ("");
      entry_artist->set_sensitive (false);

      entry_album->set_text ("");
      entry_album->set_sensitive (false);

      entry_title->set_text ("");
      entry_title->set_sensitive (false);

      entry_genre->set_text ("");
      entry_genre->set_sensitive (false);

      entry_comment->set_text ("");
      entry_comment->set_sensitive (false);

      label_track->set_text ("");
      label_track->set_sensitive (false);

      label_date->set_text ("");
      label_date->set_sensitive (false);

      label_bitrate->set_text ("");
      label_bitrate->set_sensitive (false);

      label_count->set_text ("");
      label_count->set_sensitive (false);

      i_rating->set (rating[0]);
      i_rating->set_sensitive (false);

      table_mutable_info->hide ();

      Glib::RefPtr<Gtk::TextBuffer> buffer = lyrics->get_buffer ();
      buffer->set_text ("");
      lyrics->set_sensitive (false);
      b_lyrics->set_sensitive (false);
      location.reset();

      ebox->clear_artist_album ();
      ebox->xds_unset ();
      Glib::RefPtr<Gdk::Pixbuf> pb_default = Gdk::Pixbuf::create_from_file (path_image_default);
      ebox->set_pixbuf (pb_default);

      image->set (path_image_default);
      image->drag_unhighlight ();
  }

  void
  TrackInfo::on_drag_start (int n_tracks)
  {
      set_sensitive (false);
      n_drag_tracks = n_tracks;
      n_drag_current = 0;
      copy_progress->show_all ();
      copy_progress->set_fraction (0.0);
      copy_progress->set_text ((progress_text % 0 % n_tracks).str());
  }

  void
  TrackInfo::on_drag_step  ()
  {
      n_drag_current++;
      copy_progress->set_fraction ((n_drag_current*1.)/(n_drag_tracks*1.));
      copy_progress->set_text ((progress_text % n_drag_current % n_drag_tracks).str());

      while (gtk_events_pending()) gtk_main_iteration ();
  }

  void
  TrackInfo::on_drag_end   ()
  {
      copy_progress->set_fraction (0.0);
      copy_progress->set_text ("");
      copy_progress->hide ();
      set_sensitive (true);
  }

  void
  TrackInfo::show (Glib::ustring const& uri)
  {
      Bmp::DB::Row row;
      try { library->get_metadata (uri, row); } catch (Bmp::Library::Exception& cxe) { return; }

      track = Bmp::Library::Track (row);

      if (location && (track.location.get() == location.get()))
        {
          Gtk::Window::show ();
          return;
        }

      clear ();

      location = track.location.get (); 

      if (library->is_library_track (location.get())) table_mutable_info->show_all();

      set_size_request (600, -1);
      set_keep_above (mcs->key_get<bool>("bmp","keep-above"));
      set_sensitive (false);
      Gtk::Window::show ();

      if (track.location)
        {
          Bmp::URI uri (track.location.get());
          uri.unescape ();
          entry_file->set_text (Glib::ustring(uri));
          entry_file->set_sensitive (true);
        }

      if (track.artist)
        {
          entry_artist->set_text (track.artist.get());
          entry_artist->set_sensitive (true);
          ebox->set_artist (track.artist.get());
        }

      if (track.album)
        {
          entry_album->set_text (track.album.get());
          entry_album->set_sensitive (true);
          ebox->set_album (track.album.get());
        }

      if (track.title)
        {
          entry_title->set_text (track.title.get());
          entry_title->set_sensitive (true);
        }

      if (track.genre)
        {
          entry_genre->set_text (track.genre.get());
          entry_genre->set_sensitive (true);
        }

      if (track.comment)
        {
          entry_comment->set_text (track.comment.get());
          entry_comment->set_sensitive (true);
        }
  
      if (track.date)
        {
          label_date->set_text ((boost::format ("%d") % track.date.get()).str ());
          label_date->set_sensitive (true);
        }

      if (track.tracknumber)
        {
          label_track->set_text ((boost::format ("%d") % track.tracknumber.get()).str ());
          label_track->set_sensitive (true);
        } 

      if (track.bitrate)
        {
          label_bitrate->set_text ((boost::format ("%d") % track.bitrate.get()).str () + " kbit/s");
          label_bitrate->set_sensitive (true);
        }

      if (track.count)
        {
          label_count->set_text ((boost::format ("%d") % track.count.get()).str ());
          label_count->set_sensitive (true);
        }

      if (track.rating)
        {
          i_rating->set (rating[track.rating.get()]);
          i_rating->set_sensitive (true);
        }

      if (track.artist && track.album)
        {
          ebox->xds_set ();
        }

      if (track.artist && track.title)
        {
          b_lyrics->set_sensitive (Bmp::Network::is_connected());
        }

      Glib::RefPtr<Gdk::Pixbuf> cover;
      try
        {
          Bmp::Amazon::get_cover (uri, cover);
          if (cover)
            {
              image->set (cover);
              image->drag_highlight ();
              ebox->set_pixbuf (cover);
              while (gtk_events_pending ()) gtk_main_iteration ();
            }
        }
      catch (...) {}

      Glib::ustring text;
      if (Bmp::Lyrictracker::get_lyrics (track, text, true))
        {
          Glib::RefPtr<Gtk::TextBuffer> buffer = lyrics->get_buffer ();
          buffer->set_text (text);
          lyrics->set_sensitive (true);
          if (Bmp::Network::is_connected ())
            {
              b_lyrics->set_sensitive (false);
            }
        }
     set_sensitive (true);
  }
} // namespace Bmp
